// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    config::CpuConfig,
    instruction::{
        mop_enum, AluBranchMOp, L2RegisterFileMOp, LoadStoreMOp, MOpTrait, UnitOutRegNum,
    },
    register::PRegValue,
    unit::unit_base::UnitForwardingInfo,
};
use fayalite::{
    bundle::{Bundle, BundleType},
    intern::{Intern, Interned},
    prelude::*,
    util::ready_valid::ReadyValid,
};

pub mod alu_branch;
pub mod unit_base;

macro_rules! all_units {
    (
        #[hdl_unit_kind = $HdlUnitKind:ident]
        #[unit_kind = $UnitKind:ident]
        #[hdl]
        $(#[$enum_meta:meta])*
        $vis:vis enum $UnitMOpEnum:ident<$DestReg:ident: Type, $SrcRegWidth:ident: Size> {
            $(
                #[create_dyn_unit_fn = $create_dyn_unit_fn:expr]
                #[extract = $extract:ident]
                $(#[$variant_meta:meta])*
                $Unit:ident($Op:ty),
            )*
        }
    ) => {
        $(#[$enum_meta])*
        #[derive(Copy, Clone, Eq, PartialEq, Ord, PartialOrd, Hash, Debug)]
        $vis enum $UnitKind {
            $(
                $(#[$variant_meta])*
                $Unit,
            )*
        }

        impl $UnitKind {
            pub fn unit(self, config: &CpuConfig, unit_index: usize) -> DynUnit {
                match self {
                    $($UnitKind::$Unit => $create_dyn_unit_fn(config, unit_index),)*
                }
            }
        }

        impl ToExpr for $UnitKind {
            type Type = $HdlUnitKind;

            fn to_expr(&self) -> Expr<Self::Type> {
                match self {
                    $($UnitKind::$Unit => $HdlUnitKind.$Unit(),)*
                }
            }
        }

        #[hdl]
        $(#[$enum_meta])*
        $vis enum $HdlUnitKind {
            $(
                $(#[$variant_meta])*
                $Unit,
            )*
        }

        mop_enum! {
            #[hdl]
            $(#[$enum_meta])*
            $vis enum $UnitMOpEnum<$DestReg: Type, $SrcRegWidth: Size> {
                $(
                    $(#[$variant_meta])*
                    $Unit($Op),
                )*
            }
        }

        impl<$DestReg: Type, $SrcRegWidth: Size> $UnitMOpEnum<$DestReg, $SrcRegWidth> {
            #[hdl]
            $vis fn kind(expr: impl ToExpr<Type = Self>) -> Expr<$HdlUnitKind> {
                #[hdl]
                let unit_kind = wire();
                #[hdl]
                match expr {
                    $($UnitMOpEnum::<_, _>::$Unit(_) => connect(unit_kind, $HdlUnitKind.$Unit()),)*
                }
                unit_kind
            }
            $(
                #[hdl]
                $vis fn $extract(expr: impl ToExpr<Type = Self>) -> Expr<HdlOption<$Op>> {
                    let expr = expr.to_expr();
                    let ty = Expr::ty(expr);
                    #[hdl]
                    let $extract = wire(HdlOption[ty.$Unit]);
                    connect($extract, HdlOption[ty.$Unit].HdlNone());
                    #[hdl]
                    if let $UnitMOpEnum::<_, _>::$Unit(v) = expr {
                        connect($extract, HdlSome(v));
                    }
                    $extract
                }
            )*
        }

        impl CpuConfig {
            #[hdl]
            pub fn available_units_for_kind(&self, unit_kind: impl ToExpr<Type = $HdlUnitKind>) -> Expr<Array<Bool>> {
                #[hdl]
                let available_units_for_kind = wire(Array[Bool][self.units.len()]);
                #[hdl]
                match unit_kind {
                    $($HdlUnitKind::$Unit => for (index, unit) in self.units.iter().enumerate() {
                        connect(available_units_for_kind[index], unit.kind == $UnitKind::$Unit);
                    })*
                }
                available_units_for_kind
            }
        }
    };
}

all_units! {
    #[hdl_unit_kind = HdlUnitKind]
    #[unit_kind = UnitKind]
    #[hdl]
    pub enum UnitMOp<DestReg: Type, SrcRegWidth: Size> {
        #[create_dyn_unit_fn = |config, unit_index| alu_branch::AluBranch::new(config, unit_index).to_dyn()]
        #[extract = alu_branch_mop]
        AluBranch(AluBranchMOp<DestReg, SrcRegWidth>),
        #[create_dyn_unit_fn = |config, unit_index| todo!()]
        #[extract = l2_register_file_mop]
        L2RegisterFile(L2RegisterFileMOp<DestReg, SrcRegWidth>),
        #[create_dyn_unit_fn = |config, unit_index| todo!()]
        #[extract = load_store_mop]
        LoadStore(LoadStoreMOp<DestReg, SrcRegWidth>),
    }
}

#[hdl]
pub struct UnitResultCompleted<ExtraOut> {
    pub value: PRegValue,
    pub extra_out: ExtraOut,
}

#[hdl]
pub struct UnitOutputWrite<OutRegNumWidth: Size> {
    pub which: UnitOutRegNum<OutRegNumWidth>,
    pub value: PRegValue,
}

#[hdl]
pub struct TrapData {
    // TODO
}

#[hdl]
pub enum UnitResult<ExtraOut> {
    Completed(UnitResultCompleted<ExtraOut>),
    Trap(TrapData),
}

#[hdl]
pub struct UnitOutput<OutRegNumWidth: Size, ExtraOut> {
    pub which: UnitOutRegNum<OutRegNumWidth>,
    pub result: UnitResult<ExtraOut>,
}

#[hdl]
pub struct UnitCancelInput<OutRegNumWidth: Size> {
    pub which: UnitOutRegNum<OutRegNumWidth>,
}

pub trait UnitTrait:
    'static + Send + Sync + std::fmt::Debug + fayalite::intern::SupportsPtrEqWithTypeId
{
    type Type: BundleType;
    type ExtraOut: Type;
    type MOp: Type;

    fn ty(&self) -> Self::Type;
    fn extra_out_ty(&self) -> Self::ExtraOut;
    fn mop_ty(&self) -> Self::MOp;

    fn unit_kind(&self) -> UnitKind;

    fn extract_mop(
        &self,
        mop: Expr<UnitMOp<UnitOutRegNum<DynSize>, DynSize>>,
    ) -> Expr<HdlOption<Self::MOp>>;

    fn module(&self) -> Interned<Module<Self::Type>>;

    fn input_insn(&self, this: Expr<Self::Type>) -> Expr<ReadyValid<Self::MOp>>;

    fn cancel_input(&self, this: Expr<Self::Type>) -> Expr<ReadyValid<UnitCancelInput<DynSize>>>;

    fn unit_forwarding_info(
        &self,
        this: Expr<Self::Type>,
    ) -> Expr<UnitForwardingInfo<DynSize, DynSize, DynSize>>;

    fn output(
        &self,
        this: Expr<Self::Type>,
    ) -> Expr<ReadyValid<UnitOutput<DynSize, Self::ExtraOut>>>;

    fn cd(&self, this: Expr<Self::Type>) -> Expr<ClockDomain>;

    fn to_dyn(&self) -> DynUnit;
}

type DynUnitTrait = dyn UnitTrait<Type = Bundle, ExtraOut = CanonicalType, MOp = CanonicalType>;

impl fayalite::intern::InternedCompare for DynUnitTrait {
    type InternedCompareKey = fayalite::intern::PtrEqWithTypeId;
    fn interned_compare_key_ref(this: &Self) -> Self::InternedCompareKey {
        Self::get_ptr_eq_with_type_id(this)
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct DynUnit {
    ty: Bundle,
    extra_out_ty: CanonicalType,
    mop_ty: CanonicalType,
    unit_kind: UnitKind,
    unit: Interned<DynUnitTrait>,
}

impl UnitTrait for DynUnit {
    type Type = Bundle;
    type ExtraOut = CanonicalType;
    type MOp = CanonicalType;

    fn ty(&self) -> Self::Type {
        self.ty
    }

    fn extra_out_ty(&self) -> Self::ExtraOut {
        self.extra_out_ty
    }

    fn mop_ty(&self) -> Self::MOp {
        self.mop_ty
    }

    fn unit_kind(&self) -> UnitKind {
        self.unit_kind
    }

    fn extract_mop(
        &self,
        mop: Expr<UnitMOp<UnitOutRegNum<DynSize>, DynSize>>,
    ) -> Expr<HdlOption<Self::MOp>> {
        self.unit.extract_mop(mop)
    }

    fn module(&self) -> Interned<Module<Self::Type>> {
        self.unit.module()
    }

    fn input_insn(&self, this: Expr<Self::Type>) -> Expr<ReadyValid<Self::MOp>> {
        self.unit.input_insn(this)
    }

    fn cancel_input(&self, this: Expr<Self::Type>) -> Expr<ReadyValid<UnitCancelInput<DynSize>>> {
        self.unit.cancel_input(this)
    }

    fn unit_forwarding_info(
        &self,
        this: Expr<Self::Type>,
    ) -> Expr<UnitForwardingInfo<DynSize, DynSize, DynSize>> {
        self.unit.unit_forwarding_info(this)
    }

    fn output(
        &self,
        this: Expr<Self::Type>,
    ) -> Expr<ReadyValid<UnitOutput<DynSize, Self::ExtraOut>>> {
        self.unit.output(this)
    }

    fn cd(&self, this: Expr<Self::Type>) -> Expr<ClockDomain> {
        self.unit.cd(this)
    }

    fn to_dyn(&self) -> DynUnit {
        *self
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct DynUnitWrapper<T>(pub T);

impl<T: UnitTrait + Clone + std::hash::Hash + Eq> UnitTrait for DynUnitWrapper<T> {
    type Type = Bundle;
    type ExtraOut = CanonicalType;
    type MOp = CanonicalType;

    fn ty(&self) -> Self::Type {
        Bundle::from_canonical(self.0.ty().canonical())
    }

    fn extra_out_ty(&self) -> Self::ExtraOut {
        self.0.extra_out_ty().canonical()
    }

    fn mop_ty(&self) -> Self::MOp {
        self.0.mop_ty().canonical()
    }

    fn unit_kind(&self) -> UnitKind {
        self.0.unit_kind()
    }

    fn extract_mop(
        &self,
        mop: Expr<UnitMOp<UnitOutRegNum<DynSize>, DynSize>>,
    ) -> Expr<HdlOption<Self::MOp>> {
        Expr::from_enum(Expr::as_enum(self.0.extract_mop(mop)))
    }

    fn module(&self) -> Interned<Module<Self::Type>> {
        self.0.module().canonical().intern_sized()
    }

    fn input_insn(&self, this: Expr<Self::Type>) -> Expr<ReadyValid<Self::MOp>> {
        Expr::from_bundle(Expr::as_bundle(self.0.input_insn(Expr::from_bundle(this))))
    }

    fn cancel_input(&self, this: Expr<Self::Type>) -> Expr<ReadyValid<UnitCancelInput<DynSize>>> {
        self.0.cancel_input(Expr::from_bundle(this))
    }

    fn unit_forwarding_info(
        &self,
        this: Expr<Self::Type>,
    ) -> Expr<UnitForwardingInfo<DynSize, DynSize, DynSize>> {
        self.0.unit_forwarding_info(Expr::from_bundle(this))
    }

    fn output(
        &self,
        this: Expr<Self::Type>,
    ) -> Expr<ReadyValid<UnitOutput<DynSize, Self::ExtraOut>>> {
        Expr::from_bundle(Expr::as_bundle(self.0.output(Expr::from_bundle(this))))
    }

    fn cd(&self, this: Expr<Self::Type>) -> Expr<ClockDomain> {
        self.0.cd(Expr::from_bundle(this))
    }

    fn to_dyn(&self) -> DynUnit {
        let unit = self.intern();
        DynUnit {
            ty: unit.ty(),
            extra_out_ty: unit.extra_out_ty(),
            mop_ty: unit.mop_ty(),
            unit_kind: unit.unit_kind(),
            unit: Interned::cast_unchecked(unit, |v: &Self| -> &DynUnitTrait { v }),
        }
    }
}
