// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    config::CpuConfig,
    instruction::{AluBranchMOp, UnitOutRegNum},
    unit::{
        unit_base::{unit_base, UnitForwardingInfo},
        DynUnit, DynUnitWrapper, UnitCancelInput, UnitKind, UnitMOp, UnitOutput, UnitTrait,
    },
};
use fayalite::{
    intern::{Intern, Interned},
    prelude::*,
    util::ready_valid::ReadyValid,
};

#[hdl_module]
pub fn alu_branch(config: &CpuConfig, unit_index: usize) {
    #[hdl]
    let cd: ClockDomain = m.input();
    #[hdl]
    let input_insn: ReadyValid<AluBranchMOp<UnitOutRegNum<DynSize>, DynSize>> =
        m.input(ReadyValid[AluBranchMOp[config.unit_out_reg_num()][config.p_reg_num_width()]]);
    #[hdl]
    let unit_forwarding_info: UnitForwardingInfo<DynSize, DynSize, DynSize> =
        m.input(config.unit_forwarding_info());
    #[hdl]
    let cancel_input: ReadyValid<UnitCancelInput<DynSize>> =
        m.input(ReadyValid[config.unit_cancel_input()]);
    #[hdl]
    let output: ReadyValid<UnitOutput<DynSize, ()>> =
        m.output(ReadyValid[UnitOutput[config.out_reg_num_width][()]]);
    #[hdl]
    let unit_base = instance(unit_base(
        config,
        unit_index,
        Expr::ty(input_insn).data.HdlSome,
    ));
    connect(unit_base.input_insn, input_insn);
    connect(unit_base.cd, cd);
    connect(unit_base.unit_forwarding_info, unit_forwarding_info);
    connect(unit_base.cancel_input, cancel_input);
    // TODO: finish
    connect(unit_base.ready_mop.ready, true);
    connect(output.data, Expr::ty(output.data).HdlNone());
}

#[derive(Debug, Copy, Clone, PartialEq, Eq, Hash)]
pub struct AluBranch {
    config: Interned<CpuConfig>,
    module: Interned<Module<alu_branch>>,
}

impl AluBranch {
    pub fn new(config: &CpuConfig, unit_index: usize) -> Self {
        Self {
            config: config.intern(),
            module: alu_branch(config, unit_index),
        }
    }
}

impl UnitTrait for AluBranch {
    type Type = alu_branch;
    type ExtraOut = ();
    type MOp = AluBranchMOp<UnitOutRegNum<DynSize>, DynSize>;

    fn ty(&self) -> Self::Type {
        self.module.io_ty()
    }

    fn extra_out_ty(&self) -> Self::ExtraOut {
        ()
    }

    fn mop_ty(&self) -> Self::MOp {
        self.module.io_ty().input_insn.data.HdlSome
    }

    fn unit_kind(&self) -> UnitKind {
        UnitKind::AluBranch
    }

    fn extract_mop(
        &self,
        mop: Expr<UnitMOp<UnitOutRegNum<DynSize>, DynSize>>,
    ) -> Expr<HdlOption<Self::MOp>> {
        UnitMOp::alu_branch_mop(mop)
    }

    fn module(&self) -> Interned<Module<Self::Type>> {
        self.module
    }

    fn input_insn(&self, this: Expr<Self::Type>) -> Expr<ReadyValid<Self::MOp>> {
        this.input_insn
    }

    fn cancel_input(&self, this: Expr<Self::Type>) -> Expr<ReadyValid<UnitCancelInput<DynSize>>> {
        this.cancel_input
    }

    fn unit_forwarding_info(
        &self,
        this: Expr<Self::Type>,
    ) -> Expr<UnitForwardingInfo<DynSize, DynSize, DynSize>> {
        this.unit_forwarding_info
    }

    fn output(
        &self,
        this: Expr<Self::Type>,
    ) -> Expr<ReadyValid<UnitOutput<DynSize, Self::ExtraOut>>> {
        this.output
    }

    fn cd(&self, this: Expr<Self::Type>) -> Expr<ClockDomain> {
        this.cd
    }

    fn to_dyn(&self) -> DynUnit {
        DynUnitWrapper(*self).to_dyn()
    }
}
