// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    config::CpuConfig,
    instruction::{MOpTrait, UnitOutRegNum, COMMON_MOP_SRC_LEN},
    register::PRegValue,
    unit::{UnitCancelInput, UnitOutputWrite},
    util::tree_reduce::tree_reduce,
};
use fayalite::{module::wire_with_loc, prelude::*, ty::StaticType, util::ready_valid::ReadyValid};
use std::marker::PhantomData;

#[hdl]
pub struct UnitForwardingInfo<UnitNumWidth: Size, OutRegNumWidth: Size, UnitCount: Size> {
    pub unit_output_writes: ArrayType<HdlOption<UnitOutputWrite<OutRegNumWidth>>, UnitCount>,
    pub _phantom: PhantomData<UnitNumWidth>,
}

#[hdl]
pub struct ReadyMOp<MOp: Type + MOpTrait<DestReg = UnitOutRegNum<DynSize>>> {
    pub mop: MOp,
    pub src_values: Array<PRegValue, { COMMON_MOP_SRC_LEN }>,
}

#[hdl]
struct InFlightOp<MOp: Type + MOpTrait<DestReg = UnitOutRegNum<DynSize>>> {
    pub mop: MOp,
    pub src_values: Array<HdlOption<PRegValue>, { COMMON_MOP_SRC_LEN }>,
}

#[hdl_module]
pub fn unit_base<MOp: Type + MOpTrait<DestReg = UnitOutRegNum<DynSize>>>(
    config: &CpuConfig,
    unit_index: usize,
    mop_ty: MOp,
) {
    #[hdl]
    let cd: ClockDomain = m.input();
    #[hdl]
    let unit_forwarding_info: UnitForwardingInfo<DynSize, DynSize, DynSize> =
        m.input(config.unit_forwarding_info());
    #[hdl]
    let input_insn: ReadyValid<MOp> = m.input(ReadyValid[mop_ty]);
    connect(input_insn.ready, false);
    #[hdl]
    let cancel_input: ReadyValid<UnitCancelInput<DynSize>> =
        m.input(ReadyValid[config.unit_cancel_input()]);
    connect(cancel_input.ready, true);
    #[hdl]
    let ready_mop: ReadyValid<ReadyMOp<MOp>> = m.output(ReadyValid[ReadyMOp[mop_ty]]);
    connect(ready_mop.data, Expr::ty(ready_mop.data).HdlNone());
    let max_in_flight = config.unit_max_in_flight(unit_index).get();
    #[hdl]
    let in_flight_ops = reg_builder().clock_domain(cd).reset(repeat(
        HdlOption[InFlightOp[mop_ty]].HdlNone(),
        max_in_flight,
    ));
    let in_flight_op_index_ty = UInt::range(0..max_in_flight);
    #[hdl]
    let input_index = wire(HdlOption[in_flight_op_index_ty]);
    connect(
        input_index,
        tree_reduce(
            (0..max_in_flight).map(|i| -> Expr<HdlOption<UInt>> {
                HdlOption::map(in_flight_ops[i], |_| i.cast_to(in_flight_op_index_ty))
            }),
            HdlOption::or,
        )
        .expect("max_in_flight is known to be non-zero"),
    );
    #[hdl]
    let input_in_flight_op = wire(HdlOption[InFlightOp[mop_ty]]);
    connect(input_in_flight_op, Expr::ty(input_in_flight_op).HdlNone());
    #[hdl]
    if let HdlSome(mop) = ReadyValid::firing_data(input_insn) {
        let src_values = wire_with_loc(
            "input_in_flight_op_src_values",
            SourceLocation::caller(),
            StaticType::TYPE,
        );
        connect(
            src_values,
            [HdlSome(PRegValue::zeroed()); COMMON_MOP_SRC_LEN],
        );
        MOp::for_each_src_reg(mop, &mut |src_reg, src_index| {
            #[hdl]
            if config
                .p_reg_num()
                .const_zero()
                .cast_to_bits()
                .cmp_ne(src_reg)
            {
                connect(src_values[src_index], HdlNone());
            }
        });
        connect(
            input_in_flight_op,
            HdlSome(
                #[hdl]
                InFlightOp::<_> { mop, src_values },
            ),
        );
    }
    for in_flight_op_index in 0..max_in_flight {
        #[hdl]
        if let HdlSome(in_flight_op) = in_flight_ops[in_flight_op_index] {
            #[hdl]
            if let HdlSome(cancel_input) = ReadyValid::firing_data(cancel_input) {
                #[hdl]
                let UnitCancelInput::<_> { which } = cancel_input;
                #[hdl]
                if which.value.cmp_eq(MOp::dest_reg(in_flight_op.mop).value) {
                    // TODO: if it needs extra time to cancel (e.g. still in pipeline), handle that here
                    connect(
                        in_flight_ops[in_flight_op_index],
                        HdlOption[InFlightOp[mop_ty]].HdlNone(),
                    );
                }
            }
            // TODO: finish
        } else if let HdlSome(input_index) = input_index {
            connect(input_insn.ready, true);
            #[hdl]
            if input_index.cmp_eq(in_flight_op_index) {
                connect(in_flight_ops[in_flight_op_index], input_in_flight_op);
            }
        }
    }
}
