// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use cpu::{
    config::{CpuConfig, UnitConfig},
    instruction::{
        AddSubMOp, AluCommonMOp, CommonMOp, LogicalMOp, MOp, MOpDestReg, MOpRegNum,
        OutputIntegerMode, COMMON_MOP_2_IMM_WIDTH, COMMON_MOP_3_IMM_WIDTH,
    },
    reg_alloc::{reg_alloc, FetchedDecodedMOp},
    unit::UnitKind,
};
use fayalite::{
    assert_export_firrtl,
    firrtl::ExportOptions,
    prelude::*,
    sim::{time::SimDuration, vcd::VcdWriterDecls, Simulation},
    util::RcWriter,
};
use std::num::NonZeroUsize;

#[hdl]
#[test]
fn test_reg_alloc() {
    let _n = SourceLocation::normalize_files_for_tests();
    let mut config = CpuConfig::new(vec![
        UnitConfig::new(UnitKind::AluBranch),
        UnitConfig::new(UnitKind::AluBranch),
    ]);
    config.fetch_width = NonZeroUsize::new(2).unwrap();
    let m = reg_alloc(&config);
    let mut sim = Simulation::new(m);
    let mut writer = RcWriter::default();
    sim.add_trace_writer(VcdWriterDecls::new(writer.clone()));
    let fetch_decode_interface = sim.io().fetch_decode_interface;
    sim.write_clock(sim.io().cd.clk, false);
    sim.write_reset(sim.io().cd.rst, true);
    sim.write_bool(fetch_decode_interface.fetch_decode_special_op.ready, true);
    sim.write(
        fetch_decode_interface.decoded_insns[0].data,
        HdlSome(
            #[hdl]
            FetchedDecodedMOp {
                mop: MOp.AluBranch(MOp.AluBranch.AddSub(
                    #[hdl]
                    AddSubMOp {
                        alu_common: #[hdl]
                        AluCommonMOp {
                            common: CommonMOp::new(
                                0_hdl_u0,
                                #[hdl]
                                MOpDestReg {
                                    normal_regs: #[hdl]
                                    [
                                        #[hdl]
                                        MOpRegNum { value: 1u8 },
                                        MOpRegNum::const_zero(),
                                    ],
                                    flag_regs: #[hdl]
                                    [HdlSome(()), HdlNone()],
                                },
                                [2u8, 3u8, 4u8],
                                0x1234.cast_to(SInt[COMMON_MOP_3_IMM_WIDTH]),
                            ),
                            output_integer_mode: OutputIntegerMode.Full64(),
                        },
                        invert_src0: true,
                        invert_carry_in: true,
                        invert_carry_out: true,
                        add_pc: true,
                    },
                )),
                is_unrelated_pc: true,
                pc: 0x1000_hdl_u64,
            },
        ),
    );
    sim.write(
        fetch_decode_interface.decoded_insns[1].data,
        HdlSome(
            #[hdl]
            FetchedDecodedMOp {
                mop: MOp.AluBranch(MOp.AluBranch.Logical(
                    #[hdl]
                    LogicalMOp {
                        alu_common: #[hdl]
                        AluCommonMOp {
                            common: CommonMOp::new(
                                0_hdl_u0,
                                #[hdl]
                                MOpDestReg {
                                    normal_regs: [
                                        #[hdl]
                                        MOpRegNum { value: 2u8 },
                                        MOpRegNum::const_zero(),
                                    ],
                                    flag_regs: [HdlNone(), HdlSome(())],
                                },
                                [3u8, 4u8],
                                SInt[COMMON_MOP_2_IMM_WIDTH].zero(),
                            ),
                            output_integer_mode: OutputIntegerMode.Full64(),
                        },
                        lut: 0b0110_hdl_u4,
                    },
                )),
                is_unrelated_pc: false,
                pc: 0x1004_hdl_u64,
            },
        ),
    );
    for cycle in 0..20 {
        sim.advance_time(SimDuration::from_nanos(500));
        sim.write_clock(sim.io().cd.clk, true);
        sim.advance_time(SimDuration::from_nanos(500));
        sim.write_clock(sim.io().cd.clk, false);
        sim.write_reset(sim.io().cd.rst, false);
    }
    // FIXME: vcd is just whatever reg_alloc does now, which isn't known to be correct
    let vcd = String::from_utf8(writer.take()).unwrap();
    println!("####### VCD:\n{vcd}\n#######");
    if vcd != include_str!("expected/reg_alloc.vcd") {
        panic!();
    }
    #[rustfmt::skip] // work around https://github.com/rust-lang/rustfmt/issues/6161
    assert_export_firrtl! {
        m =>
        options: ExportOptions {
            simplify_enums: None,
            ..ExportOptions::default()
        },
        "/test/reg_alloc.fir": r"FIRRTL version 3.2.0
circuit reg_alloc:
    type Ty0 = {clk: Clock, rst: Reset}
    type Ty1 = {value: UInt<8>}
    type Ty2 = {}
    type Ty3 = {|HdlNone, HdlSome: Ty2|}
    type Ty4 = {normal_regs: Ty1[2], flag_regs: Ty3[2]}
    type Ty5 = {prefix_pad: UInt<0>, dest: Ty4, src: UInt<8>[3], imm_low: UInt<25>, imm_sign: SInt<1>, _phantom: Ty2}
    type Ty6 = {|Full64, DupLow32, ZeroExt32, SignExt32, ZeroExt16, SignExt16, ZeroExt8, SignExt8|}
    type Ty7 = {common: Ty5, output_integer_mode: Ty6}
    type Ty8 = {alu_common: Ty7, invert_src0: UInt<1>, invert_carry_in: UInt<1>, invert_carry_out: UInt<1>, add_pc: UInt<1>}
    type Ty9 = {alu_common: Ty7, lut: UInt<4>}
    type Ty10 = {|AddSub: Ty8, AddSubI: Ty8, Logical: Ty9|}
    type Ty11 = {prefix_pad: UInt<1>, dest: Ty4, src: UInt<8>[3], imm_low: UInt<25>, imm_sign: SInt<1>, _phantom: Ty2}
    type Ty12 = {common: Ty11}
    type Ty13 = {|ReadL2Reg: Ty12, WriteL2Reg: Ty12|}
    type Ty14 = {|Load: Ty11, Store: Ty11|}
    type Ty15 = {|AluBranch: Ty10, L2RegisterFile: Ty13, LoadStore: Ty14|}
    type Ty16 = {mop: Ty15, is_unrelated_pc: UInt<1>, pc: UInt<64>}
    type Ty17 = {|HdlNone, HdlSome: Ty16|}
    type Ty18 = {data: Ty17, flip ready: UInt<1>}
    type Ty19 = {|Trap: Ty2, ICacheFlush|}
    type Ty20 = {|HdlNone, HdlSome: Ty19|}
    type Ty21 = {data: Ty20, flip ready: UInt<1>}
    type Ty22 = {decoded_insns: Ty18[2], flip fetch_decode_special_op: Ty21}
    type Ty23 = {adj_value: UInt<2>}
    type Ty24 = {value: UInt<4>}
    type Ty25 = {unit_num: Ty23, unit_out_reg: Ty24}
    type Ty26 = {addr: UInt<8>, en: UInt<1>, clk: Clock, flip data: Ty25}
    type Ty27 = {adj_value: UInt<1>}
    type Ty28 = {value: UInt<1>}
    type Ty29 = {unit_num: Ty27, unit_out_reg: Ty28}
    type Ty30 = {addr: UInt<8>, en: UInt<1>, clk: Clock, data: Ty25, mask: Ty29}
    type Ty31 = {addr: UInt<8>, en: UInt<1>, clk: Clock, flip data: UInt<2>}
    type Ty32 = {addr: UInt<8>, en: UInt<1>, clk: Clock, data: UInt<2>, mask: UInt<1>}
    type Ty33 = {addr: UInt<8>, en: UInt<1>, clk: Clock, flip data: UInt<4>}
    type Ty34 = {addr: UInt<8>, en: UInt<1>, clk: Clock, data: UInt<4>, mask: UInt<1>}
    type Ty35 = {addr: UInt<1>, en: UInt<1>, clk: Clock, flip data: Ty25}
    type Ty36 = {addr: UInt<1>, en: UInt<1>, clk: Clock, data: Ty25, mask: Ty29}
    type Ty37 = {addr: UInt<1>, en: UInt<1>, clk: Clock, flip data: UInt<2>}
    type Ty38 = {addr: UInt<1>, en: UInt<1>, clk: Clock, data: UInt<2>, mask: UInt<1>}
    type Ty39 = {addr: UInt<1>, en: UInt<1>, clk: Clock, flip data: UInt<4>}
    type Ty40 = {addr: UInt<1>, en: UInt<1>, clk: Clock, data: UInt<4>, mask: UInt<1>}
    type Ty41 = {|HdlNone, HdlSome: UInt<2>|}
    type Ty42 = {prefix_pad: UInt<0>, dest: Ty24, src: UInt<6>[3], imm_low: UInt<25>, imm_sign: SInt<1>, _phantom: Ty2}
    type Ty43 = {common: Ty42, output_integer_mode: Ty6}
    type Ty44 = {alu_common: Ty43, invert_src0: UInt<1>, invert_carry_in: UInt<1>, invert_carry_out: UInt<1>, add_pc: UInt<1>}
    type Ty45 = {alu_common: Ty43, lut: UInt<4>}
    type Ty46 = {|AddSub: Ty44, AddSubI: Ty44, Logical: Ty45|}
    type Ty47 = {prefix_pad: UInt<1>, dest: Ty24, src: UInt<6>[3], imm_low: UInt<25>, imm_sign: SInt<1>, _phantom: Ty2}
    type Ty48 = {common: Ty47}
    type Ty49 = {|ReadL2Reg: Ty48, WriteL2Reg: Ty48|}
    type Ty50 = {|Load: Ty47, Store: Ty47|}
    type Ty51 = {|AluBranch: Ty46, L2RegisterFile: Ty49, LoadStore: Ty50|}
    type Ty52 = {|HdlNone, HdlSome: Ty51|}
    type Ty53 = {|HdlNone, HdlSome: Ty25|}
    type Ty54 = {addr: Ty1, flip data: Ty25}
    type Ty55 = {addr: UInt<0>, en: UInt<1>, clk: Clock, data: Ty25, mask: Ty29}
    type Ty56 = {|AluBranch, L2RegisterFile, LoadStore|}
    type Ty57 = {unit_num: UInt<2>, unit_out_reg: UInt<4>}
    type Ty58 = {imm_low: UInt<25>, reversed_src: UInt<8>[0], imm_sign: SInt<1>}
    type Ty59 = {imm_low: UInt<25>, reversed_src: UInt<0>, imm_sign: UInt<1>}
    type Ty60 = {imm_low: UInt<25>, reversed_src: UInt<8>[1], imm_sign: SInt<1>}
    type Ty61 = {imm_low: UInt<25>, reversed_src: UInt<8>, imm_sign: UInt<1>}
    type Ty62 = {|HdlNone, HdlSome: Ty46|}
    type Ty63 = {data: Ty62, flip ready: UInt<1>}
    type Ty64 = {pwr_ca_x86_cf: UInt<1>, pwr_ca32_x86_af: UInt<1>, pwr_ov_x86_of: UInt<1>, pwr_ov32_x86_df: UInt<1>, pwr_cr_lt_x86_sf: UInt<1>, pwr_cr_gt_x86_pf: UInt<1>, pwr_cr_eq_x86_zf: UInt<1>, pwr_so: UInt<1>}
    type Ty65 = {int_fp: UInt<64>, flags: Ty64}
    type Ty66 = {which: Ty24, value: Ty65}
    type Ty67 = {|HdlNone, HdlSome: Ty66|}
    type Ty68 = {unit_output_writes: Ty67[2], _phantom: Ty2}
    type Ty69 = {which: Ty24}
    type Ty70 = {|HdlNone, HdlSome: Ty69|}
    type Ty71 = {data: Ty70, flip ready: UInt<1>}
    type Ty72 = {value: Ty65, extra_out: Ty2}
    type Ty73 = {|Completed: Ty72, Trap: Ty2|}
    type Ty74 = {which: Ty24, result: Ty73}
    type Ty75 = {|HdlNone, HdlSome: Ty74|}
    type Ty76 = {data: Ty75, flip ready: UInt<1>}
    type Ty77 = {flip cd: Ty0, flip input_insn: Ty63, flip unit_forwarding_info: Ty68, flip cancel_input: Ty71, `output`: Ty76}
    type Ty78 = {|HdlNone, HdlSome: UInt<4>|}
    type Ty79 = {data: Ty78, flip ready: UInt<1>}
    type Ty80 = {flip cd: Ty0, flip free_in: Ty79[1], alloc_out: Ty79[1]}
    type Ty81 = {mop: Ty46, src_values: Ty65[3]}
    type Ty82 = {|HdlNone, HdlSome: Ty81|}
    type Ty83 = {data: Ty82, flip ready: UInt<1>}
    type Ty84 = {flip cd: Ty0, flip unit_forwarding_info: Ty68, flip input_insn: Ty63, flip cancel_input: Ty71, ready_mop: Ty83}
    type Ty85 = {|HdlNone, HdlSome: UInt<3>|}
    type Ty86 = {|HdlNone, HdlSome: Ty65|}
    type Ty87 = {mop: Ty46, src_values: Ty86[3]}
    type Ty88 = {|HdlNone, HdlSome: Ty87|}
    module reg_alloc: @[reg_alloc.rs 49:1]
        input cd: Ty0 @[reg_alloc.rs 53:29]
        input fetch_decode_interface: Ty22 @[reg_alloc.rs 56:11]
        mem rename_table_normal_mem_unit_num_adj_value: @[reg_alloc.rs 74:25]
            data-type => UInt<2>
            depth => 253
            read-latency => 0
            write-latency => 1
            read-under-write => old
            reader => r0
            reader => r1
            reader => r2
            reader => r5
            reader => r6
            reader => r7
            writer => w3
            writer => w4
            writer => w8
            writer => w9
        mem rename_table_normal_mem_unit_out_reg_value: @[reg_alloc.rs 74:25]
            data-type => UInt<4>
            depth => 253
            read-latency => 0
            write-latency => 1
            read-under-write => old
            reader => r0
            reader => r1
            reader => r2
            reader => r5
            reader => r6
            reader => r7
            writer => w3
            writer => w4
            writer => w8
            writer => w9
        mem rename_table_special_mem_unit_num_adj_value: @[reg_alloc.rs 74:25]
            data-type => UInt<2>
            depth => 2
            read-latency => 0
            write-latency => 1
            read-under-write => old
            reader => r0
            reader => r1
            reader => r2
            reader => r7
            reader => r8
            reader => r9
            writer => w3
            writer => w4
            writer => w5
            writer => w6
            writer => w10
            writer => w11
            writer => w12
            writer => w13
        mem rename_table_special_mem_unit_out_reg_value: @[reg_alloc.rs 74:25]
            data-type => UInt<4>
            depth => 2
            read-latency => 0
            write-latency => 1
            read-under-write => old
            reader => r0
            reader => r1
            reader => r2
            reader => r7
            reader => r8
            reader => r9
            writer => w3
            writer => w4
            writer => w5
            writer => w6
            writer => w10
            writer => w11
            writer => w12
            writer => w13
        wire rename_table_normal_mem_r0: Ty26 @[reg_alloc.rs 119:37]
        wire rename_table_normal_mem_r1: Ty26 @[reg_alloc.rs 119:37]
        wire rename_table_normal_mem_r2: Ty26 @[reg_alloc.rs 119:37]
        wire rename_table_normal_mem_w3: Ty30 @[reg_alloc.rs 166:39]
        wire rename_table_normal_mem_w4: Ty30 @[reg_alloc.rs 166:39]
        wire rename_table_normal_mem_r5: Ty26 @[reg_alloc.rs 119:37]
        wire rename_table_normal_mem_r6: Ty26 @[reg_alloc.rs 119:37]
        wire rename_table_normal_mem_r7: Ty26 @[reg_alloc.rs 119:37]
        wire rename_table_normal_mem_w8: Ty30 @[reg_alloc.rs 166:39]
        wire rename_table_normal_mem_w9: Ty30 @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_r0.data.unit_num.adj_value, rename_table_normal_mem_unit_num_adj_value.r0.data @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_r1.data.unit_num.adj_value, rename_table_normal_mem_unit_num_adj_value.r1.data @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_r2.data.unit_num.adj_value, rename_table_normal_mem_unit_num_adj_value.r2.data @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_num_adj_value.w3.data, rename_table_normal_mem_w3.data.unit_num.adj_value @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_num_adj_value.w3.mask, rename_table_normal_mem_w3.mask.unit_num.adj_value @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_num_adj_value.w4.data, rename_table_normal_mem_w4.data.unit_num.adj_value @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_num_adj_value.w4.mask, rename_table_normal_mem_w4.mask.unit_num.adj_value @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_r5.data.unit_num.adj_value, rename_table_normal_mem_unit_num_adj_value.r5.data @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_r6.data.unit_num.adj_value, rename_table_normal_mem_unit_num_adj_value.r6.data @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_r7.data.unit_num.adj_value, rename_table_normal_mem_unit_num_adj_value.r7.data @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_num_adj_value.w8.data, rename_table_normal_mem_w8.data.unit_num.adj_value @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_num_adj_value.w8.mask, rename_table_normal_mem_w8.mask.unit_num.adj_value @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_num_adj_value.w9.data, rename_table_normal_mem_w9.data.unit_num.adj_value @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_num_adj_value.w9.mask, rename_table_normal_mem_w9.mask.unit_num.adj_value @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_num_adj_value.r0.addr, rename_table_normal_mem_r0.addr @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_num_adj_value.r0.clk, rename_table_normal_mem_r0.clk @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_num_adj_value.r0.en, rename_table_normal_mem_r0.en @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_num_adj_value.r1.addr, rename_table_normal_mem_r1.addr @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_num_adj_value.r1.clk, rename_table_normal_mem_r1.clk @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_num_adj_value.r1.en, rename_table_normal_mem_r1.en @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_num_adj_value.r2.addr, rename_table_normal_mem_r2.addr @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_num_adj_value.r2.clk, rename_table_normal_mem_r2.clk @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_num_adj_value.r2.en, rename_table_normal_mem_r2.en @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_num_adj_value.w3.addr, rename_table_normal_mem_w3.addr @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_num_adj_value.w3.clk, rename_table_normal_mem_w3.clk @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_num_adj_value.w3.en, rename_table_normal_mem_w3.en @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_num_adj_value.w4.addr, rename_table_normal_mem_w4.addr @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_num_adj_value.w4.clk, rename_table_normal_mem_w4.clk @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_num_adj_value.w4.en, rename_table_normal_mem_w4.en @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_num_adj_value.r5.addr, rename_table_normal_mem_r5.addr @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_num_adj_value.r5.clk, rename_table_normal_mem_r5.clk @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_num_adj_value.r5.en, rename_table_normal_mem_r5.en @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_num_adj_value.r6.addr, rename_table_normal_mem_r6.addr @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_num_adj_value.r6.clk, rename_table_normal_mem_r6.clk @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_num_adj_value.r6.en, rename_table_normal_mem_r6.en @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_num_adj_value.r7.addr, rename_table_normal_mem_r7.addr @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_num_adj_value.r7.clk, rename_table_normal_mem_r7.clk @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_num_adj_value.r7.en, rename_table_normal_mem_r7.en @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_num_adj_value.w8.addr, rename_table_normal_mem_w8.addr @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_num_adj_value.w8.clk, rename_table_normal_mem_w8.clk @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_num_adj_value.w8.en, rename_table_normal_mem_w8.en @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_num_adj_value.w9.addr, rename_table_normal_mem_w9.addr @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_num_adj_value.w9.clk, rename_table_normal_mem_w9.clk @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_num_adj_value.w9.en, rename_table_normal_mem_w9.en @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_r0.data.unit_out_reg.value, rename_table_normal_mem_unit_out_reg_value.r0.data @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_r1.data.unit_out_reg.value, rename_table_normal_mem_unit_out_reg_value.r1.data @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_r2.data.unit_out_reg.value, rename_table_normal_mem_unit_out_reg_value.r2.data @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_out_reg_value.w3.data, rename_table_normal_mem_w3.data.unit_out_reg.value @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_out_reg_value.w3.mask, rename_table_normal_mem_w3.mask.unit_out_reg.value @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_out_reg_value.w4.data, rename_table_normal_mem_w4.data.unit_out_reg.value @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_out_reg_value.w4.mask, rename_table_normal_mem_w4.mask.unit_out_reg.value @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_r5.data.unit_out_reg.value, rename_table_normal_mem_unit_out_reg_value.r5.data @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_r6.data.unit_out_reg.value, rename_table_normal_mem_unit_out_reg_value.r6.data @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_r7.data.unit_out_reg.value, rename_table_normal_mem_unit_out_reg_value.r7.data @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_out_reg_value.w8.data, rename_table_normal_mem_w8.data.unit_out_reg.value @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_out_reg_value.w8.mask, rename_table_normal_mem_w8.mask.unit_out_reg.value @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_out_reg_value.w9.data, rename_table_normal_mem_w9.data.unit_out_reg.value @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_out_reg_value.w9.mask, rename_table_normal_mem_w9.mask.unit_out_reg.value @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_out_reg_value.r0.addr, rename_table_normal_mem_r0.addr @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_out_reg_value.r0.clk, rename_table_normal_mem_r0.clk @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_out_reg_value.r0.en, rename_table_normal_mem_r0.en @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_out_reg_value.r1.addr, rename_table_normal_mem_r1.addr @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_out_reg_value.r1.clk, rename_table_normal_mem_r1.clk @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_out_reg_value.r1.en, rename_table_normal_mem_r1.en @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_out_reg_value.r2.addr, rename_table_normal_mem_r2.addr @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_out_reg_value.r2.clk, rename_table_normal_mem_r2.clk @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_out_reg_value.r2.en, rename_table_normal_mem_r2.en @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_out_reg_value.w3.addr, rename_table_normal_mem_w3.addr @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_out_reg_value.w3.clk, rename_table_normal_mem_w3.clk @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_out_reg_value.w3.en, rename_table_normal_mem_w3.en @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_out_reg_value.w4.addr, rename_table_normal_mem_w4.addr @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_out_reg_value.w4.clk, rename_table_normal_mem_w4.clk @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_out_reg_value.w4.en, rename_table_normal_mem_w4.en @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_out_reg_value.r5.addr, rename_table_normal_mem_r5.addr @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_out_reg_value.r5.clk, rename_table_normal_mem_r5.clk @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_out_reg_value.r5.en, rename_table_normal_mem_r5.en @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_out_reg_value.r6.addr, rename_table_normal_mem_r6.addr @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_out_reg_value.r6.clk, rename_table_normal_mem_r6.clk @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_out_reg_value.r6.en, rename_table_normal_mem_r6.en @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_out_reg_value.r7.addr, rename_table_normal_mem_r7.addr @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_out_reg_value.r7.clk, rename_table_normal_mem_r7.clk @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_out_reg_value.r7.en, rename_table_normal_mem_r7.en @[reg_alloc.rs 119:37]
        connect rename_table_normal_mem_unit_out_reg_value.w8.addr, rename_table_normal_mem_w8.addr @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_out_reg_value.w8.clk, rename_table_normal_mem_w8.clk @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_out_reg_value.w8.en, rename_table_normal_mem_w8.en @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_out_reg_value.w9.addr, rename_table_normal_mem_w9.addr @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_out_reg_value.w9.clk, rename_table_normal_mem_w9.clk @[reg_alloc.rs 166:39]
        connect rename_table_normal_mem_unit_out_reg_value.w9.en, rename_table_normal_mem_w9.en @[reg_alloc.rs 166:39]
        wire rename_table_special_mem_r0: Ty35 @[reg_alloc.rs 119:37]
        wire rename_table_special_mem_r1: Ty35 @[reg_alloc.rs 119:37]
        wire rename_table_special_mem_r2: Ty35 @[reg_alloc.rs 119:37]
        wire rename_table_special_mem_w3: Ty36 @[reg_alloc.rs 166:39]
        wire rename_table_special_mem_w4: Ty36 @[reg_alloc.rs 166:39]
        wire rename_table_special_mem_w5: Ty36 @[reg_alloc.rs 166:39]
        wire rename_table_special_mem_w6: Ty36 @[reg_alloc.rs 166:39]
        wire rename_table_special_mem_r7: Ty35 @[reg_alloc.rs 119:37]
        wire rename_table_special_mem_r8: Ty35 @[reg_alloc.rs 119:37]
        wire rename_table_special_mem_r9: Ty35 @[reg_alloc.rs 119:37]
        wire rename_table_special_mem_w10: Ty36 @[reg_alloc.rs 166:39]
        wire rename_table_special_mem_w11: Ty36 @[reg_alloc.rs 166:39]
        wire rename_table_special_mem_w12: Ty36 @[reg_alloc.rs 166:39]
        wire rename_table_special_mem_w13: Ty36 @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_r0.data.unit_num.adj_value, rename_table_special_mem_unit_num_adj_value.r0.data @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_r1.data.unit_num.adj_value, rename_table_special_mem_unit_num_adj_value.r1.data @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_r2.data.unit_num.adj_value, rename_table_special_mem_unit_num_adj_value.r2.data @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_num_adj_value.w3.data, rename_table_special_mem_w3.data.unit_num.adj_value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w3.mask, rename_table_special_mem_w3.mask.unit_num.adj_value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w4.data, rename_table_special_mem_w4.data.unit_num.adj_value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w4.mask, rename_table_special_mem_w4.mask.unit_num.adj_value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w5.data, rename_table_special_mem_w5.data.unit_num.adj_value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w5.mask, rename_table_special_mem_w5.mask.unit_num.adj_value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w6.data, rename_table_special_mem_w6.data.unit_num.adj_value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w6.mask, rename_table_special_mem_w6.mask.unit_num.adj_value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_r7.data.unit_num.adj_value, rename_table_special_mem_unit_num_adj_value.r7.data @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_r8.data.unit_num.adj_value, rename_table_special_mem_unit_num_adj_value.r8.data @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_r9.data.unit_num.adj_value, rename_table_special_mem_unit_num_adj_value.r9.data @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_num_adj_value.w10.data, rename_table_special_mem_w10.data.unit_num.adj_value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w10.mask, rename_table_special_mem_w10.mask.unit_num.adj_value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w11.data, rename_table_special_mem_w11.data.unit_num.adj_value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w11.mask, rename_table_special_mem_w11.mask.unit_num.adj_value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w12.data, rename_table_special_mem_w12.data.unit_num.adj_value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w12.mask, rename_table_special_mem_w12.mask.unit_num.adj_value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w13.data, rename_table_special_mem_w13.data.unit_num.adj_value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w13.mask, rename_table_special_mem_w13.mask.unit_num.adj_value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.r0.addr, rename_table_special_mem_r0.addr @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_num_adj_value.r0.clk, rename_table_special_mem_r0.clk @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_num_adj_value.r0.en, rename_table_special_mem_r0.en @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_num_adj_value.r1.addr, rename_table_special_mem_r1.addr @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_num_adj_value.r1.clk, rename_table_special_mem_r1.clk @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_num_adj_value.r1.en, rename_table_special_mem_r1.en @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_num_adj_value.r2.addr, rename_table_special_mem_r2.addr @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_num_adj_value.r2.clk, rename_table_special_mem_r2.clk @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_num_adj_value.r2.en, rename_table_special_mem_r2.en @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_num_adj_value.w3.addr, rename_table_special_mem_w3.addr @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w3.clk, rename_table_special_mem_w3.clk @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w3.en, rename_table_special_mem_w3.en @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w4.addr, rename_table_special_mem_w4.addr @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w4.clk, rename_table_special_mem_w4.clk @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w4.en, rename_table_special_mem_w4.en @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w5.addr, rename_table_special_mem_w5.addr @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w5.clk, rename_table_special_mem_w5.clk @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w5.en, rename_table_special_mem_w5.en @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w6.addr, rename_table_special_mem_w6.addr @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w6.clk, rename_table_special_mem_w6.clk @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w6.en, rename_table_special_mem_w6.en @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.r7.addr, rename_table_special_mem_r7.addr @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_num_adj_value.r7.clk, rename_table_special_mem_r7.clk @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_num_adj_value.r7.en, rename_table_special_mem_r7.en @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_num_adj_value.r8.addr, rename_table_special_mem_r8.addr @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_num_adj_value.r8.clk, rename_table_special_mem_r8.clk @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_num_adj_value.r8.en, rename_table_special_mem_r8.en @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_num_adj_value.r9.addr, rename_table_special_mem_r9.addr @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_num_adj_value.r9.clk, rename_table_special_mem_r9.clk @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_num_adj_value.r9.en, rename_table_special_mem_r9.en @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_num_adj_value.w10.addr, rename_table_special_mem_w10.addr @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w10.clk, rename_table_special_mem_w10.clk @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w10.en, rename_table_special_mem_w10.en @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w11.addr, rename_table_special_mem_w11.addr @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w11.clk, rename_table_special_mem_w11.clk @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w11.en, rename_table_special_mem_w11.en @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w12.addr, rename_table_special_mem_w12.addr @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w12.clk, rename_table_special_mem_w12.clk @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w12.en, rename_table_special_mem_w12.en @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w13.addr, rename_table_special_mem_w13.addr @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w13.clk, rename_table_special_mem_w13.clk @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_num_adj_value.w13.en, rename_table_special_mem_w13.en @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_r0.data.unit_out_reg.value, rename_table_special_mem_unit_out_reg_value.r0.data @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_r1.data.unit_out_reg.value, rename_table_special_mem_unit_out_reg_value.r1.data @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_r2.data.unit_out_reg.value, rename_table_special_mem_unit_out_reg_value.r2.data @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_out_reg_value.w3.data, rename_table_special_mem_w3.data.unit_out_reg.value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w3.mask, rename_table_special_mem_w3.mask.unit_out_reg.value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w4.data, rename_table_special_mem_w4.data.unit_out_reg.value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w4.mask, rename_table_special_mem_w4.mask.unit_out_reg.value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w5.data, rename_table_special_mem_w5.data.unit_out_reg.value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w5.mask, rename_table_special_mem_w5.mask.unit_out_reg.value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w6.data, rename_table_special_mem_w6.data.unit_out_reg.value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w6.mask, rename_table_special_mem_w6.mask.unit_out_reg.value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_r7.data.unit_out_reg.value, rename_table_special_mem_unit_out_reg_value.r7.data @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_r8.data.unit_out_reg.value, rename_table_special_mem_unit_out_reg_value.r8.data @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_r9.data.unit_out_reg.value, rename_table_special_mem_unit_out_reg_value.r9.data @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_out_reg_value.w10.data, rename_table_special_mem_w10.data.unit_out_reg.value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w10.mask, rename_table_special_mem_w10.mask.unit_out_reg.value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w11.data, rename_table_special_mem_w11.data.unit_out_reg.value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w11.mask, rename_table_special_mem_w11.mask.unit_out_reg.value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w12.data, rename_table_special_mem_w12.data.unit_out_reg.value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w12.mask, rename_table_special_mem_w12.mask.unit_out_reg.value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w13.data, rename_table_special_mem_w13.data.unit_out_reg.value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w13.mask, rename_table_special_mem_w13.mask.unit_out_reg.value @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.r0.addr, rename_table_special_mem_r0.addr @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_out_reg_value.r0.clk, rename_table_special_mem_r0.clk @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_out_reg_value.r0.en, rename_table_special_mem_r0.en @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_out_reg_value.r1.addr, rename_table_special_mem_r1.addr @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_out_reg_value.r1.clk, rename_table_special_mem_r1.clk @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_out_reg_value.r1.en, rename_table_special_mem_r1.en @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_out_reg_value.r2.addr, rename_table_special_mem_r2.addr @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_out_reg_value.r2.clk, rename_table_special_mem_r2.clk @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_out_reg_value.r2.en, rename_table_special_mem_r2.en @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_out_reg_value.w3.addr, rename_table_special_mem_w3.addr @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w3.clk, rename_table_special_mem_w3.clk @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w3.en, rename_table_special_mem_w3.en @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w4.addr, rename_table_special_mem_w4.addr @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w4.clk, rename_table_special_mem_w4.clk @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w4.en, rename_table_special_mem_w4.en @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w5.addr, rename_table_special_mem_w5.addr @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w5.clk, rename_table_special_mem_w5.clk @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w5.en, rename_table_special_mem_w5.en @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w6.addr, rename_table_special_mem_w6.addr @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w6.clk, rename_table_special_mem_w6.clk @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w6.en, rename_table_special_mem_w6.en @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.r7.addr, rename_table_special_mem_r7.addr @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_out_reg_value.r7.clk, rename_table_special_mem_r7.clk @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_out_reg_value.r7.en, rename_table_special_mem_r7.en @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_out_reg_value.r8.addr, rename_table_special_mem_r8.addr @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_out_reg_value.r8.clk, rename_table_special_mem_r8.clk @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_out_reg_value.r8.en, rename_table_special_mem_r8.en @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_out_reg_value.r9.addr, rename_table_special_mem_r9.addr @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_out_reg_value.r9.clk, rename_table_special_mem_r9.clk @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_out_reg_value.r9.en, rename_table_special_mem_r9.en @[reg_alloc.rs 119:37]
        connect rename_table_special_mem_unit_out_reg_value.w10.addr, rename_table_special_mem_w10.addr @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w10.clk, rename_table_special_mem_w10.clk @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w10.en, rename_table_special_mem_w10.en @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w11.addr, rename_table_special_mem_w11.addr @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w11.clk, rename_table_special_mem_w11.clk @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w11.en, rename_table_special_mem_w11.en @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w12.addr, rename_table_special_mem_w12.addr @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w12.clk, rename_table_special_mem_w12.clk @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w12.en, rename_table_special_mem_w12.en @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w13.addr, rename_table_special_mem_w13.addr @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w13.clk, rename_table_special_mem_w13.clk @[reg_alloc.rs 166:39]
        connect rename_table_special_mem_unit_out_reg_value.w13.en, rename_table_special_mem_w13.en @[reg_alloc.rs 166:39]
        connect fetch_decode_interface.fetch_decode_special_op.data, {|HdlNone, HdlSome: Ty19|}(HdlNone) @[reg_alloc.rs 58:5]
        wire available_units: UInt<1>[2][2] @[reg_alloc.rs 83:27]
        wire selected_unit_indexes: Ty41[2] @[reg_alloc.rs 86:9]
        wire renamed_mops: Ty52[2] @[reg_alloc.rs 88:24]
        wire renamed_mops_out_reg: Ty53[2] @[reg_alloc.rs 90:32]
        connect fetch_decode_interface.decoded_insns[0].ready, UInt<1>(0h1) @[reg_alloc.rs 92:9]
        wire _array_literal_expr: UInt<1>[2]
        connect _array_literal_expr[0], UInt<1>(0h0)
        connect _array_literal_expr[1], UInt<1>(0h0)
        connect available_units[0], _array_literal_expr @[reg_alloc.rs 96:9]
        connect renamed_mops[0], {|HdlNone, HdlSome: Ty51|}(HdlNone) @[reg_alloc.rs 100:9]
        wire rename_0_src_0: Ty54 @[reg_alloc.rs 113:17]
        wire _bundle_literal_expr: Ty1
        connect _bundle_literal_expr.value, tail(UInt<32>(0h0), 24)
        connect rename_0_src_0.addr, _bundle_literal_expr @[reg_alloc.rs 116:13]
        wire _bundle_literal_expr_1: Ty25
        wire _bundle_literal_expr_2: Ty23
        connect _bundle_literal_expr_2.adj_value, tail(UInt<64>(0h0), 62)
        connect _bundle_literal_expr_1.unit_num, _bundle_literal_expr_2
        wire _bundle_literal_expr_3: Ty24
        connect _bundle_literal_expr_3.value, tail(UInt<8>(0h0), 4)
        connect _bundle_literal_expr_1.unit_out_reg, _bundle_literal_expr_3
        connect rename_0_src_0.data, _bundle_literal_expr_1 @[reg_alloc.rs 117:13]
        connect rename_table_normal_mem_r0.clk, cd.clk @[reg_alloc.rs 120:17]
        connect rename_table_normal_mem_r0.addr, UInt<8>(0h0) @[reg_alloc.rs 121:17]
        connect rename_table_normal_mem_r0.en, UInt<1>(0h0) @[reg_alloc.rs 122:17]
        when and(geq(rename_0_src_0.addr.value, UInt<32>(0h1)), lt(rename_0_src_0.addr.value, UInt<32>(0hFE))): @[reg_alloc.rs 125:17]
            ; connect different types:
            ; lhs: UInt<8>
            ; rhs: UInt<33>
            connect rename_table_normal_mem_r0.addr, sub(rename_0_src_0.addr.value, UInt<32>(0h1)) @[reg_alloc.rs 126:21]
            connect rename_table_normal_mem_r0.en, UInt<1>(0h1) @[reg_alloc.rs 127:21]
            connect rename_0_src_0.data, rename_table_normal_mem_r0.data @[reg_alloc.rs 128:21]
        connect rename_table_special_mem_r0.clk, cd.clk @[reg_alloc.rs 120:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<8>
        connect rename_table_special_mem_r0.addr, UInt<8>(0h0) @[reg_alloc.rs 121:17]
        connect rename_table_special_mem_r0.en, UInt<1>(0h0) @[reg_alloc.rs 122:17]
        when and(geq(rename_0_src_0.addr.value, UInt<32>(0hFE)), lt(rename_0_src_0.addr.value, UInt<32>(0h100))): @[reg_alloc.rs 125:17]
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<33>
            connect rename_table_special_mem_r0.addr, sub(rename_0_src_0.addr.value, UInt<32>(0hFE)) @[reg_alloc.rs 126:21]
            connect rename_table_special_mem_r0.en, UInt<1>(0h1) @[reg_alloc.rs 127:21]
            connect rename_0_src_0.data, rename_table_special_mem_r0.data @[reg_alloc.rs 128:21]
        wire rename_0_src_1: Ty54 @[reg_alloc.rs 113:17]
        wire _bundle_literal_expr_4: Ty1
        connect _bundle_literal_expr_4.value, tail(UInt<32>(0h0), 24)
        connect rename_0_src_1.addr, _bundle_literal_expr_4 @[reg_alloc.rs 116:13]
        wire _bundle_literal_expr_5: Ty25
        wire _bundle_literal_expr_6: Ty23
        connect _bundle_literal_expr_6.adj_value, tail(UInt<64>(0h0), 62)
        connect _bundle_literal_expr_5.unit_num, _bundle_literal_expr_6
        wire _bundle_literal_expr_7: Ty24
        connect _bundle_literal_expr_7.value, tail(UInt<8>(0h0), 4)
        connect _bundle_literal_expr_5.unit_out_reg, _bundle_literal_expr_7
        connect rename_0_src_1.data, _bundle_literal_expr_5 @[reg_alloc.rs 117:13]
        connect rename_table_normal_mem_r1.clk, cd.clk @[reg_alloc.rs 120:17]
        connect rename_table_normal_mem_r1.addr, UInt<8>(0h0) @[reg_alloc.rs 121:17]
        connect rename_table_normal_mem_r1.en, UInt<1>(0h0) @[reg_alloc.rs 122:17]
        when and(geq(rename_0_src_1.addr.value, UInt<32>(0h1)), lt(rename_0_src_1.addr.value, UInt<32>(0hFE))): @[reg_alloc.rs 125:17]
            ; connect different types:
            ; lhs: UInt<8>
            ; rhs: UInt<33>
            connect rename_table_normal_mem_r1.addr, sub(rename_0_src_1.addr.value, UInt<32>(0h1)) @[reg_alloc.rs 126:21]
            connect rename_table_normal_mem_r1.en, UInt<1>(0h1) @[reg_alloc.rs 127:21]
            connect rename_0_src_1.data, rename_table_normal_mem_r1.data @[reg_alloc.rs 128:21]
        connect rename_table_special_mem_r1.clk, cd.clk @[reg_alloc.rs 120:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<8>
        connect rename_table_special_mem_r1.addr, UInt<8>(0h0) @[reg_alloc.rs 121:17]
        connect rename_table_special_mem_r1.en, UInt<1>(0h0) @[reg_alloc.rs 122:17]
        when and(geq(rename_0_src_1.addr.value, UInt<32>(0hFE)), lt(rename_0_src_1.addr.value, UInt<32>(0h100))): @[reg_alloc.rs 125:17]
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<33>
            connect rename_table_special_mem_r1.addr, sub(rename_0_src_1.addr.value, UInt<32>(0hFE)) @[reg_alloc.rs 126:21]
            connect rename_table_special_mem_r1.en, UInt<1>(0h1) @[reg_alloc.rs 127:21]
            connect rename_0_src_1.data, rename_table_special_mem_r1.data @[reg_alloc.rs 128:21]
        wire rename_0_src_2: Ty54 @[reg_alloc.rs 113:17]
        wire _bundle_literal_expr_8: Ty1
        connect _bundle_literal_expr_8.value, tail(UInt<32>(0h0), 24)
        connect rename_0_src_2.addr, _bundle_literal_expr_8 @[reg_alloc.rs 116:13]
        wire _bundle_literal_expr_9: Ty25
        wire _bundle_literal_expr_10: Ty23
        connect _bundle_literal_expr_10.adj_value, tail(UInt<64>(0h0), 62)
        connect _bundle_literal_expr_9.unit_num, _bundle_literal_expr_10
        wire _bundle_literal_expr_11: Ty24
        connect _bundle_literal_expr_11.value, tail(UInt<8>(0h0), 4)
        connect _bundle_literal_expr_9.unit_out_reg, _bundle_literal_expr_11
        connect rename_0_src_2.data, _bundle_literal_expr_9 @[reg_alloc.rs 117:13]
        connect rename_table_normal_mem_r2.clk, cd.clk @[reg_alloc.rs 120:17]
        connect rename_table_normal_mem_r2.addr, UInt<8>(0h0) @[reg_alloc.rs 121:17]
        connect rename_table_normal_mem_r2.en, UInt<1>(0h0) @[reg_alloc.rs 122:17]
        when and(geq(rename_0_src_2.addr.value, UInt<32>(0h1)), lt(rename_0_src_2.addr.value, UInt<32>(0hFE))): @[reg_alloc.rs 125:17]
            ; connect different types:
            ; lhs: UInt<8>
            ; rhs: UInt<33>
            connect rename_table_normal_mem_r2.addr, sub(rename_0_src_2.addr.value, UInt<32>(0h1)) @[reg_alloc.rs 126:21]
            connect rename_table_normal_mem_r2.en, UInt<1>(0h1) @[reg_alloc.rs 127:21]
            connect rename_0_src_2.data, rename_table_normal_mem_r2.data @[reg_alloc.rs 128:21]
        connect rename_table_special_mem_r2.clk, cd.clk @[reg_alloc.rs 120:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<8>
        connect rename_table_special_mem_r2.addr, UInt<8>(0h0) @[reg_alloc.rs 121:17]
        connect rename_table_special_mem_r2.en, UInt<1>(0h0) @[reg_alloc.rs 122:17]
        when and(geq(rename_0_src_2.addr.value, UInt<32>(0hFE)), lt(rename_0_src_2.addr.value, UInt<32>(0h100))): @[reg_alloc.rs 125:17]
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<33>
            connect rename_table_special_mem_r2.addr, sub(rename_0_src_2.addr.value, UInt<32>(0hFE)) @[reg_alloc.rs 126:21]
            connect rename_table_special_mem_r2.en, UInt<1>(0h1) @[reg_alloc.rs 127:21]
            connect rename_0_src_2.data, rename_table_special_mem_r2.data @[reg_alloc.rs 128:21]
        wire rename_table_normal_0_dest0: Ty30 @[reg_alloc.rs 170:21]
        connect rename_table_normal_mem_w3, rename_table_normal_0_dest0 @[reg_alloc.rs 173:17]
        wire _bundle_literal_expr_12: Ty55
        connect _bundle_literal_expr_12.addr, UInt<0>(0h0)
        connect _bundle_literal_expr_12.en, UInt<1>(0h0)
        connect _bundle_literal_expr_12.clk, cd.clk
        wire _uninit_expr: Ty25
        invalidate _uninit_expr
        connect _bundle_literal_expr_12.data, _uninit_expr
        wire _bundle_literal_expr_13: Ty29
        wire _bundle_literal_expr_14: Ty27
        connect _bundle_literal_expr_14.adj_value, UInt<1>(0h1)
        connect _bundle_literal_expr_13.unit_num, _bundle_literal_expr_14
        wire _bundle_literal_expr_15: Ty28
        connect _bundle_literal_expr_15.value, UInt<1>(0h1)
        connect _bundle_literal_expr_13.unit_out_reg, _bundle_literal_expr_15
        connect _bundle_literal_expr_12.mask, _bundle_literal_expr_13
        ; connect different types:
        ; lhs: Bundle {addr: UInt<8>, en: Bool, clk: Clock, data: Bundle {unit_num: Bundle {adj_value: UInt<2>}, unit_out_reg: Bundle {value: UInt<4>}}, mask: Bundle {unit_num: Bundle {adj_value: Bool}, unit_out_reg: Bundle {value: Bool}}}
        ; rhs: Bundle {addr: UInt<0>, en: Bool, clk: Clock, data: Bundle {unit_num: Bundle {adj_value: UInt<2>}, unit_out_reg: Bundle {value: UInt<4>}}, mask: Bundle {unit_num: Bundle {adj_value: Bool}, unit_out_reg: Bundle {value: Bool}}}
        connect rename_table_normal_0_dest0, _bundle_literal_expr_12 @[reg_alloc.rs 175:17]
        wire rename_table_special_0_dest0: Ty36 @[reg_alloc.rs 170:21]
        connect rename_table_special_mem_w3, rename_table_special_0_dest0 @[reg_alloc.rs 173:17]
        wire _bundle_literal_expr_16: Ty55
        connect _bundle_literal_expr_16.addr, UInt<0>(0h0)
        connect _bundle_literal_expr_16.en, UInt<1>(0h0)
        connect _bundle_literal_expr_16.clk, cd.clk
        wire _uninit_expr_1: Ty25
        invalidate _uninit_expr_1
        connect _bundle_literal_expr_16.data, _uninit_expr_1
        wire _bundle_literal_expr_17: Ty29
        wire _bundle_literal_expr_18: Ty27
        connect _bundle_literal_expr_18.adj_value, UInt<1>(0h1)
        connect _bundle_literal_expr_17.unit_num, _bundle_literal_expr_18
        wire _bundle_literal_expr_19: Ty28
        connect _bundle_literal_expr_19.value, UInt<1>(0h1)
        connect _bundle_literal_expr_17.unit_out_reg, _bundle_literal_expr_19
        connect _bundle_literal_expr_16.mask, _bundle_literal_expr_17
        ; connect different types:
        ; lhs: Bundle {addr: UInt<1>, en: Bool, clk: Clock, data: Bundle {unit_num: Bundle {adj_value: UInt<2>}, unit_out_reg: Bundle {value: UInt<4>}}, mask: Bundle {unit_num: Bundle {adj_value: Bool}, unit_out_reg: Bundle {value: Bool}}}
        ; rhs: Bundle {addr: UInt<0>, en: Bool, clk: Clock, data: Bundle {unit_num: Bundle {adj_value: UInt<2>}, unit_out_reg: Bundle {value: UInt<4>}}, mask: Bundle {unit_num: Bundle {adj_value: Bool}, unit_out_reg: Bundle {value: Bool}}}
        connect rename_table_special_0_dest0, _bundle_literal_expr_16 @[reg_alloc.rs 175:17]
        wire rename_table_normal_0_dest1: Ty30 @[reg_alloc.rs 170:21]
        connect rename_table_normal_mem_w4, rename_table_normal_0_dest1 @[reg_alloc.rs 173:17]
        wire _bundle_literal_expr_20: Ty55
        connect _bundle_literal_expr_20.addr, UInt<0>(0h0)
        connect _bundle_literal_expr_20.en, UInt<1>(0h0)
        connect _bundle_literal_expr_20.clk, cd.clk
        wire _uninit_expr_2: Ty25
        invalidate _uninit_expr_2
        connect _bundle_literal_expr_20.data, _uninit_expr_2
        wire _bundle_literal_expr_21: Ty29
        wire _bundle_literal_expr_22: Ty27
        connect _bundle_literal_expr_22.adj_value, UInt<1>(0h1)
        connect _bundle_literal_expr_21.unit_num, _bundle_literal_expr_22
        wire _bundle_literal_expr_23: Ty28
        connect _bundle_literal_expr_23.value, UInt<1>(0h1)
        connect _bundle_literal_expr_21.unit_out_reg, _bundle_literal_expr_23
        connect _bundle_literal_expr_20.mask, _bundle_literal_expr_21
        ; connect different types:
        ; lhs: Bundle {addr: UInt<8>, en: Bool, clk: Clock, data: Bundle {unit_num: Bundle {adj_value: UInt<2>}, unit_out_reg: Bundle {value: UInt<4>}}, mask: Bundle {unit_num: Bundle {adj_value: Bool}, unit_out_reg: Bundle {value: Bool}}}
        ; rhs: Bundle {addr: UInt<0>, en: Bool, clk: Clock, data: Bundle {unit_num: Bundle {adj_value: UInt<2>}, unit_out_reg: Bundle {value: UInt<4>}}, mask: Bundle {unit_num: Bundle {adj_value: Bool}, unit_out_reg: Bundle {value: Bool}}}
        connect rename_table_normal_0_dest1, _bundle_literal_expr_20 @[reg_alloc.rs 175:17]
        wire rename_table_special_0_dest1: Ty36 @[reg_alloc.rs 170:21]
        connect rename_table_special_mem_w4, rename_table_special_0_dest1 @[reg_alloc.rs 173:17]
        wire _bundle_literal_expr_24: Ty55
        connect _bundle_literal_expr_24.addr, UInt<0>(0h0)
        connect _bundle_literal_expr_24.en, UInt<1>(0h0)
        connect _bundle_literal_expr_24.clk, cd.clk
        wire _uninit_expr_3: Ty25
        invalidate _uninit_expr_3
        connect _bundle_literal_expr_24.data, _uninit_expr_3
        wire _bundle_literal_expr_25: Ty29
        wire _bundle_literal_expr_26: Ty27
        connect _bundle_literal_expr_26.adj_value, UInt<1>(0h1)
        connect _bundle_literal_expr_25.unit_num, _bundle_literal_expr_26
        wire _bundle_literal_expr_27: Ty28
        connect _bundle_literal_expr_27.value, UInt<1>(0h1)
        connect _bundle_literal_expr_25.unit_out_reg, _bundle_literal_expr_27
        connect _bundle_literal_expr_24.mask, _bundle_literal_expr_25
        ; connect different types:
        ; lhs: Bundle {addr: UInt<1>, en: Bool, clk: Clock, data: Bundle {unit_num: Bundle {adj_value: UInt<2>}, unit_out_reg: Bundle {value: UInt<4>}}, mask: Bundle {unit_num: Bundle {adj_value: Bool}, unit_out_reg: Bundle {value: Bool}}}
        ; rhs: Bundle {addr: UInt<0>, en: Bool, clk: Clock, data: Bundle {unit_num: Bundle {adj_value: UInt<2>}, unit_out_reg: Bundle {value: UInt<4>}}, mask: Bundle {unit_num: Bundle {adj_value: Bool}, unit_out_reg: Bundle {value: Bool}}}
        connect rename_table_special_0_dest1, _bundle_literal_expr_24 @[reg_alloc.rs 175:17]
        wire rename_table_special_0_flag0_rFE: Ty36 @[reg_alloc.rs 170:21]
        connect rename_table_special_mem_w5, rename_table_special_0_flag0_rFE @[reg_alloc.rs 173:17]
        wire _bundle_literal_expr_28: Ty55
        connect _bundle_literal_expr_28.addr, UInt<0>(0h0)
        connect _bundle_literal_expr_28.en, UInt<1>(0h0)
        connect _bundle_literal_expr_28.clk, cd.clk
        wire _uninit_expr_4: Ty25
        invalidate _uninit_expr_4
        connect _bundle_literal_expr_28.data, _uninit_expr_4
        wire _bundle_literal_expr_29: Ty29
        wire _bundle_literal_expr_30: Ty27
        connect _bundle_literal_expr_30.adj_value, UInt<1>(0h1)
        connect _bundle_literal_expr_29.unit_num, _bundle_literal_expr_30
        wire _bundle_literal_expr_31: Ty28
        connect _bundle_literal_expr_31.value, UInt<1>(0h1)
        connect _bundle_literal_expr_29.unit_out_reg, _bundle_literal_expr_31
        connect _bundle_literal_expr_28.mask, _bundle_literal_expr_29
        ; connect different types:
        ; lhs: Bundle {addr: UInt<1>, en: Bool, clk: Clock, data: Bundle {unit_num: Bundle {adj_value: UInt<2>}, unit_out_reg: Bundle {value: UInt<4>}}, mask: Bundle {unit_num: Bundle {adj_value: Bool}, unit_out_reg: Bundle {value: Bool}}}
        ; rhs: Bundle {addr: UInt<0>, en: Bool, clk: Clock, data: Bundle {unit_num: Bundle {adj_value: UInt<2>}, unit_out_reg: Bundle {value: UInt<4>}}, mask: Bundle {unit_num: Bundle {adj_value: Bool}, unit_out_reg: Bundle {value: Bool}}}
        connect rename_table_special_0_flag0_rFE, _bundle_literal_expr_28 @[reg_alloc.rs 175:17]
        wire rename_table_special_0_flag1_rFF: Ty36 @[reg_alloc.rs 170:21]
        connect rename_table_special_mem_w6, rename_table_special_0_flag1_rFF @[reg_alloc.rs 173:17]
        wire _bundle_literal_expr_32: Ty55
        connect _bundle_literal_expr_32.addr, UInt<0>(0h0)
        connect _bundle_literal_expr_32.en, UInt<1>(0h0)
        connect _bundle_literal_expr_32.clk, cd.clk
        wire _uninit_expr_5: Ty25
        invalidate _uninit_expr_5
        connect _bundle_literal_expr_32.data, _uninit_expr_5
        wire _bundle_literal_expr_33: Ty29
        wire _bundle_literal_expr_34: Ty27
        connect _bundle_literal_expr_34.adj_value, UInt<1>(0h1)
        connect _bundle_literal_expr_33.unit_num, _bundle_literal_expr_34
        wire _bundle_literal_expr_35: Ty28
        connect _bundle_literal_expr_35.value, UInt<1>(0h1)
        connect _bundle_literal_expr_33.unit_out_reg, _bundle_literal_expr_35
        connect _bundle_literal_expr_32.mask, _bundle_literal_expr_33
        ; connect different types:
        ; lhs: Bundle {addr: UInt<1>, en: Bool, clk: Clock, data: Bundle {unit_num: Bundle {adj_value: UInt<2>}, unit_out_reg: Bundle {value: UInt<4>}}, mask: Bundle {unit_num: Bundle {adj_value: Bool}, unit_out_reg: Bundle {value: Bool}}}
        ; rhs: Bundle {addr: UInt<0>, en: Bool, clk: Clock, data: Bundle {unit_num: Bundle {adj_value: UInt<2>}, unit_out_reg: Bundle {value: UInt<4>}}, mask: Bundle {unit_num: Bundle {adj_value: Bool}, unit_out_reg: Bundle {value: Bool}}}
        connect rename_table_special_0_flag1_rFF, _bundle_literal_expr_32 @[reg_alloc.rs 175:17]
        match fetch_decode_interface.decoded_insns[0].data: @[reg_alloc.rs 189:9]
            HdlNone:
                skip
            HdlSome(_match_arm_value):
                wire unit_kind: Ty56 @[unit.rs 129:1]
                match _match_arm_value.mop: @[unit.rs 129:1]
                    AluBranch(_match_arm_value_1):
                        connect unit_kind, {|AluBranch, L2RegisterFile, LoadStore|}(AluBranch) @[unit.rs 129:1]
                    L2RegisterFile(_match_arm_value_2):
                        connect unit_kind, {|AluBranch, L2RegisterFile, LoadStore|}(L2RegisterFile) @[unit.rs 129:1]
                    LoadStore(_match_arm_value_3):
                        connect unit_kind, {|AluBranch, L2RegisterFile, LoadStore|}(LoadStore) @[unit.rs 129:1]
                wire available_units_for_kind: UInt<1>[2] @[unit.rs 129:1]
                match unit_kind: @[unit.rs 129:1]
                    AluBranch:
                        connect available_units_for_kind[0], UInt<1>(0h1) @[unit.rs 129:1]
                        connect available_units_for_kind[1], UInt<1>(0h1) @[unit.rs 129:1]
                    L2RegisterFile:
                        connect available_units_for_kind[0], UInt<1>(0h0) @[unit.rs 129:1]
                        connect available_units_for_kind[1], UInt<1>(0h0) @[unit.rs 129:1]
                    LoadStore:
                        connect available_units_for_kind[0], UInt<1>(0h0) @[unit.rs 129:1]
                        connect available_units_for_kind[1], UInt<1>(0h0) @[unit.rs 129:1]
                connect available_units[0], available_units_for_kind @[reg_alloc.rs 190:13]
                match renamed_mops_out_reg[0]: @[reg_alloc.rs 195:13]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_4):
                        wire dest_reg: Ty4 @[unit.rs 129:1]
                        match _match_arm_value.mop: @[unit.rs 129:1]
                            AluBranch(_match_arm_value_5):
                                wire dest_reg_1: Ty4 @[instruction.rs 502:1]
                                match _match_arm_value_5: @[instruction.rs 502:1]
                                    AddSub(_match_arm_value_6):
                                        connect dest_reg_1, _match_arm_value_6.alu_common.common.dest @[instruction.rs 502:1]
                                    AddSubI(_match_arm_value_7):
                                        connect dest_reg_1, _match_arm_value_7.alu_common.common.dest @[instruction.rs 502:1]
                                    Logical(_match_arm_value_8):
                                        connect dest_reg_1, _match_arm_value_8.alu_common.common.dest @[instruction.rs 502:1]
                                connect dest_reg, dest_reg_1 @[unit.rs 129:1]
                            L2RegisterFile(_match_arm_value_9):
                                wire dest_reg_2: Ty4 @[instruction.rs 529:1]
                                match _match_arm_value_9: @[instruction.rs 529:1]
                                    ReadL2Reg(_match_arm_value_10):
                                        connect dest_reg_2, _match_arm_value_10.common.dest @[instruction.rs 529:1]
                                    WriteL2Reg(_match_arm_value_11):
                                        connect dest_reg_2, _match_arm_value_11.common.dest @[instruction.rs 529:1]
                                connect dest_reg, dest_reg_2 @[unit.rs 129:1]
                            LoadStore(_match_arm_value_12):
                                wire dest_reg_3: Ty4 @[instruction.rs 564:1]
                                match _match_arm_value_12: @[instruction.rs 564:1]
                                    Load(_match_arm_value_13):
                                        connect dest_reg_3, _match_arm_value_13.dest @[instruction.rs 564:1]
                                    Store(_match_arm_value_14):
                                        connect dest_reg_3, _match_arm_value_14.dest @[instruction.rs 564:1]
                                connect dest_reg, dest_reg_3 @[unit.rs 129:1]
                        wire mapped_regs: Ty51 @[unit.rs 129:1]
                        match _match_arm_value.mop: @[unit.rs 129:1]
                            AluBranch(_match_arm_value_15):
                                wire mapped_regs_1: Ty46 @[instruction.rs 502:1]
                                match _match_arm_value_15: @[instruction.rs 502:1]
                                    AddSub(_match_arm_value_16):
                                        wire _bundle_literal_expr_36: Ty1
                                        connect _bundle_literal_expr_36.value, _match_arm_value_16.alu_common.common.src[0]
                                        connect rename_0_src_0.addr, _bundle_literal_expr_36 @[reg_alloc.rs 204:29]
                                        wire _bundle_literal_expr_37: Ty1
                                        connect _bundle_literal_expr_37.value, _match_arm_value_16.alu_common.common.src[1]
                                        connect rename_0_src_1.addr, _bundle_literal_expr_37 @[reg_alloc.rs 204:29]
                                        wire _bundle_literal_expr_38: Ty1
                                        connect _bundle_literal_expr_38.value, _match_arm_value_16.alu_common.common.src[2]
                                        connect rename_0_src_2.addr, _bundle_literal_expr_38 @[reg_alloc.rs 204:29]
                                        wire _bundle_literal_expr_39: Ty44
                                        wire _bundle_literal_expr_40: Ty43
                                        wire _bundle_literal_expr_41: Ty42
                                        connect _bundle_literal_expr_41.prefix_pad, _match_arm_value_16.alu_common.common.prefix_pad
                                        connect _bundle_literal_expr_41.dest, _match_arm_value_4.unit_out_reg
                                        wire _array_literal_expr_1: UInt<6>[3]
                                        wire _array_literal_expr_2: UInt<6>[3]
                                        wire _cast_bundle_to_bits_expr: Ty57
                                        connect _cast_bundle_to_bits_expr.unit_num, rename_0_src_0.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr.unit_out_reg, rename_0_src_0.data.unit_out_reg.value
                                        wire _cast_to_bits_expr: UInt<6>
                                        connect _cast_to_bits_expr, cat(_cast_bundle_to_bits_expr.unit_out_reg, _cast_bundle_to_bits_expr.unit_num)
                                        connect _array_literal_expr_2[0], _cast_to_bits_expr
                                        wire _cast_bundle_to_bits_expr_1: Ty57
                                        connect _cast_bundle_to_bits_expr_1.unit_num, rename_0_src_1.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_1.unit_out_reg, rename_0_src_1.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_1: UInt<6>
                                        connect _cast_to_bits_expr_1, cat(_cast_bundle_to_bits_expr_1.unit_out_reg, _cast_bundle_to_bits_expr_1.unit_num)
                                        connect _array_literal_expr_2[1], _cast_to_bits_expr_1
                                        wire _cast_bundle_to_bits_expr_2: Ty57
                                        connect _cast_bundle_to_bits_expr_2.unit_num, rename_0_src_2.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_2.unit_out_reg, rename_0_src_2.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_2: UInt<6>
                                        connect _cast_to_bits_expr_2, cat(_cast_bundle_to_bits_expr_2.unit_out_reg, _cast_bundle_to_bits_expr_2.unit_num)
                                        connect _array_literal_expr_2[2], _cast_to_bits_expr_2
                                        connect _array_literal_expr_1[0], _array_literal_expr_2[0]
                                        wire _array_literal_expr_3: UInt<6>[3]
                                        wire _cast_bundle_to_bits_expr_3: Ty57
                                        connect _cast_bundle_to_bits_expr_3.unit_num, rename_0_src_0.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_3.unit_out_reg, rename_0_src_0.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_3: UInt<6>
                                        connect _cast_to_bits_expr_3, cat(_cast_bundle_to_bits_expr_3.unit_out_reg, _cast_bundle_to_bits_expr_3.unit_num)
                                        connect _array_literal_expr_3[0], _cast_to_bits_expr_3
                                        wire _cast_bundle_to_bits_expr_4: Ty57
                                        connect _cast_bundle_to_bits_expr_4.unit_num, rename_0_src_1.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_4.unit_out_reg, rename_0_src_1.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_4: UInt<6>
                                        connect _cast_to_bits_expr_4, cat(_cast_bundle_to_bits_expr_4.unit_out_reg, _cast_bundle_to_bits_expr_4.unit_num)
                                        connect _array_literal_expr_3[1], _cast_to_bits_expr_4
                                        wire _cast_bundle_to_bits_expr_5: Ty57
                                        connect _cast_bundle_to_bits_expr_5.unit_num, rename_0_src_2.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_5.unit_out_reg, rename_0_src_2.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_5: UInt<6>
                                        connect _cast_to_bits_expr_5, cat(_cast_bundle_to_bits_expr_5.unit_out_reg, _cast_bundle_to_bits_expr_5.unit_num)
                                        connect _array_literal_expr_3[2], _cast_to_bits_expr_5
                                        connect _array_literal_expr_1[1], _array_literal_expr_3[1]
                                        wire _array_literal_expr_4: UInt<6>[3]
                                        wire _cast_bundle_to_bits_expr_6: Ty57
                                        connect _cast_bundle_to_bits_expr_6.unit_num, rename_0_src_0.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_6.unit_out_reg, rename_0_src_0.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_6: UInt<6>
                                        connect _cast_to_bits_expr_6, cat(_cast_bundle_to_bits_expr_6.unit_out_reg, _cast_bundle_to_bits_expr_6.unit_num)
                                        connect _array_literal_expr_4[0], _cast_to_bits_expr_6
                                        wire _cast_bundle_to_bits_expr_7: Ty57
                                        connect _cast_bundle_to_bits_expr_7.unit_num, rename_0_src_1.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_7.unit_out_reg, rename_0_src_1.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_7: UInt<6>
                                        connect _cast_to_bits_expr_7, cat(_cast_bundle_to_bits_expr_7.unit_out_reg, _cast_bundle_to_bits_expr_7.unit_num)
                                        connect _array_literal_expr_4[1], _cast_to_bits_expr_7
                                        wire _cast_bundle_to_bits_expr_8: Ty57
                                        connect _cast_bundle_to_bits_expr_8.unit_num, rename_0_src_2.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_8.unit_out_reg, rename_0_src_2.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_8: UInt<6>
                                        connect _cast_to_bits_expr_8, cat(_cast_bundle_to_bits_expr_8.unit_out_reg, _cast_bundle_to_bits_expr_8.unit_num)
                                        connect _array_literal_expr_4[2], _cast_to_bits_expr_8
                                        connect _array_literal_expr_1[2], _array_literal_expr_4[2]
                                        connect _bundle_literal_expr_41.src, _array_literal_expr_1
                                        wire _bundle_literal_expr_42: Ty58
                                        connect _bundle_literal_expr_42.imm_low, _match_arm_value_16.alu_common.common.imm_low
                                        wire _array_literal_expr_5: UInt<8>[0]
                                        invalidate _array_literal_expr_5
                                        connect _bundle_literal_expr_42.reversed_src, _array_literal_expr_5
                                        connect _bundle_literal_expr_42.imm_sign, _match_arm_value_16.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_9: Ty59
                                        connect _cast_bundle_to_bits_expr_9.imm_low, _bundle_literal_expr_42.imm_low
                                        connect _cast_bundle_to_bits_expr_9.reversed_src, UInt<0>(0)
                                        connect _cast_bundle_to_bits_expr_9.imm_sign, asUInt(_bundle_literal_expr_42.imm_sign)
                                        wire _cast_to_bits_expr_9: UInt<26>
                                        connect _cast_to_bits_expr_9, cat(_cast_bundle_to_bits_expr_9.imm_sign, cat(_cast_bundle_to_bits_expr_9.reversed_src, _cast_bundle_to_bits_expr_9.imm_low))
                                        connect _bundle_literal_expr_41.imm_low, bits(asSInt(_cast_to_bits_expr_9), 24, 0)
                                        wire _bundle_literal_expr_43: Ty58
                                        connect _bundle_literal_expr_43.imm_low, _match_arm_value_16.alu_common.common.imm_low
                                        wire _array_literal_expr_6: UInt<8>[0]
                                        invalidate _array_literal_expr_6
                                        connect _bundle_literal_expr_43.reversed_src, _array_literal_expr_6
                                        connect _bundle_literal_expr_43.imm_sign, _match_arm_value_16.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_10: Ty59
                                        connect _cast_bundle_to_bits_expr_10.imm_low, _bundle_literal_expr_43.imm_low
                                        connect _cast_bundle_to_bits_expr_10.reversed_src, UInt<0>(0)
                                        connect _cast_bundle_to_bits_expr_10.imm_sign, asUInt(_bundle_literal_expr_43.imm_sign)
                                        wire _cast_to_bits_expr_10: UInt<26>
                                        connect _cast_to_bits_expr_10, cat(_cast_bundle_to_bits_expr_10.imm_sign, cat(_cast_bundle_to_bits_expr_10.reversed_src, _cast_bundle_to_bits_expr_10.imm_low))
                                        connect _bundle_literal_expr_41.imm_sign, shr(asSInt(_cast_to_bits_expr_10), 25)
                                        wire _bundle_literal_expr_44: Ty2
                                        invalidate _bundle_literal_expr_44
                                        connect _bundle_literal_expr_41._phantom, _bundle_literal_expr_44
                                        connect _bundle_literal_expr_40.common, _bundle_literal_expr_41
                                        connect _bundle_literal_expr_40.output_integer_mode, _match_arm_value_16.alu_common.output_integer_mode
                                        connect _bundle_literal_expr_39.alu_common, _bundle_literal_expr_40
                                        connect _bundle_literal_expr_39.invert_src0, _match_arm_value_16.invert_src0
                                        connect _bundle_literal_expr_39.invert_carry_in, _match_arm_value_16.invert_carry_in
                                        connect _bundle_literal_expr_39.invert_carry_out, _match_arm_value_16.invert_carry_out
                                        connect _bundle_literal_expr_39.add_pc, _match_arm_value_16.add_pc
                                        connect mapped_regs_1, {|AddSub: Ty44, AddSubI: Ty44, Logical: Ty45|}(AddSub, _bundle_literal_expr_39) @[instruction.rs 502:1]
                                    AddSubI(_match_arm_value_17):
                                        wire _bundle_literal_expr_45: Ty1
                                        connect _bundle_literal_expr_45.value, _match_arm_value_17.alu_common.common.src[0]
                                        connect rename_0_src_0.addr, _bundle_literal_expr_45 @[reg_alloc.rs 204:29]
                                        wire _bundle_literal_expr_46: Ty1
                                        connect _bundle_literal_expr_46.value, _match_arm_value_17.alu_common.common.src[1]
                                        connect rename_0_src_1.addr, _bundle_literal_expr_46 @[reg_alloc.rs 204:29]
                                        wire _bundle_literal_expr_47: Ty44
                                        wire _bundle_literal_expr_48: Ty43
                                        wire _bundle_literal_expr_49: Ty42
                                        connect _bundle_literal_expr_49.prefix_pad, _match_arm_value_17.alu_common.common.prefix_pad
                                        connect _bundle_literal_expr_49.dest, _match_arm_value_4.unit_out_reg
                                        wire _array_literal_expr_7: UInt<6>[3]
                                        wire _array_literal_expr_8: UInt<6>[2]
                                        wire _cast_bundle_to_bits_expr_11: Ty57
                                        connect _cast_bundle_to_bits_expr_11.unit_num, rename_0_src_0.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_11.unit_out_reg, rename_0_src_0.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_11: UInt<6>
                                        connect _cast_to_bits_expr_11, cat(_cast_bundle_to_bits_expr_11.unit_out_reg, _cast_bundle_to_bits_expr_11.unit_num)
                                        connect _array_literal_expr_8[0], _cast_to_bits_expr_11
                                        wire _cast_bundle_to_bits_expr_12: Ty57
                                        connect _cast_bundle_to_bits_expr_12.unit_num, rename_0_src_1.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_12.unit_out_reg, rename_0_src_1.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_12: UInt<6>
                                        connect _cast_to_bits_expr_12, cat(_cast_bundle_to_bits_expr_12.unit_out_reg, _cast_bundle_to_bits_expr_12.unit_num)
                                        connect _array_literal_expr_8[1], _cast_to_bits_expr_12
                                        connect _array_literal_expr_7[0], _array_literal_expr_8[0]
                                        wire _array_literal_expr_9: UInt<6>[2]
                                        wire _cast_bundle_to_bits_expr_13: Ty57
                                        connect _cast_bundle_to_bits_expr_13.unit_num, rename_0_src_0.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_13.unit_out_reg, rename_0_src_0.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_13: UInt<6>
                                        connect _cast_to_bits_expr_13, cat(_cast_bundle_to_bits_expr_13.unit_out_reg, _cast_bundle_to_bits_expr_13.unit_num)
                                        connect _array_literal_expr_9[0], _cast_to_bits_expr_13
                                        wire _cast_bundle_to_bits_expr_14: Ty57
                                        connect _cast_bundle_to_bits_expr_14.unit_num, rename_0_src_1.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_14.unit_out_reg, rename_0_src_1.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_14: UInt<6>
                                        connect _cast_to_bits_expr_14, cat(_cast_bundle_to_bits_expr_14.unit_out_reg, _cast_bundle_to_bits_expr_14.unit_num)
                                        connect _array_literal_expr_9[1], _cast_to_bits_expr_14
                                        connect _array_literal_expr_7[1], _array_literal_expr_9[1]
                                        wire _bundle_literal_expr_50: Ty60
                                        connect _bundle_literal_expr_50.imm_low, _match_arm_value_17.alu_common.common.imm_low
                                        wire _array_literal_expr_10: UInt<8>[1]
                                        connect _array_literal_expr_10[0], _match_arm_value_17.alu_common.common.src[2]
                                        connect _bundle_literal_expr_50.reversed_src, _array_literal_expr_10
                                        connect _bundle_literal_expr_50.imm_sign, _match_arm_value_17.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_15: Ty61
                                        connect _cast_bundle_to_bits_expr_15.imm_low, _bundle_literal_expr_50.imm_low
                                        connect _cast_bundle_to_bits_expr_15.reversed_src, _bundle_literal_expr_50.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_15.imm_sign, asUInt(_bundle_literal_expr_50.imm_sign)
                                        wire _cast_to_bits_expr_15: UInt<34>
                                        connect _cast_to_bits_expr_15, cat(_cast_bundle_to_bits_expr_15.imm_sign, cat(_cast_bundle_to_bits_expr_15.reversed_src, _cast_bundle_to_bits_expr_15.imm_low))
                                        wire _cast_bits_to_bundle_expr: Ty60
                                        wire _cast_bits_to_bundle_expr_flattened: Ty61
                                        connect _cast_bits_to_bundle_expr_flattened.imm_low, bits(asUInt(asSInt(_cast_to_bits_expr_15)), 24, 0)
                                        connect _cast_bits_to_bundle_expr.imm_low, _cast_bits_to_bundle_expr_flattened.imm_low
                                        connect _cast_bits_to_bundle_expr_flattened.reversed_src, bits(asUInt(asSInt(_cast_to_bits_expr_15)), 32, 25)
                                        wire _cast_bits_to_array_expr: UInt<8>[1]
                                        wire _cast_bits_to_array_expr_flattened: UInt<8>[1]
                                        connect _cast_bits_to_array_expr_flattened[0], bits(_cast_bits_to_bundle_expr_flattened.reversed_src, 7, 0)
                                        connect _cast_bits_to_array_expr[0], _cast_bits_to_array_expr_flattened[0]
                                        connect _cast_bits_to_bundle_expr.reversed_src, _cast_bits_to_array_expr
                                        connect _cast_bits_to_bundle_expr_flattened.imm_sign, bits(asUInt(asSInt(_cast_to_bits_expr_15)), 33, 33)
                                        connect _cast_bits_to_bundle_expr.imm_sign, asSInt(_cast_bits_to_bundle_expr_flattened.imm_sign)
                                        connect _array_literal_expr_7[2], tail(_cast_bits_to_bundle_expr.reversed_src[0], 2)
                                        connect _bundle_literal_expr_49.src, _array_literal_expr_7
                                        wire _bundle_literal_expr_51: Ty60
                                        connect _bundle_literal_expr_51.imm_low, _match_arm_value_17.alu_common.common.imm_low
                                        wire _array_literal_expr_11: UInt<8>[1]
                                        connect _array_literal_expr_11[0], _match_arm_value_17.alu_common.common.src[2]
                                        connect _bundle_literal_expr_51.reversed_src, _array_literal_expr_11
                                        connect _bundle_literal_expr_51.imm_sign, _match_arm_value_17.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_16: Ty61
                                        connect _cast_bundle_to_bits_expr_16.imm_low, _bundle_literal_expr_51.imm_low
                                        connect _cast_bundle_to_bits_expr_16.reversed_src, _bundle_literal_expr_51.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_16.imm_sign, asUInt(_bundle_literal_expr_51.imm_sign)
                                        wire _cast_to_bits_expr_16: UInt<34>
                                        connect _cast_to_bits_expr_16, cat(_cast_bundle_to_bits_expr_16.imm_sign, cat(_cast_bundle_to_bits_expr_16.reversed_src, _cast_bundle_to_bits_expr_16.imm_low))
                                        connect _bundle_literal_expr_49.imm_low, bits(asSInt(_cast_to_bits_expr_16), 24, 0)
                                        wire _bundle_literal_expr_52: Ty60
                                        connect _bundle_literal_expr_52.imm_low, _match_arm_value_17.alu_common.common.imm_low
                                        wire _array_literal_expr_12: UInt<8>[1]
                                        connect _array_literal_expr_12[0], _match_arm_value_17.alu_common.common.src[2]
                                        connect _bundle_literal_expr_52.reversed_src, _array_literal_expr_12
                                        connect _bundle_literal_expr_52.imm_sign, _match_arm_value_17.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_17: Ty61
                                        connect _cast_bundle_to_bits_expr_17.imm_low, _bundle_literal_expr_52.imm_low
                                        connect _cast_bundle_to_bits_expr_17.reversed_src, _bundle_literal_expr_52.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_17.imm_sign, asUInt(_bundle_literal_expr_52.imm_sign)
                                        wire _cast_to_bits_expr_17: UInt<34>
                                        connect _cast_to_bits_expr_17, cat(_cast_bundle_to_bits_expr_17.imm_sign, cat(_cast_bundle_to_bits_expr_17.reversed_src, _cast_bundle_to_bits_expr_17.imm_low))
                                        connect _bundle_literal_expr_49.imm_sign, shr(asSInt(_cast_to_bits_expr_17), 33)
                                        wire _bundle_literal_expr_53: Ty2
                                        invalidate _bundle_literal_expr_53
                                        connect _bundle_literal_expr_49._phantom, _bundle_literal_expr_53
                                        connect _bundle_literal_expr_48.common, _bundle_literal_expr_49
                                        connect _bundle_literal_expr_48.output_integer_mode, _match_arm_value_17.alu_common.output_integer_mode
                                        connect _bundle_literal_expr_47.alu_common, _bundle_literal_expr_48
                                        connect _bundle_literal_expr_47.invert_src0, _match_arm_value_17.invert_src0
                                        connect _bundle_literal_expr_47.invert_carry_in, _match_arm_value_17.invert_carry_in
                                        connect _bundle_literal_expr_47.invert_carry_out, _match_arm_value_17.invert_carry_out
                                        connect _bundle_literal_expr_47.add_pc, _match_arm_value_17.add_pc
                                        connect mapped_regs_1, {|AddSub: Ty44, AddSubI: Ty44, Logical: Ty45|}(AddSubI, _bundle_literal_expr_47) @[instruction.rs 502:1]
                                    Logical(_match_arm_value_18):
                                        wire _bundle_literal_expr_54: Ty1
                                        connect _bundle_literal_expr_54.value, _match_arm_value_18.alu_common.common.src[0]
                                        connect rename_0_src_0.addr, _bundle_literal_expr_54 @[reg_alloc.rs 204:29]
                                        wire _bundle_literal_expr_55: Ty1
                                        connect _bundle_literal_expr_55.value, _match_arm_value_18.alu_common.common.src[1]
                                        connect rename_0_src_1.addr, _bundle_literal_expr_55 @[reg_alloc.rs 204:29]
                                        wire _bundle_literal_expr_56: Ty45
                                        wire _bundle_literal_expr_57: Ty43
                                        wire _bundle_literal_expr_58: Ty42
                                        connect _bundle_literal_expr_58.prefix_pad, _match_arm_value_18.alu_common.common.prefix_pad
                                        connect _bundle_literal_expr_58.dest, _match_arm_value_4.unit_out_reg
                                        wire _array_literal_expr_13: UInt<6>[3]
                                        wire _array_literal_expr_14: UInt<6>[2]
                                        wire _cast_bundle_to_bits_expr_18: Ty57
                                        connect _cast_bundle_to_bits_expr_18.unit_num, rename_0_src_0.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_18.unit_out_reg, rename_0_src_0.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_18: UInt<6>
                                        connect _cast_to_bits_expr_18, cat(_cast_bundle_to_bits_expr_18.unit_out_reg, _cast_bundle_to_bits_expr_18.unit_num)
                                        connect _array_literal_expr_14[0], _cast_to_bits_expr_18
                                        wire _cast_bundle_to_bits_expr_19: Ty57
                                        connect _cast_bundle_to_bits_expr_19.unit_num, rename_0_src_1.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_19.unit_out_reg, rename_0_src_1.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_19: UInt<6>
                                        connect _cast_to_bits_expr_19, cat(_cast_bundle_to_bits_expr_19.unit_out_reg, _cast_bundle_to_bits_expr_19.unit_num)
                                        connect _array_literal_expr_14[1], _cast_to_bits_expr_19
                                        connect _array_literal_expr_13[0], _array_literal_expr_14[0]
                                        wire _array_literal_expr_15: UInt<6>[2]
                                        wire _cast_bundle_to_bits_expr_20: Ty57
                                        connect _cast_bundle_to_bits_expr_20.unit_num, rename_0_src_0.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_20.unit_out_reg, rename_0_src_0.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_20: UInt<6>
                                        connect _cast_to_bits_expr_20, cat(_cast_bundle_to_bits_expr_20.unit_out_reg, _cast_bundle_to_bits_expr_20.unit_num)
                                        connect _array_literal_expr_15[0], _cast_to_bits_expr_20
                                        wire _cast_bundle_to_bits_expr_21: Ty57
                                        connect _cast_bundle_to_bits_expr_21.unit_num, rename_0_src_1.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_21.unit_out_reg, rename_0_src_1.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_21: UInt<6>
                                        connect _cast_to_bits_expr_21, cat(_cast_bundle_to_bits_expr_21.unit_out_reg, _cast_bundle_to_bits_expr_21.unit_num)
                                        connect _array_literal_expr_15[1], _cast_to_bits_expr_21
                                        connect _array_literal_expr_13[1], _array_literal_expr_15[1]
                                        wire _bundle_literal_expr_59: Ty60
                                        connect _bundle_literal_expr_59.imm_low, _match_arm_value_18.alu_common.common.imm_low
                                        wire _array_literal_expr_16: UInt<8>[1]
                                        connect _array_literal_expr_16[0], _match_arm_value_18.alu_common.common.src[2]
                                        connect _bundle_literal_expr_59.reversed_src, _array_literal_expr_16
                                        connect _bundle_literal_expr_59.imm_sign, _match_arm_value_18.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_22: Ty61
                                        connect _cast_bundle_to_bits_expr_22.imm_low, _bundle_literal_expr_59.imm_low
                                        connect _cast_bundle_to_bits_expr_22.reversed_src, _bundle_literal_expr_59.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_22.imm_sign, asUInt(_bundle_literal_expr_59.imm_sign)
                                        wire _cast_to_bits_expr_22: UInt<34>
                                        connect _cast_to_bits_expr_22, cat(_cast_bundle_to_bits_expr_22.imm_sign, cat(_cast_bundle_to_bits_expr_22.reversed_src, _cast_bundle_to_bits_expr_22.imm_low))
                                        wire _cast_bits_to_bundle_expr_1: Ty60
                                        wire _cast_bits_to_bundle_expr_flattened_1: Ty61
                                        connect _cast_bits_to_bundle_expr_flattened_1.imm_low, bits(asUInt(asSInt(_cast_to_bits_expr_22)), 24, 0)
                                        connect _cast_bits_to_bundle_expr_1.imm_low, _cast_bits_to_bundle_expr_flattened_1.imm_low
                                        connect _cast_bits_to_bundle_expr_flattened_1.reversed_src, bits(asUInt(asSInt(_cast_to_bits_expr_22)), 32, 25)
                                        wire _cast_bits_to_array_expr_1: UInt<8>[1]
                                        wire _cast_bits_to_array_expr_flattened_1: UInt<8>[1]
                                        connect _cast_bits_to_array_expr_flattened_1[0], bits(_cast_bits_to_bundle_expr_flattened_1.reversed_src, 7, 0)
                                        connect _cast_bits_to_array_expr_1[0], _cast_bits_to_array_expr_flattened_1[0]
                                        connect _cast_bits_to_bundle_expr_1.reversed_src, _cast_bits_to_array_expr_1
                                        connect _cast_bits_to_bundle_expr_flattened_1.imm_sign, bits(asUInt(asSInt(_cast_to_bits_expr_22)), 33, 33)
                                        connect _cast_bits_to_bundle_expr_1.imm_sign, asSInt(_cast_bits_to_bundle_expr_flattened_1.imm_sign)
                                        connect _array_literal_expr_13[2], tail(_cast_bits_to_bundle_expr_1.reversed_src[0], 2)
                                        connect _bundle_literal_expr_58.src, _array_literal_expr_13
                                        wire _bundle_literal_expr_60: Ty60
                                        connect _bundle_literal_expr_60.imm_low, _match_arm_value_18.alu_common.common.imm_low
                                        wire _array_literal_expr_17: UInt<8>[1]
                                        connect _array_literal_expr_17[0], _match_arm_value_18.alu_common.common.src[2]
                                        connect _bundle_literal_expr_60.reversed_src, _array_literal_expr_17
                                        connect _bundle_literal_expr_60.imm_sign, _match_arm_value_18.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_23: Ty61
                                        connect _cast_bundle_to_bits_expr_23.imm_low, _bundle_literal_expr_60.imm_low
                                        connect _cast_bundle_to_bits_expr_23.reversed_src, _bundle_literal_expr_60.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_23.imm_sign, asUInt(_bundle_literal_expr_60.imm_sign)
                                        wire _cast_to_bits_expr_23: UInt<34>
                                        connect _cast_to_bits_expr_23, cat(_cast_bundle_to_bits_expr_23.imm_sign, cat(_cast_bundle_to_bits_expr_23.reversed_src, _cast_bundle_to_bits_expr_23.imm_low))
                                        connect _bundle_literal_expr_58.imm_low, bits(asSInt(_cast_to_bits_expr_23), 24, 0)
                                        wire _bundle_literal_expr_61: Ty60
                                        connect _bundle_literal_expr_61.imm_low, _match_arm_value_18.alu_common.common.imm_low
                                        wire _array_literal_expr_18: UInt<8>[1]
                                        connect _array_literal_expr_18[0], _match_arm_value_18.alu_common.common.src[2]
                                        connect _bundle_literal_expr_61.reversed_src, _array_literal_expr_18
                                        connect _bundle_literal_expr_61.imm_sign, _match_arm_value_18.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_24: Ty61
                                        connect _cast_bundle_to_bits_expr_24.imm_low, _bundle_literal_expr_61.imm_low
                                        connect _cast_bundle_to_bits_expr_24.reversed_src, _bundle_literal_expr_61.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_24.imm_sign, asUInt(_bundle_literal_expr_61.imm_sign)
                                        wire _cast_to_bits_expr_24: UInt<34>
                                        connect _cast_to_bits_expr_24, cat(_cast_bundle_to_bits_expr_24.imm_sign, cat(_cast_bundle_to_bits_expr_24.reversed_src, _cast_bundle_to_bits_expr_24.imm_low))
                                        connect _bundle_literal_expr_58.imm_sign, shr(asSInt(_cast_to_bits_expr_24), 33)
                                        wire _bundle_literal_expr_62: Ty2
                                        invalidate _bundle_literal_expr_62
                                        connect _bundle_literal_expr_58._phantom, _bundle_literal_expr_62
                                        connect _bundle_literal_expr_57.common, _bundle_literal_expr_58
                                        connect _bundle_literal_expr_57.output_integer_mode, _match_arm_value_18.alu_common.output_integer_mode
                                        connect _bundle_literal_expr_56.alu_common, _bundle_literal_expr_57
                                        connect _bundle_literal_expr_56.lut, _match_arm_value_18.lut
                                        connect mapped_regs_1, {|AddSub: Ty44, AddSubI: Ty44, Logical: Ty45|}(Logical, _bundle_literal_expr_56) @[instruction.rs 502:1]
                                connect mapped_regs, {|AluBranch: Ty46, L2RegisterFile: Ty49, LoadStore: Ty50|}(AluBranch, mapped_regs_1) @[unit.rs 129:1]
                            L2RegisterFile(_match_arm_value_19):
                                wire mapped_regs_2: Ty49 @[instruction.rs 529:1]
                                match _match_arm_value_19: @[instruction.rs 529:1]
                                    ReadL2Reg(_match_arm_value_20):
                                        wire _bundle_literal_expr_63: Ty48
                                        wire _bundle_literal_expr_64: Ty47
                                        connect _bundle_literal_expr_64.prefix_pad, _match_arm_value_20.common.prefix_pad
                                        connect _bundle_literal_expr_64.dest, _match_arm_value_4.unit_out_reg
                                        wire _array_literal_expr_19: UInt<6>[3]
                                        connect _array_literal_expr_19[0], pad(UInt<0>(0h0), 6)
                                        connect _array_literal_expr_19[1], pad(UInt<0>(0h0), 6)
                                        wire _bundle_literal_expr_65: Ty60
                                        connect _bundle_literal_expr_65.imm_low, _match_arm_value_20.common.imm_low
                                        wire _array_literal_expr_20: UInt<8>[1]
                                        connect _array_literal_expr_20[0], _match_arm_value_20.common.src[2]
                                        connect _bundle_literal_expr_65.reversed_src, _array_literal_expr_20
                                        connect _bundle_literal_expr_65.imm_sign, _match_arm_value_20.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_25: Ty61
                                        connect _cast_bundle_to_bits_expr_25.imm_low, _bundle_literal_expr_65.imm_low
                                        connect _cast_bundle_to_bits_expr_25.reversed_src, _bundle_literal_expr_65.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_25.imm_sign, asUInt(_bundle_literal_expr_65.imm_sign)
                                        wire _cast_to_bits_expr_25: UInt<34>
                                        connect _cast_to_bits_expr_25, cat(_cast_bundle_to_bits_expr_25.imm_sign, cat(_cast_bundle_to_bits_expr_25.reversed_src, _cast_bundle_to_bits_expr_25.imm_low))
                                        wire _cast_bits_to_bundle_expr_2: Ty60
                                        wire _cast_bits_to_bundle_expr_flattened_2: Ty61
                                        connect _cast_bits_to_bundle_expr_flattened_2.imm_low, bits(asUInt(asSInt(_cast_to_bits_expr_25)), 24, 0)
                                        connect _cast_bits_to_bundle_expr_2.imm_low, _cast_bits_to_bundle_expr_flattened_2.imm_low
                                        connect _cast_bits_to_bundle_expr_flattened_2.reversed_src, bits(asUInt(asSInt(_cast_to_bits_expr_25)), 32, 25)
                                        wire _cast_bits_to_array_expr_2: UInt<8>[1]
                                        wire _cast_bits_to_array_expr_flattened_2: UInt<8>[1]
                                        connect _cast_bits_to_array_expr_flattened_2[0], bits(_cast_bits_to_bundle_expr_flattened_2.reversed_src, 7, 0)
                                        connect _cast_bits_to_array_expr_2[0], _cast_bits_to_array_expr_flattened_2[0]
                                        connect _cast_bits_to_bundle_expr_2.reversed_src, _cast_bits_to_array_expr_2
                                        connect _cast_bits_to_bundle_expr_flattened_2.imm_sign, bits(asUInt(asSInt(_cast_to_bits_expr_25)), 33, 33)
                                        connect _cast_bits_to_bundle_expr_2.imm_sign, asSInt(_cast_bits_to_bundle_expr_flattened_2.imm_sign)
                                        connect _array_literal_expr_19[2], tail(_cast_bits_to_bundle_expr_2.reversed_src[0], 2)
                                        connect _bundle_literal_expr_64.src, _array_literal_expr_19
                                        wire _bundle_literal_expr_66: Ty60
                                        connect _bundle_literal_expr_66.imm_low, _match_arm_value_20.common.imm_low
                                        wire _array_literal_expr_21: UInt<8>[1]
                                        connect _array_literal_expr_21[0], _match_arm_value_20.common.src[2]
                                        connect _bundle_literal_expr_66.reversed_src, _array_literal_expr_21
                                        connect _bundle_literal_expr_66.imm_sign, _match_arm_value_20.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_26: Ty61
                                        connect _cast_bundle_to_bits_expr_26.imm_low, _bundle_literal_expr_66.imm_low
                                        connect _cast_bundle_to_bits_expr_26.reversed_src, _bundle_literal_expr_66.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_26.imm_sign, asUInt(_bundle_literal_expr_66.imm_sign)
                                        wire _cast_to_bits_expr_26: UInt<34>
                                        connect _cast_to_bits_expr_26, cat(_cast_bundle_to_bits_expr_26.imm_sign, cat(_cast_bundle_to_bits_expr_26.reversed_src, _cast_bundle_to_bits_expr_26.imm_low))
                                        connect _bundle_literal_expr_64.imm_low, bits(asSInt(_cast_to_bits_expr_26), 24, 0)
                                        wire _bundle_literal_expr_67: Ty60
                                        connect _bundle_literal_expr_67.imm_low, _match_arm_value_20.common.imm_low
                                        wire _array_literal_expr_22: UInt<8>[1]
                                        connect _array_literal_expr_22[0], _match_arm_value_20.common.src[2]
                                        connect _bundle_literal_expr_67.reversed_src, _array_literal_expr_22
                                        connect _bundle_literal_expr_67.imm_sign, _match_arm_value_20.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_27: Ty61
                                        connect _cast_bundle_to_bits_expr_27.imm_low, _bundle_literal_expr_67.imm_low
                                        connect _cast_bundle_to_bits_expr_27.reversed_src, _bundle_literal_expr_67.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_27.imm_sign, asUInt(_bundle_literal_expr_67.imm_sign)
                                        wire _cast_to_bits_expr_27: UInt<34>
                                        connect _cast_to_bits_expr_27, cat(_cast_bundle_to_bits_expr_27.imm_sign, cat(_cast_bundle_to_bits_expr_27.reversed_src, _cast_bundle_to_bits_expr_27.imm_low))
                                        connect _bundle_literal_expr_64.imm_sign, shr(asSInt(_cast_to_bits_expr_27), 33)
                                        wire _bundle_literal_expr_68: Ty2
                                        invalidate _bundle_literal_expr_68
                                        connect _bundle_literal_expr_64._phantom, _bundle_literal_expr_68
                                        connect _bundle_literal_expr_63.common, _bundle_literal_expr_64
                                        connect mapped_regs_2, {|ReadL2Reg: Ty48, WriteL2Reg: Ty48|}(ReadL2Reg, _bundle_literal_expr_63) @[instruction.rs 529:1]
                                    WriteL2Reg(_match_arm_value_21):
                                        wire _bundle_literal_expr_69: Ty1
                                        connect _bundle_literal_expr_69.value, _match_arm_value_21.common.src[0]
                                        connect rename_0_src_0.addr, _bundle_literal_expr_69 @[reg_alloc.rs 204:29]
                                        wire _bundle_literal_expr_70: Ty48
                                        wire _bundle_literal_expr_71: Ty47
                                        connect _bundle_literal_expr_71.prefix_pad, _match_arm_value_21.common.prefix_pad
                                        connect _bundle_literal_expr_71.dest, _match_arm_value_4.unit_out_reg
                                        wire _array_literal_expr_23: UInt<6>[3]
                                        wire _array_literal_expr_24: UInt<6>[1]
                                        wire _cast_bundle_to_bits_expr_28: Ty57
                                        connect _cast_bundle_to_bits_expr_28.unit_num, rename_0_src_0.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_28.unit_out_reg, rename_0_src_0.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_28: UInt<6>
                                        connect _cast_to_bits_expr_28, cat(_cast_bundle_to_bits_expr_28.unit_out_reg, _cast_bundle_to_bits_expr_28.unit_num)
                                        connect _array_literal_expr_24[0], _cast_to_bits_expr_28
                                        connect _array_literal_expr_23[0], _array_literal_expr_24[0]
                                        connect _array_literal_expr_23[1], pad(UInt<0>(0h0), 6)
                                        wire _bundle_literal_expr_72: Ty60
                                        connect _bundle_literal_expr_72.imm_low, _match_arm_value_21.common.imm_low
                                        wire _array_literal_expr_25: UInt<8>[1]
                                        connect _array_literal_expr_25[0], _match_arm_value_21.common.src[2]
                                        connect _bundle_literal_expr_72.reversed_src, _array_literal_expr_25
                                        connect _bundle_literal_expr_72.imm_sign, _match_arm_value_21.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_29: Ty61
                                        connect _cast_bundle_to_bits_expr_29.imm_low, _bundle_literal_expr_72.imm_low
                                        connect _cast_bundle_to_bits_expr_29.reversed_src, _bundle_literal_expr_72.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_29.imm_sign, asUInt(_bundle_literal_expr_72.imm_sign)
                                        wire _cast_to_bits_expr_29: UInt<34>
                                        connect _cast_to_bits_expr_29, cat(_cast_bundle_to_bits_expr_29.imm_sign, cat(_cast_bundle_to_bits_expr_29.reversed_src, _cast_bundle_to_bits_expr_29.imm_low))
                                        wire _cast_bits_to_bundle_expr_3: Ty60
                                        wire _cast_bits_to_bundle_expr_flattened_3: Ty61
                                        connect _cast_bits_to_bundle_expr_flattened_3.imm_low, bits(asUInt(asSInt(_cast_to_bits_expr_29)), 24, 0)
                                        connect _cast_bits_to_bundle_expr_3.imm_low, _cast_bits_to_bundle_expr_flattened_3.imm_low
                                        connect _cast_bits_to_bundle_expr_flattened_3.reversed_src, bits(asUInt(asSInt(_cast_to_bits_expr_29)), 32, 25)
                                        wire _cast_bits_to_array_expr_3: UInt<8>[1]
                                        wire _cast_bits_to_array_expr_flattened_3: UInt<8>[1]
                                        connect _cast_bits_to_array_expr_flattened_3[0], bits(_cast_bits_to_bundle_expr_flattened_3.reversed_src, 7, 0)
                                        connect _cast_bits_to_array_expr_3[0], _cast_bits_to_array_expr_flattened_3[0]
                                        connect _cast_bits_to_bundle_expr_3.reversed_src, _cast_bits_to_array_expr_3
                                        connect _cast_bits_to_bundle_expr_flattened_3.imm_sign, bits(asUInt(asSInt(_cast_to_bits_expr_29)), 33, 33)
                                        connect _cast_bits_to_bundle_expr_3.imm_sign, asSInt(_cast_bits_to_bundle_expr_flattened_3.imm_sign)
                                        connect _array_literal_expr_23[2], tail(_cast_bits_to_bundle_expr_3.reversed_src[0], 2)
                                        connect _bundle_literal_expr_71.src, _array_literal_expr_23
                                        wire _bundle_literal_expr_73: Ty60
                                        connect _bundle_literal_expr_73.imm_low, _match_arm_value_21.common.imm_low
                                        wire _array_literal_expr_26: UInt<8>[1]
                                        connect _array_literal_expr_26[0], _match_arm_value_21.common.src[2]
                                        connect _bundle_literal_expr_73.reversed_src, _array_literal_expr_26
                                        connect _bundle_literal_expr_73.imm_sign, _match_arm_value_21.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_30: Ty61
                                        connect _cast_bundle_to_bits_expr_30.imm_low, _bundle_literal_expr_73.imm_low
                                        connect _cast_bundle_to_bits_expr_30.reversed_src, _bundle_literal_expr_73.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_30.imm_sign, asUInt(_bundle_literal_expr_73.imm_sign)
                                        wire _cast_to_bits_expr_30: UInt<34>
                                        connect _cast_to_bits_expr_30, cat(_cast_bundle_to_bits_expr_30.imm_sign, cat(_cast_bundle_to_bits_expr_30.reversed_src, _cast_bundle_to_bits_expr_30.imm_low))
                                        connect _bundle_literal_expr_71.imm_low, bits(asSInt(_cast_to_bits_expr_30), 24, 0)
                                        wire _bundle_literal_expr_74: Ty60
                                        connect _bundle_literal_expr_74.imm_low, _match_arm_value_21.common.imm_low
                                        wire _array_literal_expr_27: UInt<8>[1]
                                        connect _array_literal_expr_27[0], _match_arm_value_21.common.src[2]
                                        connect _bundle_literal_expr_74.reversed_src, _array_literal_expr_27
                                        connect _bundle_literal_expr_74.imm_sign, _match_arm_value_21.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_31: Ty61
                                        connect _cast_bundle_to_bits_expr_31.imm_low, _bundle_literal_expr_74.imm_low
                                        connect _cast_bundle_to_bits_expr_31.reversed_src, _bundle_literal_expr_74.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_31.imm_sign, asUInt(_bundle_literal_expr_74.imm_sign)
                                        wire _cast_to_bits_expr_31: UInt<34>
                                        connect _cast_to_bits_expr_31, cat(_cast_bundle_to_bits_expr_31.imm_sign, cat(_cast_bundle_to_bits_expr_31.reversed_src, _cast_bundle_to_bits_expr_31.imm_low))
                                        connect _bundle_literal_expr_71.imm_sign, shr(asSInt(_cast_to_bits_expr_31), 33)
                                        wire _bundle_literal_expr_75: Ty2
                                        invalidate _bundle_literal_expr_75
                                        connect _bundle_literal_expr_71._phantom, _bundle_literal_expr_75
                                        connect _bundle_literal_expr_70.common, _bundle_literal_expr_71
                                        connect mapped_regs_2, {|ReadL2Reg: Ty48, WriteL2Reg: Ty48|}(WriteL2Reg, _bundle_literal_expr_70) @[instruction.rs 529:1]
                                connect mapped_regs, {|AluBranch: Ty46, L2RegisterFile: Ty49, LoadStore: Ty50|}(L2RegisterFile, mapped_regs_2) @[unit.rs 129:1]
                            LoadStore(_match_arm_value_22):
                                wire mapped_regs_3: Ty50 @[instruction.rs 564:1]
                                match _match_arm_value_22: @[instruction.rs 564:1]
                                    Load(_match_arm_value_23):
                                        wire _bundle_literal_expr_76: Ty47
                                        connect _bundle_literal_expr_76.prefix_pad, _match_arm_value_23.prefix_pad
                                        connect _bundle_literal_expr_76.dest, _match_arm_value_4.unit_out_reg
                                        wire _array_literal_expr_28: UInt<6>[3]
                                        connect _array_literal_expr_28[0], pad(UInt<0>(0h0), 6)
                                        connect _array_literal_expr_28[1], pad(UInt<0>(0h0), 6)
                                        wire _bundle_literal_expr_77: Ty60
                                        connect _bundle_literal_expr_77.imm_low, _match_arm_value_23.imm_low
                                        wire _array_literal_expr_29: UInt<8>[1]
                                        connect _array_literal_expr_29[0], _match_arm_value_23.src[2]
                                        connect _bundle_literal_expr_77.reversed_src, _array_literal_expr_29
                                        connect _bundle_literal_expr_77.imm_sign, _match_arm_value_23.imm_sign
                                        wire _cast_bundle_to_bits_expr_32: Ty61
                                        connect _cast_bundle_to_bits_expr_32.imm_low, _bundle_literal_expr_77.imm_low
                                        connect _cast_bundle_to_bits_expr_32.reversed_src, _bundle_literal_expr_77.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_32.imm_sign, asUInt(_bundle_literal_expr_77.imm_sign)
                                        wire _cast_to_bits_expr_32: UInt<34>
                                        connect _cast_to_bits_expr_32, cat(_cast_bundle_to_bits_expr_32.imm_sign, cat(_cast_bundle_to_bits_expr_32.reversed_src, _cast_bundle_to_bits_expr_32.imm_low))
                                        wire _cast_bits_to_bundle_expr_4: Ty60
                                        wire _cast_bits_to_bundle_expr_flattened_4: Ty61
                                        connect _cast_bits_to_bundle_expr_flattened_4.imm_low, bits(asUInt(asSInt(_cast_to_bits_expr_32)), 24, 0)
                                        connect _cast_bits_to_bundle_expr_4.imm_low, _cast_bits_to_bundle_expr_flattened_4.imm_low
                                        connect _cast_bits_to_bundle_expr_flattened_4.reversed_src, bits(asUInt(asSInt(_cast_to_bits_expr_32)), 32, 25)
                                        wire _cast_bits_to_array_expr_4: UInt<8>[1]
                                        wire _cast_bits_to_array_expr_flattened_4: UInt<8>[1]
                                        connect _cast_bits_to_array_expr_flattened_4[0], bits(_cast_bits_to_bundle_expr_flattened_4.reversed_src, 7, 0)
                                        connect _cast_bits_to_array_expr_4[0], _cast_bits_to_array_expr_flattened_4[0]
                                        connect _cast_bits_to_bundle_expr_4.reversed_src, _cast_bits_to_array_expr_4
                                        connect _cast_bits_to_bundle_expr_flattened_4.imm_sign, bits(asUInt(asSInt(_cast_to_bits_expr_32)), 33, 33)
                                        connect _cast_bits_to_bundle_expr_4.imm_sign, asSInt(_cast_bits_to_bundle_expr_flattened_4.imm_sign)
                                        connect _array_literal_expr_28[2], tail(_cast_bits_to_bundle_expr_4.reversed_src[0], 2)
                                        connect _bundle_literal_expr_76.src, _array_literal_expr_28
                                        wire _bundle_literal_expr_78: Ty60
                                        connect _bundle_literal_expr_78.imm_low, _match_arm_value_23.imm_low
                                        wire _array_literal_expr_30: UInt<8>[1]
                                        connect _array_literal_expr_30[0], _match_arm_value_23.src[2]
                                        connect _bundle_literal_expr_78.reversed_src, _array_literal_expr_30
                                        connect _bundle_literal_expr_78.imm_sign, _match_arm_value_23.imm_sign
                                        wire _cast_bundle_to_bits_expr_33: Ty61
                                        connect _cast_bundle_to_bits_expr_33.imm_low, _bundle_literal_expr_78.imm_low
                                        connect _cast_bundle_to_bits_expr_33.reversed_src, _bundle_literal_expr_78.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_33.imm_sign, asUInt(_bundle_literal_expr_78.imm_sign)
                                        wire _cast_to_bits_expr_33: UInt<34>
                                        connect _cast_to_bits_expr_33, cat(_cast_bundle_to_bits_expr_33.imm_sign, cat(_cast_bundle_to_bits_expr_33.reversed_src, _cast_bundle_to_bits_expr_33.imm_low))
                                        connect _bundle_literal_expr_76.imm_low, bits(asSInt(_cast_to_bits_expr_33), 24, 0)
                                        wire _bundle_literal_expr_79: Ty60
                                        connect _bundle_literal_expr_79.imm_low, _match_arm_value_23.imm_low
                                        wire _array_literal_expr_31: UInt<8>[1]
                                        connect _array_literal_expr_31[0], _match_arm_value_23.src[2]
                                        connect _bundle_literal_expr_79.reversed_src, _array_literal_expr_31
                                        connect _bundle_literal_expr_79.imm_sign, _match_arm_value_23.imm_sign
                                        wire _cast_bundle_to_bits_expr_34: Ty61
                                        connect _cast_bundle_to_bits_expr_34.imm_low, _bundle_literal_expr_79.imm_low
                                        connect _cast_bundle_to_bits_expr_34.reversed_src, _bundle_literal_expr_79.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_34.imm_sign, asUInt(_bundle_literal_expr_79.imm_sign)
                                        wire _cast_to_bits_expr_34: UInt<34>
                                        connect _cast_to_bits_expr_34, cat(_cast_bundle_to_bits_expr_34.imm_sign, cat(_cast_bundle_to_bits_expr_34.reversed_src, _cast_bundle_to_bits_expr_34.imm_low))
                                        connect _bundle_literal_expr_76.imm_sign, shr(asSInt(_cast_to_bits_expr_34), 33)
                                        wire _bundle_literal_expr_80: Ty2
                                        invalidate _bundle_literal_expr_80
                                        connect _bundle_literal_expr_76._phantom, _bundle_literal_expr_80
                                        connect mapped_regs_3, {|Load: Ty47, Store: Ty47|}(Load, _bundle_literal_expr_76) @[instruction.rs 564:1]
                                    Store(_match_arm_value_24):
                                        wire _bundle_literal_expr_81: Ty1
                                        connect _bundle_literal_expr_81.value, _match_arm_value_24.src[0]
                                        connect rename_0_src_0.addr, _bundle_literal_expr_81 @[reg_alloc.rs 204:29]
                                        wire _bundle_literal_expr_82: Ty47
                                        connect _bundle_literal_expr_82.prefix_pad, _match_arm_value_24.prefix_pad
                                        connect _bundle_literal_expr_82.dest, _match_arm_value_4.unit_out_reg
                                        wire _array_literal_expr_32: UInt<6>[3]
                                        wire _array_literal_expr_33: UInt<6>[1]
                                        wire _cast_bundle_to_bits_expr_35: Ty57
                                        connect _cast_bundle_to_bits_expr_35.unit_num, rename_0_src_0.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_35.unit_out_reg, rename_0_src_0.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_35: UInt<6>
                                        connect _cast_to_bits_expr_35, cat(_cast_bundle_to_bits_expr_35.unit_out_reg, _cast_bundle_to_bits_expr_35.unit_num)
                                        connect _array_literal_expr_33[0], _cast_to_bits_expr_35
                                        connect _array_literal_expr_32[0], _array_literal_expr_33[0]
                                        connect _array_literal_expr_32[1], pad(UInt<0>(0h0), 6)
                                        wire _bundle_literal_expr_83: Ty60
                                        connect _bundle_literal_expr_83.imm_low, _match_arm_value_24.imm_low
                                        wire _array_literal_expr_34: UInt<8>[1]
                                        connect _array_literal_expr_34[0], _match_arm_value_24.src[2]
                                        connect _bundle_literal_expr_83.reversed_src, _array_literal_expr_34
                                        connect _bundle_literal_expr_83.imm_sign, _match_arm_value_24.imm_sign
                                        wire _cast_bundle_to_bits_expr_36: Ty61
                                        connect _cast_bundle_to_bits_expr_36.imm_low, _bundle_literal_expr_83.imm_low
                                        connect _cast_bundle_to_bits_expr_36.reversed_src, _bundle_literal_expr_83.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_36.imm_sign, asUInt(_bundle_literal_expr_83.imm_sign)
                                        wire _cast_to_bits_expr_36: UInt<34>
                                        connect _cast_to_bits_expr_36, cat(_cast_bundle_to_bits_expr_36.imm_sign, cat(_cast_bundle_to_bits_expr_36.reversed_src, _cast_bundle_to_bits_expr_36.imm_low))
                                        wire _cast_bits_to_bundle_expr_5: Ty60
                                        wire _cast_bits_to_bundle_expr_flattened_5: Ty61
                                        connect _cast_bits_to_bundle_expr_flattened_5.imm_low, bits(asUInt(asSInt(_cast_to_bits_expr_36)), 24, 0)
                                        connect _cast_bits_to_bundle_expr_5.imm_low, _cast_bits_to_bundle_expr_flattened_5.imm_low
                                        connect _cast_bits_to_bundle_expr_flattened_5.reversed_src, bits(asUInt(asSInt(_cast_to_bits_expr_36)), 32, 25)
                                        wire _cast_bits_to_array_expr_5: UInt<8>[1]
                                        wire _cast_bits_to_array_expr_flattened_5: UInt<8>[1]
                                        connect _cast_bits_to_array_expr_flattened_5[0], bits(_cast_bits_to_bundle_expr_flattened_5.reversed_src, 7, 0)
                                        connect _cast_bits_to_array_expr_5[0], _cast_bits_to_array_expr_flattened_5[0]
                                        connect _cast_bits_to_bundle_expr_5.reversed_src, _cast_bits_to_array_expr_5
                                        connect _cast_bits_to_bundle_expr_flattened_5.imm_sign, bits(asUInt(asSInt(_cast_to_bits_expr_36)), 33, 33)
                                        connect _cast_bits_to_bundle_expr_5.imm_sign, asSInt(_cast_bits_to_bundle_expr_flattened_5.imm_sign)
                                        connect _array_literal_expr_32[2], tail(_cast_bits_to_bundle_expr_5.reversed_src[0], 2)
                                        connect _bundle_literal_expr_82.src, _array_literal_expr_32
                                        wire _bundle_literal_expr_84: Ty60
                                        connect _bundle_literal_expr_84.imm_low, _match_arm_value_24.imm_low
                                        wire _array_literal_expr_35: UInt<8>[1]
                                        connect _array_literal_expr_35[0], _match_arm_value_24.src[2]
                                        connect _bundle_literal_expr_84.reversed_src, _array_literal_expr_35
                                        connect _bundle_literal_expr_84.imm_sign, _match_arm_value_24.imm_sign
                                        wire _cast_bundle_to_bits_expr_37: Ty61
                                        connect _cast_bundle_to_bits_expr_37.imm_low, _bundle_literal_expr_84.imm_low
                                        connect _cast_bundle_to_bits_expr_37.reversed_src, _bundle_literal_expr_84.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_37.imm_sign, asUInt(_bundle_literal_expr_84.imm_sign)
                                        wire _cast_to_bits_expr_37: UInt<34>
                                        connect _cast_to_bits_expr_37, cat(_cast_bundle_to_bits_expr_37.imm_sign, cat(_cast_bundle_to_bits_expr_37.reversed_src, _cast_bundle_to_bits_expr_37.imm_low))
                                        connect _bundle_literal_expr_82.imm_low, bits(asSInt(_cast_to_bits_expr_37), 24, 0)
                                        wire _bundle_literal_expr_85: Ty60
                                        connect _bundle_literal_expr_85.imm_low, _match_arm_value_24.imm_low
                                        wire _array_literal_expr_36: UInt<8>[1]
                                        connect _array_literal_expr_36[0], _match_arm_value_24.src[2]
                                        connect _bundle_literal_expr_85.reversed_src, _array_literal_expr_36
                                        connect _bundle_literal_expr_85.imm_sign, _match_arm_value_24.imm_sign
                                        wire _cast_bundle_to_bits_expr_38: Ty61
                                        connect _cast_bundle_to_bits_expr_38.imm_low, _bundle_literal_expr_85.imm_low
                                        connect _cast_bundle_to_bits_expr_38.reversed_src, _bundle_literal_expr_85.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_38.imm_sign, asUInt(_bundle_literal_expr_85.imm_sign)
                                        wire _cast_to_bits_expr_38: UInt<34>
                                        connect _cast_to_bits_expr_38, cat(_cast_bundle_to_bits_expr_38.imm_sign, cat(_cast_bundle_to_bits_expr_38.reversed_src, _cast_bundle_to_bits_expr_38.imm_low))
                                        connect _bundle_literal_expr_82.imm_sign, shr(asSInt(_cast_to_bits_expr_38), 33)
                                        wire _bundle_literal_expr_86: Ty2
                                        invalidate _bundle_literal_expr_86
                                        connect _bundle_literal_expr_82._phantom, _bundle_literal_expr_86
                                        connect mapped_regs_3, {|Load: Ty47, Store: Ty47|}(Store, _bundle_literal_expr_82) @[instruction.rs 564:1]
                                connect mapped_regs, {|AluBranch: Ty46, L2RegisterFile: Ty49, LoadStore: Ty50|}(LoadStore, mapped_regs_3) @[unit.rs 129:1]
                        connect renamed_mops[0], {|HdlNone, HdlSome: Ty51|}(HdlSome, mapped_regs) @[reg_alloc.rs 197:17]
                        wire flag_reg: Ty1 @[instruction.rs 831:32]
                        wire _bundle_literal_expr_87: Ty1
                        connect _bundle_literal_expr_87.value, tail(UInt<32>(0h0), 24)
                        connect flag_reg, _bundle_literal_expr_87 @[instruction.rs 832:17]
                        match dest_reg.flag_regs[0]: @[instruction.rs 834:17]
                            HdlNone:
                                skip
                            HdlSome(_match_arm_value_25):
                                wire _bundle_literal_expr_88: Ty1
                                connect _bundle_literal_expr_88.value, tail(UInt<32>(0hFE), 24)
                                connect flag_reg, _bundle_literal_expr_88 @[instruction.rs 836:21]
                        wire flag_reg_1: Ty1 @[instruction.rs 831:32]
                        wire _bundle_literal_expr_89: Ty1
                        connect _bundle_literal_expr_89.value, tail(UInt<32>(0h0), 24)
                        connect flag_reg_1, _bundle_literal_expr_89 @[instruction.rs 832:17]
                        match dest_reg.flag_regs[1]: @[instruction.rs 834:17]
                            HdlNone:
                                skip
                            HdlSome(_match_arm_value_26):
                                wire _bundle_literal_expr_90: Ty1
                                connect _bundle_literal_expr_90.value, tail(UInt<32>(0hFF), 24)
                                connect flag_reg_1, _bundle_literal_expr_90 @[instruction.rs 836:21]
                        when and(geq(dest_reg.normal_regs[0].value, UInt<32>(0h1)), lt(dest_reg.normal_regs[0].value, UInt<32>(0hFE))): @[reg_alloc.rs 228:25]
                            connect rename_table_normal_0_dest0.data, _match_arm_value_4 @[reg_alloc.rs 229:29]
                            ; connect different types:
                            ; lhs: UInt<8>
                            ; rhs: UInt<33>
                            connect rename_table_normal_0_dest0.addr, sub(dest_reg.normal_regs[0].value, UInt<32>(0h1)) @[reg_alloc.rs 233:33]
                            connect rename_table_normal_0_dest0.en, UInt<1>(0h1) @[reg_alloc.rs 235:29]
                        when and(geq(dest_reg.normal_regs[0].value, UInt<32>(0hFE)), lt(dest_reg.normal_regs[0].value, UInt<32>(0h100))): @[reg_alloc.rs 228:25]
                            connect rename_table_special_0_dest0.data, _match_arm_value_4 @[reg_alloc.rs 229:29]
                            ; connect different types:
                            ; lhs: UInt<1>
                            ; rhs: UInt<33>
                            connect rename_table_special_0_dest0.addr, sub(dest_reg.normal_regs[0].value, UInt<32>(0hFE)) @[reg_alloc.rs 233:33]
                            connect rename_table_special_0_dest0.en, UInt<1>(0h1) @[reg_alloc.rs 235:29]
                        when and(geq(dest_reg.normal_regs[1].value, UInt<32>(0h1)), lt(dest_reg.normal_regs[1].value, UInt<32>(0hFE))): @[reg_alloc.rs 228:25]
                            connect rename_table_normal_0_dest1.data, _match_arm_value_4 @[reg_alloc.rs 229:29]
                            ; connect different types:
                            ; lhs: UInt<8>
                            ; rhs: UInt<33>
                            connect rename_table_normal_0_dest1.addr, sub(dest_reg.normal_regs[1].value, UInt<32>(0h1)) @[reg_alloc.rs 233:33]
                            connect rename_table_normal_0_dest1.en, UInt<1>(0h1) @[reg_alloc.rs 235:29]
                        when and(geq(dest_reg.normal_regs[1].value, UInt<32>(0hFE)), lt(dest_reg.normal_regs[1].value, UInt<32>(0h100))): @[reg_alloc.rs 228:25]
                            connect rename_table_special_0_dest1.data, _match_arm_value_4 @[reg_alloc.rs 229:29]
                            ; connect different types:
                            ; lhs: UInt<1>
                            ; rhs: UInt<33>
                            connect rename_table_special_0_dest1.addr, sub(dest_reg.normal_regs[1].value, UInt<32>(0hFE)) @[reg_alloc.rs 233:33]
                            connect rename_table_special_0_dest1.en, UInt<1>(0h1) @[reg_alloc.rs 235:29]
                        when and(geq(flag_reg.value, UInt<32>(0hFE)), lt(flag_reg.value, UInt<32>(0h100))): @[reg_alloc.rs 228:25]
                            connect rename_table_special_0_flag0_rFE.data, _match_arm_value_4 @[reg_alloc.rs 229:29]
                            ; connect different types:
                            ; lhs: UInt<1>
                            ; rhs: UInt<32>
                            connect rename_table_special_0_flag0_rFE.addr, UInt<32>(0h0) @[reg_alloc.rs 231:33]
                            connect rename_table_special_0_flag0_rFE.en, UInt<1>(0h1) @[reg_alloc.rs 235:29]
                        when and(geq(flag_reg_1.value, UInt<32>(0hFE)), lt(flag_reg_1.value, UInt<32>(0h100))): @[reg_alloc.rs 228:25]
                            connect rename_table_special_0_flag1_rFF.data, _match_arm_value_4 @[reg_alloc.rs 229:29]
                            ; connect different types:
                            ; lhs: UInt<1>
                            ; rhs: UInt<32>
                            connect rename_table_special_0_flag1_rFF.addr, UInt<32>(0h1) @[reg_alloc.rs 231:33]
                            connect rename_table_special_0_flag1_rFF.en, UInt<1>(0h1) @[reg_alloc.rs 235:29]
        wire selected_unit_index_leaf_0_0: Ty41 @[reg_alloc.rs 249:25]
        connect selected_unit_index_leaf_0_0, {|HdlNone, HdlSome: UInt<2>|}(HdlNone) @[reg_alloc.rs 252:21]
        wire unit_index_0_0: UInt<2> @[reg_alloc.rs 258:25]
        ; connect different types:
        ; lhs: UInt<2>
        ; rhs: UInt<64>
        connect unit_index_0_0, UInt<64>(0h0) @[reg_alloc.rs 261:21]
        when available_units[0][0]: @[reg_alloc.rs 263:21]
            connect selected_unit_index_leaf_0_0, {|HdlNone, HdlSome: UInt<2>|}(HdlSome, unit_index_0_0) @[reg_alloc.rs 264:25]
        wire selected_unit_index_leaf_0_1: Ty41 @[reg_alloc.rs 249:25]
        connect selected_unit_index_leaf_0_1, {|HdlNone, HdlSome: UInt<2>|}(HdlNone) @[reg_alloc.rs 252:21]
        wire unit_index_0_1: UInt<2> @[reg_alloc.rs 258:25]
        ; connect different types:
        ; lhs: UInt<2>
        ; rhs: UInt<64>
        connect unit_index_0_1, UInt<64>(0h1) @[reg_alloc.rs 261:21]
        when available_units[0][1]: @[reg_alloc.rs 263:21]
            connect selected_unit_index_leaf_0_1, {|HdlNone, HdlSome: UInt<2>|}(HdlSome, unit_index_0_1) @[reg_alloc.rs 264:25]
        wire selected_unit_index_node_0_0: Ty41 @[reg_alloc.rs 271:25]
        connect selected_unit_index_node_0_0, selected_unit_index_leaf_0_0 @[reg_alloc.rs 275:21]
        match selected_unit_index_leaf_0_0: @[reg_alloc.rs 277:21]
            HdlNone:
                connect selected_unit_index_node_0_0, selected_unit_index_leaf_0_1 @[reg_alloc.rs 278:25]
            HdlSome(_match_arm_value_27):
                skip
        connect selected_unit_indexes[0], selected_unit_index_node_0_0 @[reg_alloc.rs 241:9]
        connect fetch_decode_interface.decoded_insns[1].ready, UInt<1>(0h1) @[reg_alloc.rs 92:9]
        wire _array_literal_expr_37: UInt<1>[2]
        connect _array_literal_expr_37[0], UInt<1>(0h0)
        connect _array_literal_expr_37[1], UInt<1>(0h0)
        connect available_units[1], _array_literal_expr_37 @[reg_alloc.rs 96:9]
        connect renamed_mops[1], {|HdlNone, HdlSome: Ty51|}(HdlNone) @[reg_alloc.rs 100:9]
        wire rename_1_src_0: Ty54 @[reg_alloc.rs 113:17]
        wire _bundle_literal_expr_91: Ty1
        connect _bundle_literal_expr_91.value, tail(UInt<32>(0h0), 24)
        connect rename_1_src_0.addr, _bundle_literal_expr_91 @[reg_alloc.rs 116:13]
        wire _bundle_literal_expr_92: Ty25
        wire _bundle_literal_expr_93: Ty23
        connect _bundle_literal_expr_93.adj_value, tail(UInt<64>(0h0), 62)
        connect _bundle_literal_expr_92.unit_num, _bundle_literal_expr_93
        wire _bundle_literal_expr_94: Ty24
        connect _bundle_literal_expr_94.value, tail(UInt<8>(0h0), 4)
        connect _bundle_literal_expr_92.unit_out_reg, _bundle_literal_expr_94
        connect rename_1_src_0.data, _bundle_literal_expr_92 @[reg_alloc.rs 117:13]
        connect rename_table_normal_mem_r5.clk, cd.clk @[reg_alloc.rs 120:17]
        connect rename_table_normal_mem_r5.addr, UInt<8>(0h0) @[reg_alloc.rs 121:17]
        connect rename_table_normal_mem_r5.en, UInt<1>(0h0) @[reg_alloc.rs 122:17]
        when and(geq(rename_1_src_0.addr.value, UInt<32>(0h1)), lt(rename_1_src_0.addr.value, UInt<32>(0hFE))): @[reg_alloc.rs 125:17]
            ; connect different types:
            ; lhs: UInt<8>
            ; rhs: UInt<33>
            connect rename_table_normal_mem_r5.addr, sub(rename_1_src_0.addr.value, UInt<32>(0h1)) @[reg_alloc.rs 126:21]
            connect rename_table_normal_mem_r5.en, UInt<1>(0h1) @[reg_alloc.rs 127:21]
            connect rename_1_src_0.data, rename_table_normal_mem_r5.data @[reg_alloc.rs 128:21]
            match fetch_decode_interface.decoded_insns[0].data: @[reg_alloc.rs 131:25]
                HdlNone:
                    skip
                HdlSome(_match_arm_value_28):
                    match renamed_mops_out_reg[0]: @[reg_alloc.rs 135:29]
                        HdlNone:
                            skip
                        HdlSome(_match_arm_value_29):
                            wire dest_reg_4: Ty4 @[unit.rs 129:1]
                            match _match_arm_value_28.mop: @[unit.rs 129:1]
                                AluBranch(_match_arm_value_30):
                                    wire dest_reg_5: Ty4 @[instruction.rs 502:1]
                                    match _match_arm_value_30: @[instruction.rs 502:1]
                                        AddSub(_match_arm_value_31):
                                            connect dest_reg_5, _match_arm_value_31.alu_common.common.dest @[instruction.rs 502:1]
                                        AddSubI(_match_arm_value_32):
                                            connect dest_reg_5, _match_arm_value_32.alu_common.common.dest @[instruction.rs 502:1]
                                        Logical(_match_arm_value_33):
                                            connect dest_reg_5, _match_arm_value_33.alu_common.common.dest @[instruction.rs 502:1]
                                    connect dest_reg_4, dest_reg_5 @[unit.rs 129:1]
                                L2RegisterFile(_match_arm_value_34):
                                    wire dest_reg_6: Ty4 @[instruction.rs 529:1]
                                    match _match_arm_value_34: @[instruction.rs 529:1]
                                        ReadL2Reg(_match_arm_value_35):
                                            connect dest_reg_6, _match_arm_value_35.common.dest @[instruction.rs 529:1]
                                        WriteL2Reg(_match_arm_value_36):
                                            connect dest_reg_6, _match_arm_value_36.common.dest @[instruction.rs 529:1]
                                    connect dest_reg_4, dest_reg_6 @[unit.rs 129:1]
                                LoadStore(_match_arm_value_37):
                                    wire dest_reg_7: Ty4 @[instruction.rs 564:1]
                                    match _match_arm_value_37: @[instruction.rs 564:1]
                                        Load(_match_arm_value_38):
                                            connect dest_reg_7, _match_arm_value_38.dest @[instruction.rs 564:1]
                                        Store(_match_arm_value_39):
                                            connect dest_reg_7, _match_arm_value_39.dest @[instruction.rs 564:1]
                                    connect dest_reg_4, dest_reg_7 @[unit.rs 129:1]
                            wire flag_reg_2: Ty1 @[instruction.rs 831:32]
                            wire _bundle_literal_expr_95: Ty1
                            connect _bundle_literal_expr_95.value, tail(UInt<32>(0h0), 24)
                            connect flag_reg_2, _bundle_literal_expr_95 @[instruction.rs 832:17]
                            match dest_reg_4.flag_regs[0]: @[instruction.rs 834:17]
                                HdlNone:
                                    skip
                                HdlSome(_match_arm_value_40):
                                    wire _bundle_literal_expr_96: Ty1
                                    connect _bundle_literal_expr_96.value, tail(UInt<32>(0hFE), 24)
                                    connect flag_reg_2, _bundle_literal_expr_96 @[instruction.rs 836:21]
                            wire flag_reg_3: Ty1 @[instruction.rs 831:32]
                            wire _bundle_literal_expr_97: Ty1
                            connect _bundle_literal_expr_97.value, tail(UInt<32>(0h0), 24)
                            connect flag_reg_3, _bundle_literal_expr_97 @[instruction.rs 832:17]
                            match dest_reg_4.flag_regs[1]: @[instruction.rs 834:17]
                                HdlNone:
                                    skip
                                HdlSome(_match_arm_value_41):
                                    wire _bundle_literal_expr_98: Ty1
                                    connect _bundle_literal_expr_98.value, tail(UInt<32>(0hFF), 24)
                                    connect flag_reg_3, _bundle_literal_expr_98 @[instruction.rs 836:21]
                            when eq(dest_reg_4.normal_regs[0].value, rename_1_src_0.addr.value): @[reg_alloc.rs 145:41]
                                connect rename_1_src_0.data, _match_arm_value_29 @[reg_alloc.rs 146:45]
                            when eq(dest_reg_4.normal_regs[1].value, rename_1_src_0.addr.value): @[reg_alloc.rs 145:41]
                                connect rename_1_src_0.data, _match_arm_value_29 @[reg_alloc.rs 146:45]
        connect rename_table_special_mem_r7.clk, cd.clk @[reg_alloc.rs 120:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<8>
        connect rename_table_special_mem_r7.addr, UInt<8>(0h0) @[reg_alloc.rs 121:17]
        connect rename_table_special_mem_r7.en, UInt<1>(0h0) @[reg_alloc.rs 122:17]
        when and(geq(rename_1_src_0.addr.value, UInt<32>(0hFE)), lt(rename_1_src_0.addr.value, UInt<32>(0h100))): @[reg_alloc.rs 125:17]
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<33>
            connect rename_table_special_mem_r7.addr, sub(rename_1_src_0.addr.value, UInt<32>(0hFE)) @[reg_alloc.rs 126:21]
            connect rename_table_special_mem_r7.en, UInt<1>(0h1) @[reg_alloc.rs 127:21]
            connect rename_1_src_0.data, rename_table_special_mem_r7.data @[reg_alloc.rs 128:21]
            match fetch_decode_interface.decoded_insns[0].data: @[reg_alloc.rs 131:25]
                HdlNone:
                    skip
                HdlSome(_match_arm_value_42):
                    match renamed_mops_out_reg[0]: @[reg_alloc.rs 135:29]
                        HdlNone:
                            skip
                        HdlSome(_match_arm_value_43):
                            wire dest_reg_8: Ty4 @[unit.rs 129:1]
                            match _match_arm_value_42.mop: @[unit.rs 129:1]
                                AluBranch(_match_arm_value_44):
                                    wire dest_reg_9: Ty4 @[instruction.rs 502:1]
                                    match _match_arm_value_44: @[instruction.rs 502:1]
                                        AddSub(_match_arm_value_45):
                                            connect dest_reg_9, _match_arm_value_45.alu_common.common.dest @[instruction.rs 502:1]
                                        AddSubI(_match_arm_value_46):
                                            connect dest_reg_9, _match_arm_value_46.alu_common.common.dest @[instruction.rs 502:1]
                                        Logical(_match_arm_value_47):
                                            connect dest_reg_9, _match_arm_value_47.alu_common.common.dest @[instruction.rs 502:1]
                                    connect dest_reg_8, dest_reg_9 @[unit.rs 129:1]
                                L2RegisterFile(_match_arm_value_48):
                                    wire dest_reg_10: Ty4 @[instruction.rs 529:1]
                                    match _match_arm_value_48: @[instruction.rs 529:1]
                                        ReadL2Reg(_match_arm_value_49):
                                            connect dest_reg_10, _match_arm_value_49.common.dest @[instruction.rs 529:1]
                                        WriteL2Reg(_match_arm_value_50):
                                            connect dest_reg_10, _match_arm_value_50.common.dest @[instruction.rs 529:1]
                                    connect dest_reg_8, dest_reg_10 @[unit.rs 129:1]
                                LoadStore(_match_arm_value_51):
                                    wire dest_reg_11: Ty4 @[instruction.rs 564:1]
                                    match _match_arm_value_51: @[instruction.rs 564:1]
                                        Load(_match_arm_value_52):
                                            connect dest_reg_11, _match_arm_value_52.dest @[instruction.rs 564:1]
                                        Store(_match_arm_value_53):
                                            connect dest_reg_11, _match_arm_value_53.dest @[instruction.rs 564:1]
                                    connect dest_reg_8, dest_reg_11 @[unit.rs 129:1]
                            wire flag_reg_4: Ty1 @[instruction.rs 831:32]
                            wire _bundle_literal_expr_99: Ty1
                            connect _bundle_literal_expr_99.value, tail(UInt<32>(0h0), 24)
                            connect flag_reg_4, _bundle_literal_expr_99 @[instruction.rs 832:17]
                            match dest_reg_8.flag_regs[0]: @[instruction.rs 834:17]
                                HdlNone:
                                    skip
                                HdlSome(_match_arm_value_54):
                                    wire _bundle_literal_expr_100: Ty1
                                    connect _bundle_literal_expr_100.value, tail(UInt<32>(0hFE), 24)
                                    connect flag_reg_4, _bundle_literal_expr_100 @[instruction.rs 836:21]
                            wire flag_reg_5: Ty1 @[instruction.rs 831:32]
                            wire _bundle_literal_expr_101: Ty1
                            connect _bundle_literal_expr_101.value, tail(UInt<32>(0h0), 24)
                            connect flag_reg_5, _bundle_literal_expr_101 @[instruction.rs 832:17]
                            match dest_reg_8.flag_regs[1]: @[instruction.rs 834:17]
                                HdlNone:
                                    skip
                                HdlSome(_match_arm_value_55):
                                    wire _bundle_literal_expr_102: Ty1
                                    connect _bundle_literal_expr_102.value, tail(UInt<32>(0hFF), 24)
                                    connect flag_reg_5, _bundle_literal_expr_102 @[instruction.rs 836:21]
                            when eq(dest_reg_8.normal_regs[0].value, rename_1_src_0.addr.value): @[reg_alloc.rs 145:41]
                                connect rename_1_src_0.data, _match_arm_value_43 @[reg_alloc.rs 146:45]
                            when eq(dest_reg_8.normal_regs[1].value, rename_1_src_0.addr.value): @[reg_alloc.rs 145:41]
                                connect rename_1_src_0.data, _match_arm_value_43 @[reg_alloc.rs 146:45]
                            when eq(flag_reg_4.value, rename_1_src_0.addr.value): @[reg_alloc.rs 145:41]
                                connect rename_1_src_0.data, _match_arm_value_43 @[reg_alloc.rs 146:45]
                            when eq(flag_reg_5.value, rename_1_src_0.addr.value): @[reg_alloc.rs 145:41]
                                connect rename_1_src_0.data, _match_arm_value_43 @[reg_alloc.rs 146:45]
        wire rename_1_src_1: Ty54 @[reg_alloc.rs 113:17]
        wire _bundle_literal_expr_103: Ty1
        connect _bundle_literal_expr_103.value, tail(UInt<32>(0h0), 24)
        connect rename_1_src_1.addr, _bundle_literal_expr_103 @[reg_alloc.rs 116:13]
        wire _bundle_literal_expr_104: Ty25
        wire _bundle_literal_expr_105: Ty23
        connect _bundle_literal_expr_105.adj_value, tail(UInt<64>(0h0), 62)
        connect _bundle_literal_expr_104.unit_num, _bundle_literal_expr_105
        wire _bundle_literal_expr_106: Ty24
        connect _bundle_literal_expr_106.value, tail(UInt<8>(0h0), 4)
        connect _bundle_literal_expr_104.unit_out_reg, _bundle_literal_expr_106
        connect rename_1_src_1.data, _bundle_literal_expr_104 @[reg_alloc.rs 117:13]
        connect rename_table_normal_mem_r6.clk, cd.clk @[reg_alloc.rs 120:17]
        connect rename_table_normal_mem_r6.addr, UInt<8>(0h0) @[reg_alloc.rs 121:17]
        connect rename_table_normal_mem_r6.en, UInt<1>(0h0) @[reg_alloc.rs 122:17]
        when and(geq(rename_1_src_1.addr.value, UInt<32>(0h1)), lt(rename_1_src_1.addr.value, UInt<32>(0hFE))): @[reg_alloc.rs 125:17]
            ; connect different types:
            ; lhs: UInt<8>
            ; rhs: UInt<33>
            connect rename_table_normal_mem_r6.addr, sub(rename_1_src_1.addr.value, UInt<32>(0h1)) @[reg_alloc.rs 126:21]
            connect rename_table_normal_mem_r6.en, UInt<1>(0h1) @[reg_alloc.rs 127:21]
            connect rename_1_src_1.data, rename_table_normal_mem_r6.data @[reg_alloc.rs 128:21]
            match fetch_decode_interface.decoded_insns[0].data: @[reg_alloc.rs 131:25]
                HdlNone:
                    skip
                HdlSome(_match_arm_value_56):
                    match renamed_mops_out_reg[0]: @[reg_alloc.rs 135:29]
                        HdlNone:
                            skip
                        HdlSome(_match_arm_value_57):
                            wire dest_reg_12: Ty4 @[unit.rs 129:1]
                            match _match_arm_value_56.mop: @[unit.rs 129:1]
                                AluBranch(_match_arm_value_58):
                                    wire dest_reg_13: Ty4 @[instruction.rs 502:1]
                                    match _match_arm_value_58: @[instruction.rs 502:1]
                                        AddSub(_match_arm_value_59):
                                            connect dest_reg_13, _match_arm_value_59.alu_common.common.dest @[instruction.rs 502:1]
                                        AddSubI(_match_arm_value_60):
                                            connect dest_reg_13, _match_arm_value_60.alu_common.common.dest @[instruction.rs 502:1]
                                        Logical(_match_arm_value_61):
                                            connect dest_reg_13, _match_arm_value_61.alu_common.common.dest @[instruction.rs 502:1]
                                    connect dest_reg_12, dest_reg_13 @[unit.rs 129:1]
                                L2RegisterFile(_match_arm_value_62):
                                    wire dest_reg_14: Ty4 @[instruction.rs 529:1]
                                    match _match_arm_value_62: @[instruction.rs 529:1]
                                        ReadL2Reg(_match_arm_value_63):
                                            connect dest_reg_14, _match_arm_value_63.common.dest @[instruction.rs 529:1]
                                        WriteL2Reg(_match_arm_value_64):
                                            connect dest_reg_14, _match_arm_value_64.common.dest @[instruction.rs 529:1]
                                    connect dest_reg_12, dest_reg_14 @[unit.rs 129:1]
                                LoadStore(_match_arm_value_65):
                                    wire dest_reg_15: Ty4 @[instruction.rs 564:1]
                                    match _match_arm_value_65: @[instruction.rs 564:1]
                                        Load(_match_arm_value_66):
                                            connect dest_reg_15, _match_arm_value_66.dest @[instruction.rs 564:1]
                                        Store(_match_arm_value_67):
                                            connect dest_reg_15, _match_arm_value_67.dest @[instruction.rs 564:1]
                                    connect dest_reg_12, dest_reg_15 @[unit.rs 129:1]
                            wire flag_reg_6: Ty1 @[instruction.rs 831:32]
                            wire _bundle_literal_expr_107: Ty1
                            connect _bundle_literal_expr_107.value, tail(UInt<32>(0h0), 24)
                            connect flag_reg_6, _bundle_literal_expr_107 @[instruction.rs 832:17]
                            match dest_reg_12.flag_regs[0]: @[instruction.rs 834:17]
                                HdlNone:
                                    skip
                                HdlSome(_match_arm_value_68):
                                    wire _bundle_literal_expr_108: Ty1
                                    connect _bundle_literal_expr_108.value, tail(UInt<32>(0hFE), 24)
                                    connect flag_reg_6, _bundle_literal_expr_108 @[instruction.rs 836:21]
                            wire flag_reg_7: Ty1 @[instruction.rs 831:32]
                            wire _bundle_literal_expr_109: Ty1
                            connect _bundle_literal_expr_109.value, tail(UInt<32>(0h0), 24)
                            connect flag_reg_7, _bundle_literal_expr_109 @[instruction.rs 832:17]
                            match dest_reg_12.flag_regs[1]: @[instruction.rs 834:17]
                                HdlNone:
                                    skip
                                HdlSome(_match_arm_value_69):
                                    wire _bundle_literal_expr_110: Ty1
                                    connect _bundle_literal_expr_110.value, tail(UInt<32>(0hFF), 24)
                                    connect flag_reg_7, _bundle_literal_expr_110 @[instruction.rs 836:21]
                            when eq(dest_reg_12.normal_regs[0].value, rename_1_src_1.addr.value): @[reg_alloc.rs 145:41]
                                connect rename_1_src_1.data, _match_arm_value_57 @[reg_alloc.rs 146:45]
                            when eq(dest_reg_12.normal_regs[1].value, rename_1_src_1.addr.value): @[reg_alloc.rs 145:41]
                                connect rename_1_src_1.data, _match_arm_value_57 @[reg_alloc.rs 146:45]
        connect rename_table_special_mem_r8.clk, cd.clk @[reg_alloc.rs 120:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<8>
        connect rename_table_special_mem_r8.addr, UInt<8>(0h0) @[reg_alloc.rs 121:17]
        connect rename_table_special_mem_r8.en, UInt<1>(0h0) @[reg_alloc.rs 122:17]
        when and(geq(rename_1_src_1.addr.value, UInt<32>(0hFE)), lt(rename_1_src_1.addr.value, UInt<32>(0h100))): @[reg_alloc.rs 125:17]
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<33>
            connect rename_table_special_mem_r8.addr, sub(rename_1_src_1.addr.value, UInt<32>(0hFE)) @[reg_alloc.rs 126:21]
            connect rename_table_special_mem_r8.en, UInt<1>(0h1) @[reg_alloc.rs 127:21]
            connect rename_1_src_1.data, rename_table_special_mem_r8.data @[reg_alloc.rs 128:21]
            match fetch_decode_interface.decoded_insns[0].data: @[reg_alloc.rs 131:25]
                HdlNone:
                    skip
                HdlSome(_match_arm_value_70):
                    match renamed_mops_out_reg[0]: @[reg_alloc.rs 135:29]
                        HdlNone:
                            skip
                        HdlSome(_match_arm_value_71):
                            wire dest_reg_16: Ty4 @[unit.rs 129:1]
                            match _match_arm_value_70.mop: @[unit.rs 129:1]
                                AluBranch(_match_arm_value_72):
                                    wire dest_reg_17: Ty4 @[instruction.rs 502:1]
                                    match _match_arm_value_72: @[instruction.rs 502:1]
                                        AddSub(_match_arm_value_73):
                                            connect dest_reg_17, _match_arm_value_73.alu_common.common.dest @[instruction.rs 502:1]
                                        AddSubI(_match_arm_value_74):
                                            connect dest_reg_17, _match_arm_value_74.alu_common.common.dest @[instruction.rs 502:1]
                                        Logical(_match_arm_value_75):
                                            connect dest_reg_17, _match_arm_value_75.alu_common.common.dest @[instruction.rs 502:1]
                                    connect dest_reg_16, dest_reg_17 @[unit.rs 129:1]
                                L2RegisterFile(_match_arm_value_76):
                                    wire dest_reg_18: Ty4 @[instruction.rs 529:1]
                                    match _match_arm_value_76: @[instruction.rs 529:1]
                                        ReadL2Reg(_match_arm_value_77):
                                            connect dest_reg_18, _match_arm_value_77.common.dest @[instruction.rs 529:1]
                                        WriteL2Reg(_match_arm_value_78):
                                            connect dest_reg_18, _match_arm_value_78.common.dest @[instruction.rs 529:1]
                                    connect dest_reg_16, dest_reg_18 @[unit.rs 129:1]
                                LoadStore(_match_arm_value_79):
                                    wire dest_reg_19: Ty4 @[instruction.rs 564:1]
                                    match _match_arm_value_79: @[instruction.rs 564:1]
                                        Load(_match_arm_value_80):
                                            connect dest_reg_19, _match_arm_value_80.dest @[instruction.rs 564:1]
                                        Store(_match_arm_value_81):
                                            connect dest_reg_19, _match_arm_value_81.dest @[instruction.rs 564:1]
                                    connect dest_reg_16, dest_reg_19 @[unit.rs 129:1]
                            wire flag_reg_8: Ty1 @[instruction.rs 831:32]
                            wire _bundle_literal_expr_111: Ty1
                            connect _bundle_literal_expr_111.value, tail(UInt<32>(0h0), 24)
                            connect flag_reg_8, _bundle_literal_expr_111 @[instruction.rs 832:17]
                            match dest_reg_16.flag_regs[0]: @[instruction.rs 834:17]
                                HdlNone:
                                    skip
                                HdlSome(_match_arm_value_82):
                                    wire _bundle_literal_expr_112: Ty1
                                    connect _bundle_literal_expr_112.value, tail(UInt<32>(0hFE), 24)
                                    connect flag_reg_8, _bundle_literal_expr_112 @[instruction.rs 836:21]
                            wire flag_reg_9: Ty1 @[instruction.rs 831:32]
                            wire _bundle_literal_expr_113: Ty1
                            connect _bundle_literal_expr_113.value, tail(UInt<32>(0h0), 24)
                            connect flag_reg_9, _bundle_literal_expr_113 @[instruction.rs 832:17]
                            match dest_reg_16.flag_regs[1]: @[instruction.rs 834:17]
                                HdlNone:
                                    skip
                                HdlSome(_match_arm_value_83):
                                    wire _bundle_literal_expr_114: Ty1
                                    connect _bundle_literal_expr_114.value, tail(UInt<32>(0hFF), 24)
                                    connect flag_reg_9, _bundle_literal_expr_114 @[instruction.rs 836:21]
                            when eq(dest_reg_16.normal_regs[0].value, rename_1_src_1.addr.value): @[reg_alloc.rs 145:41]
                                connect rename_1_src_1.data, _match_arm_value_71 @[reg_alloc.rs 146:45]
                            when eq(dest_reg_16.normal_regs[1].value, rename_1_src_1.addr.value): @[reg_alloc.rs 145:41]
                                connect rename_1_src_1.data, _match_arm_value_71 @[reg_alloc.rs 146:45]
                            when eq(flag_reg_8.value, rename_1_src_1.addr.value): @[reg_alloc.rs 145:41]
                                connect rename_1_src_1.data, _match_arm_value_71 @[reg_alloc.rs 146:45]
                            when eq(flag_reg_9.value, rename_1_src_1.addr.value): @[reg_alloc.rs 145:41]
                                connect rename_1_src_1.data, _match_arm_value_71 @[reg_alloc.rs 146:45]
        wire rename_1_src_2: Ty54 @[reg_alloc.rs 113:17]
        wire _bundle_literal_expr_115: Ty1
        connect _bundle_literal_expr_115.value, tail(UInt<32>(0h0), 24)
        connect rename_1_src_2.addr, _bundle_literal_expr_115 @[reg_alloc.rs 116:13]
        wire _bundle_literal_expr_116: Ty25
        wire _bundle_literal_expr_117: Ty23
        connect _bundle_literal_expr_117.adj_value, tail(UInt<64>(0h0), 62)
        connect _bundle_literal_expr_116.unit_num, _bundle_literal_expr_117
        wire _bundle_literal_expr_118: Ty24
        connect _bundle_literal_expr_118.value, tail(UInt<8>(0h0), 4)
        connect _bundle_literal_expr_116.unit_out_reg, _bundle_literal_expr_118
        connect rename_1_src_2.data, _bundle_literal_expr_116 @[reg_alloc.rs 117:13]
        connect rename_table_normal_mem_r7.clk, cd.clk @[reg_alloc.rs 120:17]
        connect rename_table_normal_mem_r7.addr, UInt<8>(0h0) @[reg_alloc.rs 121:17]
        connect rename_table_normal_mem_r7.en, UInt<1>(0h0) @[reg_alloc.rs 122:17]
        when and(geq(rename_1_src_2.addr.value, UInt<32>(0h1)), lt(rename_1_src_2.addr.value, UInt<32>(0hFE))): @[reg_alloc.rs 125:17]
            ; connect different types:
            ; lhs: UInt<8>
            ; rhs: UInt<33>
            connect rename_table_normal_mem_r7.addr, sub(rename_1_src_2.addr.value, UInt<32>(0h1)) @[reg_alloc.rs 126:21]
            connect rename_table_normal_mem_r7.en, UInt<1>(0h1) @[reg_alloc.rs 127:21]
            connect rename_1_src_2.data, rename_table_normal_mem_r7.data @[reg_alloc.rs 128:21]
            match fetch_decode_interface.decoded_insns[0].data: @[reg_alloc.rs 131:25]
                HdlNone:
                    skip
                HdlSome(_match_arm_value_84):
                    match renamed_mops_out_reg[0]: @[reg_alloc.rs 135:29]
                        HdlNone:
                            skip
                        HdlSome(_match_arm_value_85):
                            wire dest_reg_20: Ty4 @[unit.rs 129:1]
                            match _match_arm_value_84.mop: @[unit.rs 129:1]
                                AluBranch(_match_arm_value_86):
                                    wire dest_reg_21: Ty4 @[instruction.rs 502:1]
                                    match _match_arm_value_86: @[instruction.rs 502:1]
                                        AddSub(_match_arm_value_87):
                                            connect dest_reg_21, _match_arm_value_87.alu_common.common.dest @[instruction.rs 502:1]
                                        AddSubI(_match_arm_value_88):
                                            connect dest_reg_21, _match_arm_value_88.alu_common.common.dest @[instruction.rs 502:1]
                                        Logical(_match_arm_value_89):
                                            connect dest_reg_21, _match_arm_value_89.alu_common.common.dest @[instruction.rs 502:1]
                                    connect dest_reg_20, dest_reg_21 @[unit.rs 129:1]
                                L2RegisterFile(_match_arm_value_90):
                                    wire dest_reg_22: Ty4 @[instruction.rs 529:1]
                                    match _match_arm_value_90: @[instruction.rs 529:1]
                                        ReadL2Reg(_match_arm_value_91):
                                            connect dest_reg_22, _match_arm_value_91.common.dest @[instruction.rs 529:1]
                                        WriteL2Reg(_match_arm_value_92):
                                            connect dest_reg_22, _match_arm_value_92.common.dest @[instruction.rs 529:1]
                                    connect dest_reg_20, dest_reg_22 @[unit.rs 129:1]
                                LoadStore(_match_arm_value_93):
                                    wire dest_reg_23: Ty4 @[instruction.rs 564:1]
                                    match _match_arm_value_93: @[instruction.rs 564:1]
                                        Load(_match_arm_value_94):
                                            connect dest_reg_23, _match_arm_value_94.dest @[instruction.rs 564:1]
                                        Store(_match_arm_value_95):
                                            connect dest_reg_23, _match_arm_value_95.dest @[instruction.rs 564:1]
                                    connect dest_reg_20, dest_reg_23 @[unit.rs 129:1]
                            wire flag_reg_10: Ty1 @[instruction.rs 831:32]
                            wire _bundle_literal_expr_119: Ty1
                            connect _bundle_literal_expr_119.value, tail(UInt<32>(0h0), 24)
                            connect flag_reg_10, _bundle_literal_expr_119 @[instruction.rs 832:17]
                            match dest_reg_20.flag_regs[0]: @[instruction.rs 834:17]
                                HdlNone:
                                    skip
                                HdlSome(_match_arm_value_96):
                                    wire _bundle_literal_expr_120: Ty1
                                    connect _bundle_literal_expr_120.value, tail(UInt<32>(0hFE), 24)
                                    connect flag_reg_10, _bundle_literal_expr_120 @[instruction.rs 836:21]
                            wire flag_reg_11: Ty1 @[instruction.rs 831:32]
                            wire _bundle_literal_expr_121: Ty1
                            connect _bundle_literal_expr_121.value, tail(UInt<32>(0h0), 24)
                            connect flag_reg_11, _bundle_literal_expr_121 @[instruction.rs 832:17]
                            match dest_reg_20.flag_regs[1]: @[instruction.rs 834:17]
                                HdlNone:
                                    skip
                                HdlSome(_match_arm_value_97):
                                    wire _bundle_literal_expr_122: Ty1
                                    connect _bundle_literal_expr_122.value, tail(UInt<32>(0hFF), 24)
                                    connect flag_reg_11, _bundle_literal_expr_122 @[instruction.rs 836:21]
                            when eq(dest_reg_20.normal_regs[0].value, rename_1_src_2.addr.value): @[reg_alloc.rs 145:41]
                                connect rename_1_src_2.data, _match_arm_value_85 @[reg_alloc.rs 146:45]
                            when eq(dest_reg_20.normal_regs[1].value, rename_1_src_2.addr.value): @[reg_alloc.rs 145:41]
                                connect rename_1_src_2.data, _match_arm_value_85 @[reg_alloc.rs 146:45]
        connect rename_table_special_mem_r9.clk, cd.clk @[reg_alloc.rs 120:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<8>
        connect rename_table_special_mem_r9.addr, UInt<8>(0h0) @[reg_alloc.rs 121:17]
        connect rename_table_special_mem_r9.en, UInt<1>(0h0) @[reg_alloc.rs 122:17]
        when and(geq(rename_1_src_2.addr.value, UInt<32>(0hFE)), lt(rename_1_src_2.addr.value, UInt<32>(0h100))): @[reg_alloc.rs 125:17]
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<33>
            connect rename_table_special_mem_r9.addr, sub(rename_1_src_2.addr.value, UInt<32>(0hFE)) @[reg_alloc.rs 126:21]
            connect rename_table_special_mem_r9.en, UInt<1>(0h1) @[reg_alloc.rs 127:21]
            connect rename_1_src_2.data, rename_table_special_mem_r9.data @[reg_alloc.rs 128:21]
            match fetch_decode_interface.decoded_insns[0].data: @[reg_alloc.rs 131:25]
                HdlNone:
                    skip
                HdlSome(_match_arm_value_98):
                    match renamed_mops_out_reg[0]: @[reg_alloc.rs 135:29]
                        HdlNone:
                            skip
                        HdlSome(_match_arm_value_99):
                            wire dest_reg_24: Ty4 @[unit.rs 129:1]
                            match _match_arm_value_98.mop: @[unit.rs 129:1]
                                AluBranch(_match_arm_value_100):
                                    wire dest_reg_25: Ty4 @[instruction.rs 502:1]
                                    match _match_arm_value_100: @[instruction.rs 502:1]
                                        AddSub(_match_arm_value_101):
                                            connect dest_reg_25, _match_arm_value_101.alu_common.common.dest @[instruction.rs 502:1]
                                        AddSubI(_match_arm_value_102):
                                            connect dest_reg_25, _match_arm_value_102.alu_common.common.dest @[instruction.rs 502:1]
                                        Logical(_match_arm_value_103):
                                            connect dest_reg_25, _match_arm_value_103.alu_common.common.dest @[instruction.rs 502:1]
                                    connect dest_reg_24, dest_reg_25 @[unit.rs 129:1]
                                L2RegisterFile(_match_arm_value_104):
                                    wire dest_reg_26: Ty4 @[instruction.rs 529:1]
                                    match _match_arm_value_104: @[instruction.rs 529:1]
                                        ReadL2Reg(_match_arm_value_105):
                                            connect dest_reg_26, _match_arm_value_105.common.dest @[instruction.rs 529:1]
                                        WriteL2Reg(_match_arm_value_106):
                                            connect dest_reg_26, _match_arm_value_106.common.dest @[instruction.rs 529:1]
                                    connect dest_reg_24, dest_reg_26 @[unit.rs 129:1]
                                LoadStore(_match_arm_value_107):
                                    wire dest_reg_27: Ty4 @[instruction.rs 564:1]
                                    match _match_arm_value_107: @[instruction.rs 564:1]
                                        Load(_match_arm_value_108):
                                            connect dest_reg_27, _match_arm_value_108.dest @[instruction.rs 564:1]
                                        Store(_match_arm_value_109):
                                            connect dest_reg_27, _match_arm_value_109.dest @[instruction.rs 564:1]
                                    connect dest_reg_24, dest_reg_27 @[unit.rs 129:1]
                            wire flag_reg_12: Ty1 @[instruction.rs 831:32]
                            wire _bundle_literal_expr_123: Ty1
                            connect _bundle_literal_expr_123.value, tail(UInt<32>(0h0), 24)
                            connect flag_reg_12, _bundle_literal_expr_123 @[instruction.rs 832:17]
                            match dest_reg_24.flag_regs[0]: @[instruction.rs 834:17]
                                HdlNone:
                                    skip
                                HdlSome(_match_arm_value_110):
                                    wire _bundle_literal_expr_124: Ty1
                                    connect _bundle_literal_expr_124.value, tail(UInt<32>(0hFE), 24)
                                    connect flag_reg_12, _bundle_literal_expr_124 @[instruction.rs 836:21]
                            wire flag_reg_13: Ty1 @[instruction.rs 831:32]
                            wire _bundle_literal_expr_125: Ty1
                            connect _bundle_literal_expr_125.value, tail(UInt<32>(0h0), 24)
                            connect flag_reg_13, _bundle_literal_expr_125 @[instruction.rs 832:17]
                            match dest_reg_24.flag_regs[1]: @[instruction.rs 834:17]
                                HdlNone:
                                    skip
                                HdlSome(_match_arm_value_111):
                                    wire _bundle_literal_expr_126: Ty1
                                    connect _bundle_literal_expr_126.value, tail(UInt<32>(0hFF), 24)
                                    connect flag_reg_13, _bundle_literal_expr_126 @[instruction.rs 836:21]
                            when eq(dest_reg_24.normal_regs[0].value, rename_1_src_2.addr.value): @[reg_alloc.rs 145:41]
                                connect rename_1_src_2.data, _match_arm_value_99 @[reg_alloc.rs 146:45]
                            when eq(dest_reg_24.normal_regs[1].value, rename_1_src_2.addr.value): @[reg_alloc.rs 145:41]
                                connect rename_1_src_2.data, _match_arm_value_99 @[reg_alloc.rs 146:45]
                            when eq(flag_reg_12.value, rename_1_src_2.addr.value): @[reg_alloc.rs 145:41]
                                connect rename_1_src_2.data, _match_arm_value_99 @[reg_alloc.rs 146:45]
                            when eq(flag_reg_13.value, rename_1_src_2.addr.value): @[reg_alloc.rs 145:41]
                                connect rename_1_src_2.data, _match_arm_value_99 @[reg_alloc.rs 146:45]
        wire rename_table_normal_1_dest0: Ty30 @[reg_alloc.rs 170:21]
        connect rename_table_normal_mem_w8, rename_table_normal_1_dest0 @[reg_alloc.rs 173:17]
        wire _bundle_literal_expr_127: Ty55
        connect _bundle_literal_expr_127.addr, UInt<0>(0h0)
        connect _bundle_literal_expr_127.en, UInt<1>(0h0)
        connect _bundle_literal_expr_127.clk, cd.clk
        wire _uninit_expr_6: Ty25
        invalidate _uninit_expr_6
        connect _bundle_literal_expr_127.data, _uninit_expr_6
        wire _bundle_literal_expr_128: Ty29
        wire _bundle_literal_expr_129: Ty27
        connect _bundle_literal_expr_129.adj_value, UInt<1>(0h1)
        connect _bundle_literal_expr_128.unit_num, _bundle_literal_expr_129
        wire _bundle_literal_expr_130: Ty28
        connect _bundle_literal_expr_130.value, UInt<1>(0h1)
        connect _bundle_literal_expr_128.unit_out_reg, _bundle_literal_expr_130
        connect _bundle_literal_expr_127.mask, _bundle_literal_expr_128
        ; connect different types:
        ; lhs: Bundle {addr: UInt<8>, en: Bool, clk: Clock, data: Bundle {unit_num: Bundle {adj_value: UInt<2>}, unit_out_reg: Bundle {value: UInt<4>}}, mask: Bundle {unit_num: Bundle {adj_value: Bool}, unit_out_reg: Bundle {value: Bool}}}
        ; rhs: Bundle {addr: UInt<0>, en: Bool, clk: Clock, data: Bundle {unit_num: Bundle {adj_value: UInt<2>}, unit_out_reg: Bundle {value: UInt<4>}}, mask: Bundle {unit_num: Bundle {adj_value: Bool}, unit_out_reg: Bundle {value: Bool}}}
        connect rename_table_normal_1_dest0, _bundle_literal_expr_127 @[reg_alloc.rs 175:17]
        wire rename_table_special_1_dest0: Ty36 @[reg_alloc.rs 170:21]
        connect rename_table_special_mem_w10, rename_table_special_1_dest0 @[reg_alloc.rs 173:17]
        wire _bundle_literal_expr_131: Ty55
        connect _bundle_literal_expr_131.addr, UInt<0>(0h0)
        connect _bundle_literal_expr_131.en, UInt<1>(0h0)
        connect _bundle_literal_expr_131.clk, cd.clk
        wire _uninit_expr_7: Ty25
        invalidate _uninit_expr_7
        connect _bundle_literal_expr_131.data, _uninit_expr_7
        wire _bundle_literal_expr_132: Ty29
        wire _bundle_literal_expr_133: Ty27
        connect _bundle_literal_expr_133.adj_value, UInt<1>(0h1)
        connect _bundle_literal_expr_132.unit_num, _bundle_literal_expr_133
        wire _bundle_literal_expr_134: Ty28
        connect _bundle_literal_expr_134.value, UInt<1>(0h1)
        connect _bundle_literal_expr_132.unit_out_reg, _bundle_literal_expr_134
        connect _bundle_literal_expr_131.mask, _bundle_literal_expr_132
        ; connect different types:
        ; lhs: Bundle {addr: UInt<1>, en: Bool, clk: Clock, data: Bundle {unit_num: Bundle {adj_value: UInt<2>}, unit_out_reg: Bundle {value: UInt<4>}}, mask: Bundle {unit_num: Bundle {adj_value: Bool}, unit_out_reg: Bundle {value: Bool}}}
        ; rhs: Bundle {addr: UInt<0>, en: Bool, clk: Clock, data: Bundle {unit_num: Bundle {adj_value: UInt<2>}, unit_out_reg: Bundle {value: UInt<4>}}, mask: Bundle {unit_num: Bundle {adj_value: Bool}, unit_out_reg: Bundle {value: Bool}}}
        connect rename_table_special_1_dest0, _bundle_literal_expr_131 @[reg_alloc.rs 175:17]
        wire rename_table_normal_1_dest1: Ty30 @[reg_alloc.rs 170:21]
        connect rename_table_normal_mem_w9, rename_table_normal_1_dest1 @[reg_alloc.rs 173:17]
        wire _bundle_literal_expr_135: Ty55
        connect _bundle_literal_expr_135.addr, UInt<0>(0h0)
        connect _bundle_literal_expr_135.en, UInt<1>(0h0)
        connect _bundle_literal_expr_135.clk, cd.clk
        wire _uninit_expr_8: Ty25
        invalidate _uninit_expr_8
        connect _bundle_literal_expr_135.data, _uninit_expr_8
        wire _bundle_literal_expr_136: Ty29
        wire _bundle_literal_expr_137: Ty27
        connect _bundle_literal_expr_137.adj_value, UInt<1>(0h1)
        connect _bundle_literal_expr_136.unit_num, _bundle_literal_expr_137
        wire _bundle_literal_expr_138: Ty28
        connect _bundle_literal_expr_138.value, UInt<1>(0h1)
        connect _bundle_literal_expr_136.unit_out_reg, _bundle_literal_expr_138
        connect _bundle_literal_expr_135.mask, _bundle_literal_expr_136
        ; connect different types:
        ; lhs: Bundle {addr: UInt<8>, en: Bool, clk: Clock, data: Bundle {unit_num: Bundle {adj_value: UInt<2>}, unit_out_reg: Bundle {value: UInt<4>}}, mask: Bundle {unit_num: Bundle {adj_value: Bool}, unit_out_reg: Bundle {value: Bool}}}
        ; rhs: Bundle {addr: UInt<0>, en: Bool, clk: Clock, data: Bundle {unit_num: Bundle {adj_value: UInt<2>}, unit_out_reg: Bundle {value: UInt<4>}}, mask: Bundle {unit_num: Bundle {adj_value: Bool}, unit_out_reg: Bundle {value: Bool}}}
        connect rename_table_normal_1_dest1, _bundle_literal_expr_135 @[reg_alloc.rs 175:17]
        wire rename_table_special_1_dest1: Ty36 @[reg_alloc.rs 170:21]
        connect rename_table_special_mem_w11, rename_table_special_1_dest1 @[reg_alloc.rs 173:17]
        wire _bundle_literal_expr_139: Ty55
        connect _bundle_literal_expr_139.addr, UInt<0>(0h0)
        connect _bundle_literal_expr_139.en, UInt<1>(0h0)
        connect _bundle_literal_expr_139.clk, cd.clk
        wire _uninit_expr_9: Ty25
        invalidate _uninit_expr_9
        connect _bundle_literal_expr_139.data, _uninit_expr_9
        wire _bundle_literal_expr_140: Ty29
        wire _bundle_literal_expr_141: Ty27
        connect _bundle_literal_expr_141.adj_value, UInt<1>(0h1)
        connect _bundle_literal_expr_140.unit_num, _bundle_literal_expr_141
        wire _bundle_literal_expr_142: Ty28
        connect _bundle_literal_expr_142.value, UInt<1>(0h1)
        connect _bundle_literal_expr_140.unit_out_reg, _bundle_literal_expr_142
        connect _bundle_literal_expr_139.mask, _bundle_literal_expr_140
        ; connect different types:
        ; lhs: Bundle {addr: UInt<1>, en: Bool, clk: Clock, data: Bundle {unit_num: Bundle {adj_value: UInt<2>}, unit_out_reg: Bundle {value: UInt<4>}}, mask: Bundle {unit_num: Bundle {adj_value: Bool}, unit_out_reg: Bundle {value: Bool}}}
        ; rhs: Bundle {addr: UInt<0>, en: Bool, clk: Clock, data: Bundle {unit_num: Bundle {adj_value: UInt<2>}, unit_out_reg: Bundle {value: UInt<4>}}, mask: Bundle {unit_num: Bundle {adj_value: Bool}, unit_out_reg: Bundle {value: Bool}}}
        connect rename_table_special_1_dest1, _bundle_literal_expr_139 @[reg_alloc.rs 175:17]
        wire rename_table_special_1_flag0_rFE: Ty36 @[reg_alloc.rs 170:21]
        connect rename_table_special_mem_w12, rename_table_special_1_flag0_rFE @[reg_alloc.rs 173:17]
        wire _bundle_literal_expr_143: Ty55
        connect _bundle_literal_expr_143.addr, UInt<0>(0h0)
        connect _bundle_literal_expr_143.en, UInt<1>(0h0)
        connect _bundle_literal_expr_143.clk, cd.clk
        wire _uninit_expr_10: Ty25
        invalidate _uninit_expr_10
        connect _bundle_literal_expr_143.data, _uninit_expr_10
        wire _bundle_literal_expr_144: Ty29
        wire _bundle_literal_expr_145: Ty27
        connect _bundle_literal_expr_145.adj_value, UInt<1>(0h1)
        connect _bundle_literal_expr_144.unit_num, _bundle_literal_expr_145
        wire _bundle_literal_expr_146: Ty28
        connect _bundle_literal_expr_146.value, UInt<1>(0h1)
        connect _bundle_literal_expr_144.unit_out_reg, _bundle_literal_expr_146
        connect _bundle_literal_expr_143.mask, _bundle_literal_expr_144
        ; connect different types:
        ; lhs: Bundle {addr: UInt<1>, en: Bool, clk: Clock, data: Bundle {unit_num: Bundle {adj_value: UInt<2>}, unit_out_reg: Bundle {value: UInt<4>}}, mask: Bundle {unit_num: Bundle {adj_value: Bool}, unit_out_reg: Bundle {value: Bool}}}
        ; rhs: Bundle {addr: UInt<0>, en: Bool, clk: Clock, data: Bundle {unit_num: Bundle {adj_value: UInt<2>}, unit_out_reg: Bundle {value: UInt<4>}}, mask: Bundle {unit_num: Bundle {adj_value: Bool}, unit_out_reg: Bundle {value: Bool}}}
        connect rename_table_special_1_flag0_rFE, _bundle_literal_expr_143 @[reg_alloc.rs 175:17]
        wire rename_table_special_1_flag1_rFF: Ty36 @[reg_alloc.rs 170:21]
        connect rename_table_special_mem_w13, rename_table_special_1_flag1_rFF @[reg_alloc.rs 173:17]
        wire _bundle_literal_expr_147: Ty55
        connect _bundle_literal_expr_147.addr, UInt<0>(0h0)
        connect _bundle_literal_expr_147.en, UInt<1>(0h0)
        connect _bundle_literal_expr_147.clk, cd.clk
        wire _uninit_expr_11: Ty25
        invalidate _uninit_expr_11
        connect _bundle_literal_expr_147.data, _uninit_expr_11
        wire _bundle_literal_expr_148: Ty29
        wire _bundle_literal_expr_149: Ty27
        connect _bundle_literal_expr_149.adj_value, UInt<1>(0h1)
        connect _bundle_literal_expr_148.unit_num, _bundle_literal_expr_149
        wire _bundle_literal_expr_150: Ty28
        connect _bundle_literal_expr_150.value, UInt<1>(0h1)
        connect _bundle_literal_expr_148.unit_out_reg, _bundle_literal_expr_150
        connect _bundle_literal_expr_147.mask, _bundle_literal_expr_148
        ; connect different types:
        ; lhs: Bundle {addr: UInt<1>, en: Bool, clk: Clock, data: Bundle {unit_num: Bundle {adj_value: UInt<2>}, unit_out_reg: Bundle {value: UInt<4>}}, mask: Bundle {unit_num: Bundle {adj_value: Bool}, unit_out_reg: Bundle {value: Bool}}}
        ; rhs: Bundle {addr: UInt<0>, en: Bool, clk: Clock, data: Bundle {unit_num: Bundle {adj_value: UInt<2>}, unit_out_reg: Bundle {value: UInt<4>}}, mask: Bundle {unit_num: Bundle {adj_value: Bool}, unit_out_reg: Bundle {value: Bool}}}
        connect rename_table_special_1_flag1_rFF, _bundle_literal_expr_147 @[reg_alloc.rs 175:17]
        match fetch_decode_interface.decoded_insns[1].data: @[reg_alloc.rs 189:9]
            HdlNone:
                skip
            HdlSome(_match_arm_value_112):
                wire unit_kind_1: Ty56 @[unit.rs 129:1]
                match _match_arm_value_112.mop: @[unit.rs 129:1]
                    AluBranch(_match_arm_value_113):
                        connect unit_kind_1, {|AluBranch, L2RegisterFile, LoadStore|}(AluBranch) @[unit.rs 129:1]
                    L2RegisterFile(_match_arm_value_114):
                        connect unit_kind_1, {|AluBranch, L2RegisterFile, LoadStore|}(L2RegisterFile) @[unit.rs 129:1]
                    LoadStore(_match_arm_value_115):
                        connect unit_kind_1, {|AluBranch, L2RegisterFile, LoadStore|}(LoadStore) @[unit.rs 129:1]
                wire available_units_for_kind_1: UInt<1>[2] @[unit.rs 129:1]
                match unit_kind_1: @[unit.rs 129:1]
                    AluBranch:
                        connect available_units_for_kind_1[0], UInt<1>(0h1) @[unit.rs 129:1]
                        connect available_units_for_kind_1[1], UInt<1>(0h1) @[unit.rs 129:1]
                    L2RegisterFile:
                        connect available_units_for_kind_1[0], UInt<1>(0h0) @[unit.rs 129:1]
                        connect available_units_for_kind_1[1], UInt<1>(0h0) @[unit.rs 129:1]
                    LoadStore:
                        connect available_units_for_kind_1[0], UInt<1>(0h0) @[unit.rs 129:1]
                        connect available_units_for_kind_1[1], UInt<1>(0h0) @[unit.rs 129:1]
                connect available_units[1], available_units_for_kind_1 @[reg_alloc.rs 190:13]
                match renamed_mops_out_reg[1]: @[reg_alloc.rs 195:13]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_116):
                        wire dest_reg_28: Ty4 @[unit.rs 129:1]
                        match _match_arm_value_112.mop: @[unit.rs 129:1]
                            AluBranch(_match_arm_value_117):
                                wire dest_reg_29: Ty4 @[instruction.rs 502:1]
                                match _match_arm_value_117: @[instruction.rs 502:1]
                                    AddSub(_match_arm_value_118):
                                        connect dest_reg_29, _match_arm_value_118.alu_common.common.dest @[instruction.rs 502:1]
                                    AddSubI(_match_arm_value_119):
                                        connect dest_reg_29, _match_arm_value_119.alu_common.common.dest @[instruction.rs 502:1]
                                    Logical(_match_arm_value_120):
                                        connect dest_reg_29, _match_arm_value_120.alu_common.common.dest @[instruction.rs 502:1]
                                connect dest_reg_28, dest_reg_29 @[unit.rs 129:1]
                            L2RegisterFile(_match_arm_value_121):
                                wire dest_reg_30: Ty4 @[instruction.rs 529:1]
                                match _match_arm_value_121: @[instruction.rs 529:1]
                                    ReadL2Reg(_match_arm_value_122):
                                        connect dest_reg_30, _match_arm_value_122.common.dest @[instruction.rs 529:1]
                                    WriteL2Reg(_match_arm_value_123):
                                        connect dest_reg_30, _match_arm_value_123.common.dest @[instruction.rs 529:1]
                                connect dest_reg_28, dest_reg_30 @[unit.rs 129:1]
                            LoadStore(_match_arm_value_124):
                                wire dest_reg_31: Ty4 @[instruction.rs 564:1]
                                match _match_arm_value_124: @[instruction.rs 564:1]
                                    Load(_match_arm_value_125):
                                        connect dest_reg_31, _match_arm_value_125.dest @[instruction.rs 564:1]
                                    Store(_match_arm_value_126):
                                        connect dest_reg_31, _match_arm_value_126.dest @[instruction.rs 564:1]
                                connect dest_reg_28, dest_reg_31 @[unit.rs 129:1]
                        wire mapped_regs_4: Ty51 @[unit.rs 129:1]
                        match _match_arm_value_112.mop: @[unit.rs 129:1]
                            AluBranch(_match_arm_value_127):
                                wire mapped_regs_5: Ty46 @[instruction.rs 502:1]
                                match _match_arm_value_127: @[instruction.rs 502:1]
                                    AddSub(_match_arm_value_128):
                                        wire _bundle_literal_expr_151: Ty1
                                        connect _bundle_literal_expr_151.value, _match_arm_value_128.alu_common.common.src[0]
                                        connect rename_1_src_0.addr, _bundle_literal_expr_151 @[reg_alloc.rs 204:29]
                                        wire _bundle_literal_expr_152: Ty1
                                        connect _bundle_literal_expr_152.value, _match_arm_value_128.alu_common.common.src[1]
                                        connect rename_1_src_1.addr, _bundle_literal_expr_152 @[reg_alloc.rs 204:29]
                                        wire _bundle_literal_expr_153: Ty1
                                        connect _bundle_literal_expr_153.value, _match_arm_value_128.alu_common.common.src[2]
                                        connect rename_1_src_2.addr, _bundle_literal_expr_153 @[reg_alloc.rs 204:29]
                                        wire _bundle_literal_expr_154: Ty44
                                        wire _bundle_literal_expr_155: Ty43
                                        wire _bundle_literal_expr_156: Ty42
                                        connect _bundle_literal_expr_156.prefix_pad, _match_arm_value_128.alu_common.common.prefix_pad
                                        connect _bundle_literal_expr_156.dest, _match_arm_value_116.unit_out_reg
                                        wire _array_literal_expr_38: UInt<6>[3]
                                        wire _array_literal_expr_39: UInt<6>[3]
                                        wire _cast_bundle_to_bits_expr_39: Ty57
                                        connect _cast_bundle_to_bits_expr_39.unit_num, rename_1_src_0.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_39.unit_out_reg, rename_1_src_0.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_39: UInt<6>
                                        connect _cast_to_bits_expr_39, cat(_cast_bundle_to_bits_expr_39.unit_out_reg, _cast_bundle_to_bits_expr_39.unit_num)
                                        connect _array_literal_expr_39[0], _cast_to_bits_expr_39
                                        wire _cast_bundle_to_bits_expr_40: Ty57
                                        connect _cast_bundle_to_bits_expr_40.unit_num, rename_1_src_1.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_40.unit_out_reg, rename_1_src_1.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_40: UInt<6>
                                        connect _cast_to_bits_expr_40, cat(_cast_bundle_to_bits_expr_40.unit_out_reg, _cast_bundle_to_bits_expr_40.unit_num)
                                        connect _array_literal_expr_39[1], _cast_to_bits_expr_40
                                        wire _cast_bundle_to_bits_expr_41: Ty57
                                        connect _cast_bundle_to_bits_expr_41.unit_num, rename_1_src_2.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_41.unit_out_reg, rename_1_src_2.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_41: UInt<6>
                                        connect _cast_to_bits_expr_41, cat(_cast_bundle_to_bits_expr_41.unit_out_reg, _cast_bundle_to_bits_expr_41.unit_num)
                                        connect _array_literal_expr_39[2], _cast_to_bits_expr_41
                                        connect _array_literal_expr_38[0], _array_literal_expr_39[0]
                                        wire _array_literal_expr_40: UInt<6>[3]
                                        wire _cast_bundle_to_bits_expr_42: Ty57
                                        connect _cast_bundle_to_bits_expr_42.unit_num, rename_1_src_0.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_42.unit_out_reg, rename_1_src_0.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_42: UInt<6>
                                        connect _cast_to_bits_expr_42, cat(_cast_bundle_to_bits_expr_42.unit_out_reg, _cast_bundle_to_bits_expr_42.unit_num)
                                        connect _array_literal_expr_40[0], _cast_to_bits_expr_42
                                        wire _cast_bundle_to_bits_expr_43: Ty57
                                        connect _cast_bundle_to_bits_expr_43.unit_num, rename_1_src_1.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_43.unit_out_reg, rename_1_src_1.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_43: UInt<6>
                                        connect _cast_to_bits_expr_43, cat(_cast_bundle_to_bits_expr_43.unit_out_reg, _cast_bundle_to_bits_expr_43.unit_num)
                                        connect _array_literal_expr_40[1], _cast_to_bits_expr_43
                                        wire _cast_bundle_to_bits_expr_44: Ty57
                                        connect _cast_bundle_to_bits_expr_44.unit_num, rename_1_src_2.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_44.unit_out_reg, rename_1_src_2.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_44: UInt<6>
                                        connect _cast_to_bits_expr_44, cat(_cast_bundle_to_bits_expr_44.unit_out_reg, _cast_bundle_to_bits_expr_44.unit_num)
                                        connect _array_literal_expr_40[2], _cast_to_bits_expr_44
                                        connect _array_literal_expr_38[1], _array_literal_expr_40[1]
                                        wire _array_literal_expr_41: UInt<6>[3]
                                        wire _cast_bundle_to_bits_expr_45: Ty57
                                        connect _cast_bundle_to_bits_expr_45.unit_num, rename_1_src_0.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_45.unit_out_reg, rename_1_src_0.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_45: UInt<6>
                                        connect _cast_to_bits_expr_45, cat(_cast_bundle_to_bits_expr_45.unit_out_reg, _cast_bundle_to_bits_expr_45.unit_num)
                                        connect _array_literal_expr_41[0], _cast_to_bits_expr_45
                                        wire _cast_bundle_to_bits_expr_46: Ty57
                                        connect _cast_bundle_to_bits_expr_46.unit_num, rename_1_src_1.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_46.unit_out_reg, rename_1_src_1.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_46: UInt<6>
                                        connect _cast_to_bits_expr_46, cat(_cast_bundle_to_bits_expr_46.unit_out_reg, _cast_bundle_to_bits_expr_46.unit_num)
                                        connect _array_literal_expr_41[1], _cast_to_bits_expr_46
                                        wire _cast_bundle_to_bits_expr_47: Ty57
                                        connect _cast_bundle_to_bits_expr_47.unit_num, rename_1_src_2.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_47.unit_out_reg, rename_1_src_2.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_47: UInt<6>
                                        connect _cast_to_bits_expr_47, cat(_cast_bundle_to_bits_expr_47.unit_out_reg, _cast_bundle_to_bits_expr_47.unit_num)
                                        connect _array_literal_expr_41[2], _cast_to_bits_expr_47
                                        connect _array_literal_expr_38[2], _array_literal_expr_41[2]
                                        connect _bundle_literal_expr_156.src, _array_literal_expr_38
                                        wire _bundle_literal_expr_157: Ty58
                                        connect _bundle_literal_expr_157.imm_low, _match_arm_value_128.alu_common.common.imm_low
                                        wire _array_literal_expr_42: UInt<8>[0]
                                        invalidate _array_literal_expr_42
                                        connect _bundle_literal_expr_157.reversed_src, _array_literal_expr_42
                                        connect _bundle_literal_expr_157.imm_sign, _match_arm_value_128.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_48: Ty59
                                        connect _cast_bundle_to_bits_expr_48.imm_low, _bundle_literal_expr_157.imm_low
                                        connect _cast_bundle_to_bits_expr_48.reversed_src, UInt<0>(0)
                                        connect _cast_bundle_to_bits_expr_48.imm_sign, asUInt(_bundle_literal_expr_157.imm_sign)
                                        wire _cast_to_bits_expr_48: UInt<26>
                                        connect _cast_to_bits_expr_48, cat(_cast_bundle_to_bits_expr_48.imm_sign, cat(_cast_bundle_to_bits_expr_48.reversed_src, _cast_bundle_to_bits_expr_48.imm_low))
                                        connect _bundle_literal_expr_156.imm_low, bits(asSInt(_cast_to_bits_expr_48), 24, 0)
                                        wire _bundle_literal_expr_158: Ty58
                                        connect _bundle_literal_expr_158.imm_low, _match_arm_value_128.alu_common.common.imm_low
                                        wire _array_literal_expr_43: UInt<8>[0]
                                        invalidate _array_literal_expr_43
                                        connect _bundle_literal_expr_158.reversed_src, _array_literal_expr_43
                                        connect _bundle_literal_expr_158.imm_sign, _match_arm_value_128.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_49: Ty59
                                        connect _cast_bundle_to_bits_expr_49.imm_low, _bundle_literal_expr_158.imm_low
                                        connect _cast_bundle_to_bits_expr_49.reversed_src, UInt<0>(0)
                                        connect _cast_bundle_to_bits_expr_49.imm_sign, asUInt(_bundle_literal_expr_158.imm_sign)
                                        wire _cast_to_bits_expr_49: UInt<26>
                                        connect _cast_to_bits_expr_49, cat(_cast_bundle_to_bits_expr_49.imm_sign, cat(_cast_bundle_to_bits_expr_49.reversed_src, _cast_bundle_to_bits_expr_49.imm_low))
                                        connect _bundle_literal_expr_156.imm_sign, shr(asSInt(_cast_to_bits_expr_49), 25)
                                        wire _bundle_literal_expr_159: Ty2
                                        invalidate _bundle_literal_expr_159
                                        connect _bundle_literal_expr_156._phantom, _bundle_literal_expr_159
                                        connect _bundle_literal_expr_155.common, _bundle_literal_expr_156
                                        connect _bundle_literal_expr_155.output_integer_mode, _match_arm_value_128.alu_common.output_integer_mode
                                        connect _bundle_literal_expr_154.alu_common, _bundle_literal_expr_155
                                        connect _bundle_literal_expr_154.invert_src0, _match_arm_value_128.invert_src0
                                        connect _bundle_literal_expr_154.invert_carry_in, _match_arm_value_128.invert_carry_in
                                        connect _bundle_literal_expr_154.invert_carry_out, _match_arm_value_128.invert_carry_out
                                        connect _bundle_literal_expr_154.add_pc, _match_arm_value_128.add_pc
                                        connect mapped_regs_5, {|AddSub: Ty44, AddSubI: Ty44, Logical: Ty45|}(AddSub, _bundle_literal_expr_154) @[instruction.rs 502:1]
                                    AddSubI(_match_arm_value_129):
                                        wire _bundle_literal_expr_160: Ty1
                                        connect _bundle_literal_expr_160.value, _match_arm_value_129.alu_common.common.src[0]
                                        connect rename_1_src_0.addr, _bundle_literal_expr_160 @[reg_alloc.rs 204:29]
                                        wire _bundle_literal_expr_161: Ty1
                                        connect _bundle_literal_expr_161.value, _match_arm_value_129.alu_common.common.src[1]
                                        connect rename_1_src_1.addr, _bundle_literal_expr_161 @[reg_alloc.rs 204:29]
                                        wire _bundle_literal_expr_162: Ty44
                                        wire _bundle_literal_expr_163: Ty43
                                        wire _bundle_literal_expr_164: Ty42
                                        connect _bundle_literal_expr_164.prefix_pad, _match_arm_value_129.alu_common.common.prefix_pad
                                        connect _bundle_literal_expr_164.dest, _match_arm_value_116.unit_out_reg
                                        wire _array_literal_expr_44: UInt<6>[3]
                                        wire _array_literal_expr_45: UInt<6>[2]
                                        wire _cast_bundle_to_bits_expr_50: Ty57
                                        connect _cast_bundle_to_bits_expr_50.unit_num, rename_1_src_0.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_50.unit_out_reg, rename_1_src_0.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_50: UInt<6>
                                        connect _cast_to_bits_expr_50, cat(_cast_bundle_to_bits_expr_50.unit_out_reg, _cast_bundle_to_bits_expr_50.unit_num)
                                        connect _array_literal_expr_45[0], _cast_to_bits_expr_50
                                        wire _cast_bundle_to_bits_expr_51: Ty57
                                        connect _cast_bundle_to_bits_expr_51.unit_num, rename_1_src_1.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_51.unit_out_reg, rename_1_src_1.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_51: UInt<6>
                                        connect _cast_to_bits_expr_51, cat(_cast_bundle_to_bits_expr_51.unit_out_reg, _cast_bundle_to_bits_expr_51.unit_num)
                                        connect _array_literal_expr_45[1], _cast_to_bits_expr_51
                                        connect _array_literal_expr_44[0], _array_literal_expr_45[0]
                                        wire _array_literal_expr_46: UInt<6>[2]
                                        wire _cast_bundle_to_bits_expr_52: Ty57
                                        connect _cast_bundle_to_bits_expr_52.unit_num, rename_1_src_0.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_52.unit_out_reg, rename_1_src_0.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_52: UInt<6>
                                        connect _cast_to_bits_expr_52, cat(_cast_bundle_to_bits_expr_52.unit_out_reg, _cast_bundle_to_bits_expr_52.unit_num)
                                        connect _array_literal_expr_46[0], _cast_to_bits_expr_52
                                        wire _cast_bundle_to_bits_expr_53: Ty57
                                        connect _cast_bundle_to_bits_expr_53.unit_num, rename_1_src_1.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_53.unit_out_reg, rename_1_src_1.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_53: UInt<6>
                                        connect _cast_to_bits_expr_53, cat(_cast_bundle_to_bits_expr_53.unit_out_reg, _cast_bundle_to_bits_expr_53.unit_num)
                                        connect _array_literal_expr_46[1], _cast_to_bits_expr_53
                                        connect _array_literal_expr_44[1], _array_literal_expr_46[1]
                                        wire _bundle_literal_expr_165: Ty60
                                        connect _bundle_literal_expr_165.imm_low, _match_arm_value_129.alu_common.common.imm_low
                                        wire _array_literal_expr_47: UInt<8>[1]
                                        connect _array_literal_expr_47[0], _match_arm_value_129.alu_common.common.src[2]
                                        connect _bundle_literal_expr_165.reversed_src, _array_literal_expr_47
                                        connect _bundle_literal_expr_165.imm_sign, _match_arm_value_129.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_54: Ty61
                                        connect _cast_bundle_to_bits_expr_54.imm_low, _bundle_literal_expr_165.imm_low
                                        connect _cast_bundle_to_bits_expr_54.reversed_src, _bundle_literal_expr_165.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_54.imm_sign, asUInt(_bundle_literal_expr_165.imm_sign)
                                        wire _cast_to_bits_expr_54: UInt<34>
                                        connect _cast_to_bits_expr_54, cat(_cast_bundle_to_bits_expr_54.imm_sign, cat(_cast_bundle_to_bits_expr_54.reversed_src, _cast_bundle_to_bits_expr_54.imm_low))
                                        wire _cast_bits_to_bundle_expr_6: Ty60
                                        wire _cast_bits_to_bundle_expr_flattened_6: Ty61
                                        connect _cast_bits_to_bundle_expr_flattened_6.imm_low, bits(asUInt(asSInt(_cast_to_bits_expr_54)), 24, 0)
                                        connect _cast_bits_to_bundle_expr_6.imm_low, _cast_bits_to_bundle_expr_flattened_6.imm_low
                                        connect _cast_bits_to_bundle_expr_flattened_6.reversed_src, bits(asUInt(asSInt(_cast_to_bits_expr_54)), 32, 25)
                                        wire _cast_bits_to_array_expr_6: UInt<8>[1]
                                        wire _cast_bits_to_array_expr_flattened_6: UInt<8>[1]
                                        connect _cast_bits_to_array_expr_flattened_6[0], bits(_cast_bits_to_bundle_expr_flattened_6.reversed_src, 7, 0)
                                        connect _cast_bits_to_array_expr_6[0], _cast_bits_to_array_expr_flattened_6[0]
                                        connect _cast_bits_to_bundle_expr_6.reversed_src, _cast_bits_to_array_expr_6
                                        connect _cast_bits_to_bundle_expr_flattened_6.imm_sign, bits(asUInt(asSInt(_cast_to_bits_expr_54)), 33, 33)
                                        connect _cast_bits_to_bundle_expr_6.imm_sign, asSInt(_cast_bits_to_bundle_expr_flattened_6.imm_sign)
                                        connect _array_literal_expr_44[2], tail(_cast_bits_to_bundle_expr_6.reversed_src[0], 2)
                                        connect _bundle_literal_expr_164.src, _array_literal_expr_44
                                        wire _bundle_literal_expr_166: Ty60
                                        connect _bundle_literal_expr_166.imm_low, _match_arm_value_129.alu_common.common.imm_low
                                        wire _array_literal_expr_48: UInt<8>[1]
                                        connect _array_literal_expr_48[0], _match_arm_value_129.alu_common.common.src[2]
                                        connect _bundle_literal_expr_166.reversed_src, _array_literal_expr_48
                                        connect _bundle_literal_expr_166.imm_sign, _match_arm_value_129.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_55: Ty61
                                        connect _cast_bundle_to_bits_expr_55.imm_low, _bundle_literal_expr_166.imm_low
                                        connect _cast_bundle_to_bits_expr_55.reversed_src, _bundle_literal_expr_166.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_55.imm_sign, asUInt(_bundle_literal_expr_166.imm_sign)
                                        wire _cast_to_bits_expr_55: UInt<34>
                                        connect _cast_to_bits_expr_55, cat(_cast_bundle_to_bits_expr_55.imm_sign, cat(_cast_bundle_to_bits_expr_55.reversed_src, _cast_bundle_to_bits_expr_55.imm_low))
                                        connect _bundle_literal_expr_164.imm_low, bits(asSInt(_cast_to_bits_expr_55), 24, 0)
                                        wire _bundle_literal_expr_167: Ty60
                                        connect _bundle_literal_expr_167.imm_low, _match_arm_value_129.alu_common.common.imm_low
                                        wire _array_literal_expr_49: UInt<8>[1]
                                        connect _array_literal_expr_49[0], _match_arm_value_129.alu_common.common.src[2]
                                        connect _bundle_literal_expr_167.reversed_src, _array_literal_expr_49
                                        connect _bundle_literal_expr_167.imm_sign, _match_arm_value_129.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_56: Ty61
                                        connect _cast_bundle_to_bits_expr_56.imm_low, _bundle_literal_expr_167.imm_low
                                        connect _cast_bundle_to_bits_expr_56.reversed_src, _bundle_literal_expr_167.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_56.imm_sign, asUInt(_bundle_literal_expr_167.imm_sign)
                                        wire _cast_to_bits_expr_56: UInt<34>
                                        connect _cast_to_bits_expr_56, cat(_cast_bundle_to_bits_expr_56.imm_sign, cat(_cast_bundle_to_bits_expr_56.reversed_src, _cast_bundle_to_bits_expr_56.imm_low))
                                        connect _bundle_literal_expr_164.imm_sign, shr(asSInt(_cast_to_bits_expr_56), 33)
                                        wire _bundle_literal_expr_168: Ty2
                                        invalidate _bundle_literal_expr_168
                                        connect _bundle_literal_expr_164._phantom, _bundle_literal_expr_168
                                        connect _bundle_literal_expr_163.common, _bundle_literal_expr_164
                                        connect _bundle_literal_expr_163.output_integer_mode, _match_arm_value_129.alu_common.output_integer_mode
                                        connect _bundle_literal_expr_162.alu_common, _bundle_literal_expr_163
                                        connect _bundle_literal_expr_162.invert_src0, _match_arm_value_129.invert_src0
                                        connect _bundle_literal_expr_162.invert_carry_in, _match_arm_value_129.invert_carry_in
                                        connect _bundle_literal_expr_162.invert_carry_out, _match_arm_value_129.invert_carry_out
                                        connect _bundle_literal_expr_162.add_pc, _match_arm_value_129.add_pc
                                        connect mapped_regs_5, {|AddSub: Ty44, AddSubI: Ty44, Logical: Ty45|}(AddSubI, _bundle_literal_expr_162) @[instruction.rs 502:1]
                                    Logical(_match_arm_value_130):
                                        wire _bundle_literal_expr_169: Ty1
                                        connect _bundle_literal_expr_169.value, _match_arm_value_130.alu_common.common.src[0]
                                        connect rename_1_src_0.addr, _bundle_literal_expr_169 @[reg_alloc.rs 204:29]
                                        wire _bundle_literal_expr_170: Ty1
                                        connect _bundle_literal_expr_170.value, _match_arm_value_130.alu_common.common.src[1]
                                        connect rename_1_src_1.addr, _bundle_literal_expr_170 @[reg_alloc.rs 204:29]
                                        wire _bundle_literal_expr_171: Ty45
                                        wire _bundle_literal_expr_172: Ty43
                                        wire _bundle_literal_expr_173: Ty42
                                        connect _bundle_literal_expr_173.prefix_pad, _match_arm_value_130.alu_common.common.prefix_pad
                                        connect _bundle_literal_expr_173.dest, _match_arm_value_116.unit_out_reg
                                        wire _array_literal_expr_50: UInt<6>[3]
                                        wire _array_literal_expr_51: UInt<6>[2]
                                        wire _cast_bundle_to_bits_expr_57: Ty57
                                        connect _cast_bundle_to_bits_expr_57.unit_num, rename_1_src_0.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_57.unit_out_reg, rename_1_src_0.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_57: UInt<6>
                                        connect _cast_to_bits_expr_57, cat(_cast_bundle_to_bits_expr_57.unit_out_reg, _cast_bundle_to_bits_expr_57.unit_num)
                                        connect _array_literal_expr_51[0], _cast_to_bits_expr_57
                                        wire _cast_bundle_to_bits_expr_58: Ty57
                                        connect _cast_bundle_to_bits_expr_58.unit_num, rename_1_src_1.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_58.unit_out_reg, rename_1_src_1.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_58: UInt<6>
                                        connect _cast_to_bits_expr_58, cat(_cast_bundle_to_bits_expr_58.unit_out_reg, _cast_bundle_to_bits_expr_58.unit_num)
                                        connect _array_literal_expr_51[1], _cast_to_bits_expr_58
                                        connect _array_literal_expr_50[0], _array_literal_expr_51[0]
                                        wire _array_literal_expr_52: UInt<6>[2]
                                        wire _cast_bundle_to_bits_expr_59: Ty57
                                        connect _cast_bundle_to_bits_expr_59.unit_num, rename_1_src_0.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_59.unit_out_reg, rename_1_src_0.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_59: UInt<6>
                                        connect _cast_to_bits_expr_59, cat(_cast_bundle_to_bits_expr_59.unit_out_reg, _cast_bundle_to_bits_expr_59.unit_num)
                                        connect _array_literal_expr_52[0], _cast_to_bits_expr_59
                                        wire _cast_bundle_to_bits_expr_60: Ty57
                                        connect _cast_bundle_to_bits_expr_60.unit_num, rename_1_src_1.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_60.unit_out_reg, rename_1_src_1.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_60: UInt<6>
                                        connect _cast_to_bits_expr_60, cat(_cast_bundle_to_bits_expr_60.unit_out_reg, _cast_bundle_to_bits_expr_60.unit_num)
                                        connect _array_literal_expr_52[1], _cast_to_bits_expr_60
                                        connect _array_literal_expr_50[1], _array_literal_expr_52[1]
                                        wire _bundle_literal_expr_174: Ty60
                                        connect _bundle_literal_expr_174.imm_low, _match_arm_value_130.alu_common.common.imm_low
                                        wire _array_literal_expr_53: UInt<8>[1]
                                        connect _array_literal_expr_53[0], _match_arm_value_130.alu_common.common.src[2]
                                        connect _bundle_literal_expr_174.reversed_src, _array_literal_expr_53
                                        connect _bundle_literal_expr_174.imm_sign, _match_arm_value_130.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_61: Ty61
                                        connect _cast_bundle_to_bits_expr_61.imm_low, _bundle_literal_expr_174.imm_low
                                        connect _cast_bundle_to_bits_expr_61.reversed_src, _bundle_literal_expr_174.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_61.imm_sign, asUInt(_bundle_literal_expr_174.imm_sign)
                                        wire _cast_to_bits_expr_61: UInt<34>
                                        connect _cast_to_bits_expr_61, cat(_cast_bundle_to_bits_expr_61.imm_sign, cat(_cast_bundle_to_bits_expr_61.reversed_src, _cast_bundle_to_bits_expr_61.imm_low))
                                        wire _cast_bits_to_bundle_expr_7: Ty60
                                        wire _cast_bits_to_bundle_expr_flattened_7: Ty61
                                        connect _cast_bits_to_bundle_expr_flattened_7.imm_low, bits(asUInt(asSInt(_cast_to_bits_expr_61)), 24, 0)
                                        connect _cast_bits_to_bundle_expr_7.imm_low, _cast_bits_to_bundle_expr_flattened_7.imm_low
                                        connect _cast_bits_to_bundle_expr_flattened_7.reversed_src, bits(asUInt(asSInt(_cast_to_bits_expr_61)), 32, 25)
                                        wire _cast_bits_to_array_expr_7: UInt<8>[1]
                                        wire _cast_bits_to_array_expr_flattened_7: UInt<8>[1]
                                        connect _cast_bits_to_array_expr_flattened_7[0], bits(_cast_bits_to_bundle_expr_flattened_7.reversed_src, 7, 0)
                                        connect _cast_bits_to_array_expr_7[0], _cast_bits_to_array_expr_flattened_7[0]
                                        connect _cast_bits_to_bundle_expr_7.reversed_src, _cast_bits_to_array_expr_7
                                        connect _cast_bits_to_bundle_expr_flattened_7.imm_sign, bits(asUInt(asSInt(_cast_to_bits_expr_61)), 33, 33)
                                        connect _cast_bits_to_bundle_expr_7.imm_sign, asSInt(_cast_bits_to_bundle_expr_flattened_7.imm_sign)
                                        connect _array_literal_expr_50[2], tail(_cast_bits_to_bundle_expr_7.reversed_src[0], 2)
                                        connect _bundle_literal_expr_173.src, _array_literal_expr_50
                                        wire _bundle_literal_expr_175: Ty60
                                        connect _bundle_literal_expr_175.imm_low, _match_arm_value_130.alu_common.common.imm_low
                                        wire _array_literal_expr_54: UInt<8>[1]
                                        connect _array_literal_expr_54[0], _match_arm_value_130.alu_common.common.src[2]
                                        connect _bundle_literal_expr_175.reversed_src, _array_literal_expr_54
                                        connect _bundle_literal_expr_175.imm_sign, _match_arm_value_130.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_62: Ty61
                                        connect _cast_bundle_to_bits_expr_62.imm_low, _bundle_literal_expr_175.imm_low
                                        connect _cast_bundle_to_bits_expr_62.reversed_src, _bundle_literal_expr_175.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_62.imm_sign, asUInt(_bundle_literal_expr_175.imm_sign)
                                        wire _cast_to_bits_expr_62: UInt<34>
                                        connect _cast_to_bits_expr_62, cat(_cast_bundle_to_bits_expr_62.imm_sign, cat(_cast_bundle_to_bits_expr_62.reversed_src, _cast_bundle_to_bits_expr_62.imm_low))
                                        connect _bundle_literal_expr_173.imm_low, bits(asSInt(_cast_to_bits_expr_62), 24, 0)
                                        wire _bundle_literal_expr_176: Ty60
                                        connect _bundle_literal_expr_176.imm_low, _match_arm_value_130.alu_common.common.imm_low
                                        wire _array_literal_expr_55: UInt<8>[1]
                                        connect _array_literal_expr_55[0], _match_arm_value_130.alu_common.common.src[2]
                                        connect _bundle_literal_expr_176.reversed_src, _array_literal_expr_55
                                        connect _bundle_literal_expr_176.imm_sign, _match_arm_value_130.alu_common.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_63: Ty61
                                        connect _cast_bundle_to_bits_expr_63.imm_low, _bundle_literal_expr_176.imm_low
                                        connect _cast_bundle_to_bits_expr_63.reversed_src, _bundle_literal_expr_176.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_63.imm_sign, asUInt(_bundle_literal_expr_176.imm_sign)
                                        wire _cast_to_bits_expr_63: UInt<34>
                                        connect _cast_to_bits_expr_63, cat(_cast_bundle_to_bits_expr_63.imm_sign, cat(_cast_bundle_to_bits_expr_63.reversed_src, _cast_bundle_to_bits_expr_63.imm_low))
                                        connect _bundle_literal_expr_173.imm_sign, shr(asSInt(_cast_to_bits_expr_63), 33)
                                        wire _bundle_literal_expr_177: Ty2
                                        invalidate _bundle_literal_expr_177
                                        connect _bundle_literal_expr_173._phantom, _bundle_literal_expr_177
                                        connect _bundle_literal_expr_172.common, _bundle_literal_expr_173
                                        connect _bundle_literal_expr_172.output_integer_mode, _match_arm_value_130.alu_common.output_integer_mode
                                        connect _bundle_literal_expr_171.alu_common, _bundle_literal_expr_172
                                        connect _bundle_literal_expr_171.lut, _match_arm_value_130.lut
                                        connect mapped_regs_5, {|AddSub: Ty44, AddSubI: Ty44, Logical: Ty45|}(Logical, _bundle_literal_expr_171) @[instruction.rs 502:1]
                                connect mapped_regs_4, {|AluBranch: Ty46, L2RegisterFile: Ty49, LoadStore: Ty50|}(AluBranch, mapped_regs_5) @[unit.rs 129:1]
                            L2RegisterFile(_match_arm_value_131):
                                wire mapped_regs_6: Ty49 @[instruction.rs 529:1]
                                match _match_arm_value_131: @[instruction.rs 529:1]
                                    ReadL2Reg(_match_arm_value_132):
                                        wire _bundle_literal_expr_178: Ty48
                                        wire _bundle_literal_expr_179: Ty47
                                        connect _bundle_literal_expr_179.prefix_pad, _match_arm_value_132.common.prefix_pad
                                        connect _bundle_literal_expr_179.dest, _match_arm_value_116.unit_out_reg
                                        wire _array_literal_expr_56: UInt<6>[3]
                                        connect _array_literal_expr_56[0], pad(UInt<0>(0h0), 6)
                                        connect _array_literal_expr_56[1], pad(UInt<0>(0h0), 6)
                                        wire _bundle_literal_expr_180: Ty60
                                        connect _bundle_literal_expr_180.imm_low, _match_arm_value_132.common.imm_low
                                        wire _array_literal_expr_57: UInt<8>[1]
                                        connect _array_literal_expr_57[0], _match_arm_value_132.common.src[2]
                                        connect _bundle_literal_expr_180.reversed_src, _array_literal_expr_57
                                        connect _bundle_literal_expr_180.imm_sign, _match_arm_value_132.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_64: Ty61
                                        connect _cast_bundle_to_bits_expr_64.imm_low, _bundle_literal_expr_180.imm_low
                                        connect _cast_bundle_to_bits_expr_64.reversed_src, _bundle_literal_expr_180.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_64.imm_sign, asUInt(_bundle_literal_expr_180.imm_sign)
                                        wire _cast_to_bits_expr_64: UInt<34>
                                        connect _cast_to_bits_expr_64, cat(_cast_bundle_to_bits_expr_64.imm_sign, cat(_cast_bundle_to_bits_expr_64.reversed_src, _cast_bundle_to_bits_expr_64.imm_low))
                                        wire _cast_bits_to_bundle_expr_8: Ty60
                                        wire _cast_bits_to_bundle_expr_flattened_8: Ty61
                                        connect _cast_bits_to_bundle_expr_flattened_8.imm_low, bits(asUInt(asSInt(_cast_to_bits_expr_64)), 24, 0)
                                        connect _cast_bits_to_bundle_expr_8.imm_low, _cast_bits_to_bundle_expr_flattened_8.imm_low
                                        connect _cast_bits_to_bundle_expr_flattened_8.reversed_src, bits(asUInt(asSInt(_cast_to_bits_expr_64)), 32, 25)
                                        wire _cast_bits_to_array_expr_8: UInt<8>[1]
                                        wire _cast_bits_to_array_expr_flattened_8: UInt<8>[1]
                                        connect _cast_bits_to_array_expr_flattened_8[0], bits(_cast_bits_to_bundle_expr_flattened_8.reversed_src, 7, 0)
                                        connect _cast_bits_to_array_expr_8[0], _cast_bits_to_array_expr_flattened_8[0]
                                        connect _cast_bits_to_bundle_expr_8.reversed_src, _cast_bits_to_array_expr_8
                                        connect _cast_bits_to_bundle_expr_flattened_8.imm_sign, bits(asUInt(asSInt(_cast_to_bits_expr_64)), 33, 33)
                                        connect _cast_bits_to_bundle_expr_8.imm_sign, asSInt(_cast_bits_to_bundle_expr_flattened_8.imm_sign)
                                        connect _array_literal_expr_56[2], tail(_cast_bits_to_bundle_expr_8.reversed_src[0], 2)
                                        connect _bundle_literal_expr_179.src, _array_literal_expr_56
                                        wire _bundle_literal_expr_181: Ty60
                                        connect _bundle_literal_expr_181.imm_low, _match_arm_value_132.common.imm_low
                                        wire _array_literal_expr_58: UInt<8>[1]
                                        connect _array_literal_expr_58[0], _match_arm_value_132.common.src[2]
                                        connect _bundle_literal_expr_181.reversed_src, _array_literal_expr_58
                                        connect _bundle_literal_expr_181.imm_sign, _match_arm_value_132.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_65: Ty61
                                        connect _cast_bundle_to_bits_expr_65.imm_low, _bundle_literal_expr_181.imm_low
                                        connect _cast_bundle_to_bits_expr_65.reversed_src, _bundle_literal_expr_181.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_65.imm_sign, asUInt(_bundle_literal_expr_181.imm_sign)
                                        wire _cast_to_bits_expr_65: UInt<34>
                                        connect _cast_to_bits_expr_65, cat(_cast_bundle_to_bits_expr_65.imm_sign, cat(_cast_bundle_to_bits_expr_65.reversed_src, _cast_bundle_to_bits_expr_65.imm_low))
                                        connect _bundle_literal_expr_179.imm_low, bits(asSInt(_cast_to_bits_expr_65), 24, 0)
                                        wire _bundle_literal_expr_182: Ty60
                                        connect _bundle_literal_expr_182.imm_low, _match_arm_value_132.common.imm_low
                                        wire _array_literal_expr_59: UInt<8>[1]
                                        connect _array_literal_expr_59[0], _match_arm_value_132.common.src[2]
                                        connect _bundle_literal_expr_182.reversed_src, _array_literal_expr_59
                                        connect _bundle_literal_expr_182.imm_sign, _match_arm_value_132.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_66: Ty61
                                        connect _cast_bundle_to_bits_expr_66.imm_low, _bundle_literal_expr_182.imm_low
                                        connect _cast_bundle_to_bits_expr_66.reversed_src, _bundle_literal_expr_182.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_66.imm_sign, asUInt(_bundle_literal_expr_182.imm_sign)
                                        wire _cast_to_bits_expr_66: UInt<34>
                                        connect _cast_to_bits_expr_66, cat(_cast_bundle_to_bits_expr_66.imm_sign, cat(_cast_bundle_to_bits_expr_66.reversed_src, _cast_bundle_to_bits_expr_66.imm_low))
                                        connect _bundle_literal_expr_179.imm_sign, shr(asSInt(_cast_to_bits_expr_66), 33)
                                        wire _bundle_literal_expr_183: Ty2
                                        invalidate _bundle_literal_expr_183
                                        connect _bundle_literal_expr_179._phantom, _bundle_literal_expr_183
                                        connect _bundle_literal_expr_178.common, _bundle_literal_expr_179
                                        connect mapped_regs_6, {|ReadL2Reg: Ty48, WriteL2Reg: Ty48|}(ReadL2Reg, _bundle_literal_expr_178) @[instruction.rs 529:1]
                                    WriteL2Reg(_match_arm_value_133):
                                        wire _bundle_literal_expr_184: Ty1
                                        connect _bundle_literal_expr_184.value, _match_arm_value_133.common.src[0]
                                        connect rename_1_src_0.addr, _bundle_literal_expr_184 @[reg_alloc.rs 204:29]
                                        wire _bundle_literal_expr_185: Ty48
                                        wire _bundle_literal_expr_186: Ty47
                                        connect _bundle_literal_expr_186.prefix_pad, _match_arm_value_133.common.prefix_pad
                                        connect _bundle_literal_expr_186.dest, _match_arm_value_116.unit_out_reg
                                        wire _array_literal_expr_60: UInt<6>[3]
                                        wire _array_literal_expr_61: UInt<6>[1]
                                        wire _cast_bundle_to_bits_expr_67: Ty57
                                        connect _cast_bundle_to_bits_expr_67.unit_num, rename_1_src_0.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_67.unit_out_reg, rename_1_src_0.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_67: UInt<6>
                                        connect _cast_to_bits_expr_67, cat(_cast_bundle_to_bits_expr_67.unit_out_reg, _cast_bundle_to_bits_expr_67.unit_num)
                                        connect _array_literal_expr_61[0], _cast_to_bits_expr_67
                                        connect _array_literal_expr_60[0], _array_literal_expr_61[0]
                                        connect _array_literal_expr_60[1], pad(UInt<0>(0h0), 6)
                                        wire _bundle_literal_expr_187: Ty60
                                        connect _bundle_literal_expr_187.imm_low, _match_arm_value_133.common.imm_low
                                        wire _array_literal_expr_62: UInt<8>[1]
                                        connect _array_literal_expr_62[0], _match_arm_value_133.common.src[2]
                                        connect _bundle_literal_expr_187.reversed_src, _array_literal_expr_62
                                        connect _bundle_literal_expr_187.imm_sign, _match_arm_value_133.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_68: Ty61
                                        connect _cast_bundle_to_bits_expr_68.imm_low, _bundle_literal_expr_187.imm_low
                                        connect _cast_bundle_to_bits_expr_68.reversed_src, _bundle_literal_expr_187.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_68.imm_sign, asUInt(_bundle_literal_expr_187.imm_sign)
                                        wire _cast_to_bits_expr_68: UInt<34>
                                        connect _cast_to_bits_expr_68, cat(_cast_bundle_to_bits_expr_68.imm_sign, cat(_cast_bundle_to_bits_expr_68.reversed_src, _cast_bundle_to_bits_expr_68.imm_low))
                                        wire _cast_bits_to_bundle_expr_9: Ty60
                                        wire _cast_bits_to_bundle_expr_flattened_9: Ty61
                                        connect _cast_bits_to_bundle_expr_flattened_9.imm_low, bits(asUInt(asSInt(_cast_to_bits_expr_68)), 24, 0)
                                        connect _cast_bits_to_bundle_expr_9.imm_low, _cast_bits_to_bundle_expr_flattened_9.imm_low
                                        connect _cast_bits_to_bundle_expr_flattened_9.reversed_src, bits(asUInt(asSInt(_cast_to_bits_expr_68)), 32, 25)
                                        wire _cast_bits_to_array_expr_9: UInt<8>[1]
                                        wire _cast_bits_to_array_expr_flattened_9: UInt<8>[1]
                                        connect _cast_bits_to_array_expr_flattened_9[0], bits(_cast_bits_to_bundle_expr_flattened_9.reversed_src, 7, 0)
                                        connect _cast_bits_to_array_expr_9[0], _cast_bits_to_array_expr_flattened_9[0]
                                        connect _cast_bits_to_bundle_expr_9.reversed_src, _cast_bits_to_array_expr_9
                                        connect _cast_bits_to_bundle_expr_flattened_9.imm_sign, bits(asUInt(asSInt(_cast_to_bits_expr_68)), 33, 33)
                                        connect _cast_bits_to_bundle_expr_9.imm_sign, asSInt(_cast_bits_to_bundle_expr_flattened_9.imm_sign)
                                        connect _array_literal_expr_60[2], tail(_cast_bits_to_bundle_expr_9.reversed_src[0], 2)
                                        connect _bundle_literal_expr_186.src, _array_literal_expr_60
                                        wire _bundle_literal_expr_188: Ty60
                                        connect _bundle_literal_expr_188.imm_low, _match_arm_value_133.common.imm_low
                                        wire _array_literal_expr_63: UInt<8>[1]
                                        connect _array_literal_expr_63[0], _match_arm_value_133.common.src[2]
                                        connect _bundle_literal_expr_188.reversed_src, _array_literal_expr_63
                                        connect _bundle_literal_expr_188.imm_sign, _match_arm_value_133.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_69: Ty61
                                        connect _cast_bundle_to_bits_expr_69.imm_low, _bundle_literal_expr_188.imm_low
                                        connect _cast_bundle_to_bits_expr_69.reversed_src, _bundle_literal_expr_188.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_69.imm_sign, asUInt(_bundle_literal_expr_188.imm_sign)
                                        wire _cast_to_bits_expr_69: UInt<34>
                                        connect _cast_to_bits_expr_69, cat(_cast_bundle_to_bits_expr_69.imm_sign, cat(_cast_bundle_to_bits_expr_69.reversed_src, _cast_bundle_to_bits_expr_69.imm_low))
                                        connect _bundle_literal_expr_186.imm_low, bits(asSInt(_cast_to_bits_expr_69), 24, 0)
                                        wire _bundle_literal_expr_189: Ty60
                                        connect _bundle_literal_expr_189.imm_low, _match_arm_value_133.common.imm_low
                                        wire _array_literal_expr_64: UInt<8>[1]
                                        connect _array_literal_expr_64[0], _match_arm_value_133.common.src[2]
                                        connect _bundle_literal_expr_189.reversed_src, _array_literal_expr_64
                                        connect _bundle_literal_expr_189.imm_sign, _match_arm_value_133.common.imm_sign
                                        wire _cast_bundle_to_bits_expr_70: Ty61
                                        connect _cast_bundle_to_bits_expr_70.imm_low, _bundle_literal_expr_189.imm_low
                                        connect _cast_bundle_to_bits_expr_70.reversed_src, _bundle_literal_expr_189.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_70.imm_sign, asUInt(_bundle_literal_expr_189.imm_sign)
                                        wire _cast_to_bits_expr_70: UInt<34>
                                        connect _cast_to_bits_expr_70, cat(_cast_bundle_to_bits_expr_70.imm_sign, cat(_cast_bundle_to_bits_expr_70.reversed_src, _cast_bundle_to_bits_expr_70.imm_low))
                                        connect _bundle_literal_expr_186.imm_sign, shr(asSInt(_cast_to_bits_expr_70), 33)
                                        wire _bundle_literal_expr_190: Ty2
                                        invalidate _bundle_literal_expr_190
                                        connect _bundle_literal_expr_186._phantom, _bundle_literal_expr_190
                                        connect _bundle_literal_expr_185.common, _bundle_literal_expr_186
                                        connect mapped_regs_6, {|ReadL2Reg: Ty48, WriteL2Reg: Ty48|}(WriteL2Reg, _bundle_literal_expr_185) @[instruction.rs 529:1]
                                connect mapped_regs_4, {|AluBranch: Ty46, L2RegisterFile: Ty49, LoadStore: Ty50|}(L2RegisterFile, mapped_regs_6) @[unit.rs 129:1]
                            LoadStore(_match_arm_value_134):
                                wire mapped_regs_7: Ty50 @[instruction.rs 564:1]
                                match _match_arm_value_134: @[instruction.rs 564:1]
                                    Load(_match_arm_value_135):
                                        wire _bundle_literal_expr_191: Ty47
                                        connect _bundle_literal_expr_191.prefix_pad, _match_arm_value_135.prefix_pad
                                        connect _bundle_literal_expr_191.dest, _match_arm_value_116.unit_out_reg
                                        wire _array_literal_expr_65: UInt<6>[3]
                                        connect _array_literal_expr_65[0], pad(UInt<0>(0h0), 6)
                                        connect _array_literal_expr_65[1], pad(UInt<0>(0h0), 6)
                                        wire _bundle_literal_expr_192: Ty60
                                        connect _bundle_literal_expr_192.imm_low, _match_arm_value_135.imm_low
                                        wire _array_literal_expr_66: UInt<8>[1]
                                        connect _array_literal_expr_66[0], _match_arm_value_135.src[2]
                                        connect _bundle_literal_expr_192.reversed_src, _array_literal_expr_66
                                        connect _bundle_literal_expr_192.imm_sign, _match_arm_value_135.imm_sign
                                        wire _cast_bundle_to_bits_expr_71: Ty61
                                        connect _cast_bundle_to_bits_expr_71.imm_low, _bundle_literal_expr_192.imm_low
                                        connect _cast_bundle_to_bits_expr_71.reversed_src, _bundle_literal_expr_192.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_71.imm_sign, asUInt(_bundle_literal_expr_192.imm_sign)
                                        wire _cast_to_bits_expr_71: UInt<34>
                                        connect _cast_to_bits_expr_71, cat(_cast_bundle_to_bits_expr_71.imm_sign, cat(_cast_bundle_to_bits_expr_71.reversed_src, _cast_bundle_to_bits_expr_71.imm_low))
                                        wire _cast_bits_to_bundle_expr_10: Ty60
                                        wire _cast_bits_to_bundle_expr_flattened_10: Ty61
                                        connect _cast_bits_to_bundle_expr_flattened_10.imm_low, bits(asUInt(asSInt(_cast_to_bits_expr_71)), 24, 0)
                                        connect _cast_bits_to_bundle_expr_10.imm_low, _cast_bits_to_bundle_expr_flattened_10.imm_low
                                        connect _cast_bits_to_bundle_expr_flattened_10.reversed_src, bits(asUInt(asSInt(_cast_to_bits_expr_71)), 32, 25)
                                        wire _cast_bits_to_array_expr_10: UInt<8>[1]
                                        wire _cast_bits_to_array_expr_flattened_10: UInt<8>[1]
                                        connect _cast_bits_to_array_expr_flattened_10[0], bits(_cast_bits_to_bundle_expr_flattened_10.reversed_src, 7, 0)
                                        connect _cast_bits_to_array_expr_10[0], _cast_bits_to_array_expr_flattened_10[0]
                                        connect _cast_bits_to_bundle_expr_10.reversed_src, _cast_bits_to_array_expr_10
                                        connect _cast_bits_to_bundle_expr_flattened_10.imm_sign, bits(asUInt(asSInt(_cast_to_bits_expr_71)), 33, 33)
                                        connect _cast_bits_to_bundle_expr_10.imm_sign, asSInt(_cast_bits_to_bundle_expr_flattened_10.imm_sign)
                                        connect _array_literal_expr_65[2], tail(_cast_bits_to_bundle_expr_10.reversed_src[0], 2)
                                        connect _bundle_literal_expr_191.src, _array_literal_expr_65
                                        wire _bundle_literal_expr_193: Ty60
                                        connect _bundle_literal_expr_193.imm_low, _match_arm_value_135.imm_low
                                        wire _array_literal_expr_67: UInt<8>[1]
                                        connect _array_literal_expr_67[0], _match_arm_value_135.src[2]
                                        connect _bundle_literal_expr_193.reversed_src, _array_literal_expr_67
                                        connect _bundle_literal_expr_193.imm_sign, _match_arm_value_135.imm_sign
                                        wire _cast_bundle_to_bits_expr_72: Ty61
                                        connect _cast_bundle_to_bits_expr_72.imm_low, _bundle_literal_expr_193.imm_low
                                        connect _cast_bundle_to_bits_expr_72.reversed_src, _bundle_literal_expr_193.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_72.imm_sign, asUInt(_bundle_literal_expr_193.imm_sign)
                                        wire _cast_to_bits_expr_72: UInt<34>
                                        connect _cast_to_bits_expr_72, cat(_cast_bundle_to_bits_expr_72.imm_sign, cat(_cast_bundle_to_bits_expr_72.reversed_src, _cast_bundle_to_bits_expr_72.imm_low))
                                        connect _bundle_literal_expr_191.imm_low, bits(asSInt(_cast_to_bits_expr_72), 24, 0)
                                        wire _bundle_literal_expr_194: Ty60
                                        connect _bundle_literal_expr_194.imm_low, _match_arm_value_135.imm_low
                                        wire _array_literal_expr_68: UInt<8>[1]
                                        connect _array_literal_expr_68[0], _match_arm_value_135.src[2]
                                        connect _bundle_literal_expr_194.reversed_src, _array_literal_expr_68
                                        connect _bundle_literal_expr_194.imm_sign, _match_arm_value_135.imm_sign
                                        wire _cast_bundle_to_bits_expr_73: Ty61
                                        connect _cast_bundle_to_bits_expr_73.imm_low, _bundle_literal_expr_194.imm_low
                                        connect _cast_bundle_to_bits_expr_73.reversed_src, _bundle_literal_expr_194.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_73.imm_sign, asUInt(_bundle_literal_expr_194.imm_sign)
                                        wire _cast_to_bits_expr_73: UInt<34>
                                        connect _cast_to_bits_expr_73, cat(_cast_bundle_to_bits_expr_73.imm_sign, cat(_cast_bundle_to_bits_expr_73.reversed_src, _cast_bundle_to_bits_expr_73.imm_low))
                                        connect _bundle_literal_expr_191.imm_sign, shr(asSInt(_cast_to_bits_expr_73), 33)
                                        wire _bundle_literal_expr_195: Ty2
                                        invalidate _bundle_literal_expr_195
                                        connect _bundle_literal_expr_191._phantom, _bundle_literal_expr_195
                                        connect mapped_regs_7, {|Load: Ty47, Store: Ty47|}(Load, _bundle_literal_expr_191) @[instruction.rs 564:1]
                                    Store(_match_arm_value_136):
                                        wire _bundle_literal_expr_196: Ty1
                                        connect _bundle_literal_expr_196.value, _match_arm_value_136.src[0]
                                        connect rename_1_src_0.addr, _bundle_literal_expr_196 @[reg_alloc.rs 204:29]
                                        wire _bundle_literal_expr_197: Ty47
                                        connect _bundle_literal_expr_197.prefix_pad, _match_arm_value_136.prefix_pad
                                        connect _bundle_literal_expr_197.dest, _match_arm_value_116.unit_out_reg
                                        wire _array_literal_expr_69: UInt<6>[3]
                                        wire _array_literal_expr_70: UInt<6>[1]
                                        wire _cast_bundle_to_bits_expr_74: Ty57
                                        connect _cast_bundle_to_bits_expr_74.unit_num, rename_1_src_0.data.unit_num.adj_value
                                        connect _cast_bundle_to_bits_expr_74.unit_out_reg, rename_1_src_0.data.unit_out_reg.value
                                        wire _cast_to_bits_expr_74: UInt<6>
                                        connect _cast_to_bits_expr_74, cat(_cast_bundle_to_bits_expr_74.unit_out_reg, _cast_bundle_to_bits_expr_74.unit_num)
                                        connect _array_literal_expr_70[0], _cast_to_bits_expr_74
                                        connect _array_literal_expr_69[0], _array_literal_expr_70[0]
                                        connect _array_literal_expr_69[1], pad(UInt<0>(0h0), 6)
                                        wire _bundle_literal_expr_198: Ty60
                                        connect _bundle_literal_expr_198.imm_low, _match_arm_value_136.imm_low
                                        wire _array_literal_expr_71: UInt<8>[1]
                                        connect _array_literal_expr_71[0], _match_arm_value_136.src[2]
                                        connect _bundle_literal_expr_198.reversed_src, _array_literal_expr_71
                                        connect _bundle_literal_expr_198.imm_sign, _match_arm_value_136.imm_sign
                                        wire _cast_bundle_to_bits_expr_75: Ty61
                                        connect _cast_bundle_to_bits_expr_75.imm_low, _bundle_literal_expr_198.imm_low
                                        connect _cast_bundle_to_bits_expr_75.reversed_src, _bundle_literal_expr_198.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_75.imm_sign, asUInt(_bundle_literal_expr_198.imm_sign)
                                        wire _cast_to_bits_expr_75: UInt<34>
                                        connect _cast_to_bits_expr_75, cat(_cast_bundle_to_bits_expr_75.imm_sign, cat(_cast_bundle_to_bits_expr_75.reversed_src, _cast_bundle_to_bits_expr_75.imm_low))
                                        wire _cast_bits_to_bundle_expr_11: Ty60
                                        wire _cast_bits_to_bundle_expr_flattened_11: Ty61
                                        connect _cast_bits_to_bundle_expr_flattened_11.imm_low, bits(asUInt(asSInt(_cast_to_bits_expr_75)), 24, 0)
                                        connect _cast_bits_to_bundle_expr_11.imm_low, _cast_bits_to_bundle_expr_flattened_11.imm_low
                                        connect _cast_bits_to_bundle_expr_flattened_11.reversed_src, bits(asUInt(asSInt(_cast_to_bits_expr_75)), 32, 25)
                                        wire _cast_bits_to_array_expr_11: UInt<8>[1]
                                        wire _cast_bits_to_array_expr_flattened_11: UInt<8>[1]
                                        connect _cast_bits_to_array_expr_flattened_11[0], bits(_cast_bits_to_bundle_expr_flattened_11.reversed_src, 7, 0)
                                        connect _cast_bits_to_array_expr_11[0], _cast_bits_to_array_expr_flattened_11[0]
                                        connect _cast_bits_to_bundle_expr_11.reversed_src, _cast_bits_to_array_expr_11
                                        connect _cast_bits_to_bundle_expr_flattened_11.imm_sign, bits(asUInt(asSInt(_cast_to_bits_expr_75)), 33, 33)
                                        connect _cast_bits_to_bundle_expr_11.imm_sign, asSInt(_cast_bits_to_bundle_expr_flattened_11.imm_sign)
                                        connect _array_literal_expr_69[2], tail(_cast_bits_to_bundle_expr_11.reversed_src[0], 2)
                                        connect _bundle_literal_expr_197.src, _array_literal_expr_69
                                        wire _bundle_literal_expr_199: Ty60
                                        connect _bundle_literal_expr_199.imm_low, _match_arm_value_136.imm_low
                                        wire _array_literal_expr_72: UInt<8>[1]
                                        connect _array_literal_expr_72[0], _match_arm_value_136.src[2]
                                        connect _bundle_literal_expr_199.reversed_src, _array_literal_expr_72
                                        connect _bundle_literal_expr_199.imm_sign, _match_arm_value_136.imm_sign
                                        wire _cast_bundle_to_bits_expr_76: Ty61
                                        connect _cast_bundle_to_bits_expr_76.imm_low, _bundle_literal_expr_199.imm_low
                                        connect _cast_bundle_to_bits_expr_76.reversed_src, _bundle_literal_expr_199.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_76.imm_sign, asUInt(_bundle_literal_expr_199.imm_sign)
                                        wire _cast_to_bits_expr_76: UInt<34>
                                        connect _cast_to_bits_expr_76, cat(_cast_bundle_to_bits_expr_76.imm_sign, cat(_cast_bundle_to_bits_expr_76.reversed_src, _cast_bundle_to_bits_expr_76.imm_low))
                                        connect _bundle_literal_expr_197.imm_low, bits(asSInt(_cast_to_bits_expr_76), 24, 0)
                                        wire _bundle_literal_expr_200: Ty60
                                        connect _bundle_literal_expr_200.imm_low, _match_arm_value_136.imm_low
                                        wire _array_literal_expr_73: UInt<8>[1]
                                        connect _array_literal_expr_73[0], _match_arm_value_136.src[2]
                                        connect _bundle_literal_expr_200.reversed_src, _array_literal_expr_73
                                        connect _bundle_literal_expr_200.imm_sign, _match_arm_value_136.imm_sign
                                        wire _cast_bundle_to_bits_expr_77: Ty61
                                        connect _cast_bundle_to_bits_expr_77.imm_low, _bundle_literal_expr_200.imm_low
                                        connect _cast_bundle_to_bits_expr_77.reversed_src, _bundle_literal_expr_200.reversed_src[0]
                                        connect _cast_bundle_to_bits_expr_77.imm_sign, asUInt(_bundle_literal_expr_200.imm_sign)
                                        wire _cast_to_bits_expr_77: UInt<34>
                                        connect _cast_to_bits_expr_77, cat(_cast_bundle_to_bits_expr_77.imm_sign, cat(_cast_bundle_to_bits_expr_77.reversed_src, _cast_bundle_to_bits_expr_77.imm_low))
                                        connect _bundle_literal_expr_197.imm_sign, shr(asSInt(_cast_to_bits_expr_77), 33)
                                        wire _bundle_literal_expr_201: Ty2
                                        invalidate _bundle_literal_expr_201
                                        connect _bundle_literal_expr_197._phantom, _bundle_literal_expr_201
                                        connect mapped_regs_7, {|Load: Ty47, Store: Ty47|}(Store, _bundle_literal_expr_197) @[instruction.rs 564:1]
                                connect mapped_regs_4, {|AluBranch: Ty46, L2RegisterFile: Ty49, LoadStore: Ty50|}(LoadStore, mapped_regs_7) @[unit.rs 129:1]
                        connect renamed_mops[1], {|HdlNone, HdlSome: Ty51|}(HdlSome, mapped_regs_4) @[reg_alloc.rs 197:17]
                        wire flag_reg_14: Ty1 @[instruction.rs 831:32]
                        wire _bundle_literal_expr_202: Ty1
                        connect _bundle_literal_expr_202.value, tail(UInt<32>(0h0), 24)
                        connect flag_reg_14, _bundle_literal_expr_202 @[instruction.rs 832:17]
                        match dest_reg_28.flag_regs[0]: @[instruction.rs 834:17]
                            HdlNone:
                                skip
                            HdlSome(_match_arm_value_137):
                                wire _bundle_literal_expr_203: Ty1
                                connect _bundle_literal_expr_203.value, tail(UInt<32>(0hFE), 24)
                                connect flag_reg_14, _bundle_literal_expr_203 @[instruction.rs 836:21]
                        wire flag_reg_15: Ty1 @[instruction.rs 831:32]
                        wire _bundle_literal_expr_204: Ty1
                        connect _bundle_literal_expr_204.value, tail(UInt<32>(0h0), 24)
                        connect flag_reg_15, _bundle_literal_expr_204 @[instruction.rs 832:17]
                        match dest_reg_28.flag_regs[1]: @[instruction.rs 834:17]
                            HdlNone:
                                skip
                            HdlSome(_match_arm_value_138):
                                wire _bundle_literal_expr_205: Ty1
                                connect _bundle_literal_expr_205.value, tail(UInt<32>(0hFF), 24)
                                connect flag_reg_15, _bundle_literal_expr_205 @[instruction.rs 836:21]
                        when and(geq(dest_reg_28.normal_regs[0].value, UInt<32>(0h1)), lt(dest_reg_28.normal_regs[0].value, UInt<32>(0hFE))): @[reg_alloc.rs 228:25]
                            connect rename_table_normal_1_dest0.data, _match_arm_value_116 @[reg_alloc.rs 229:29]
                            ; connect different types:
                            ; lhs: UInt<8>
                            ; rhs: UInt<33>
                            connect rename_table_normal_1_dest0.addr, sub(dest_reg_28.normal_regs[0].value, UInt<32>(0h1)) @[reg_alloc.rs 233:33]
                            connect rename_table_normal_1_dest0.en, UInt<1>(0h1) @[reg_alloc.rs 235:29]
                        when and(geq(dest_reg_28.normal_regs[0].value, UInt<32>(0hFE)), lt(dest_reg_28.normal_regs[0].value, UInt<32>(0h100))): @[reg_alloc.rs 228:25]
                            connect rename_table_special_1_dest0.data, _match_arm_value_116 @[reg_alloc.rs 229:29]
                            ; connect different types:
                            ; lhs: UInt<1>
                            ; rhs: UInt<33>
                            connect rename_table_special_1_dest0.addr, sub(dest_reg_28.normal_regs[0].value, UInt<32>(0hFE)) @[reg_alloc.rs 233:33]
                            connect rename_table_special_1_dest0.en, UInt<1>(0h1) @[reg_alloc.rs 235:29]
                        when and(geq(dest_reg_28.normal_regs[1].value, UInt<32>(0h1)), lt(dest_reg_28.normal_regs[1].value, UInt<32>(0hFE))): @[reg_alloc.rs 228:25]
                            connect rename_table_normal_1_dest1.data, _match_arm_value_116 @[reg_alloc.rs 229:29]
                            ; connect different types:
                            ; lhs: UInt<8>
                            ; rhs: UInt<33>
                            connect rename_table_normal_1_dest1.addr, sub(dest_reg_28.normal_regs[1].value, UInt<32>(0h1)) @[reg_alloc.rs 233:33]
                            connect rename_table_normal_1_dest1.en, UInt<1>(0h1) @[reg_alloc.rs 235:29]
                        when and(geq(dest_reg_28.normal_regs[1].value, UInt<32>(0hFE)), lt(dest_reg_28.normal_regs[1].value, UInt<32>(0h100))): @[reg_alloc.rs 228:25]
                            connect rename_table_special_1_dest1.data, _match_arm_value_116 @[reg_alloc.rs 229:29]
                            ; connect different types:
                            ; lhs: UInt<1>
                            ; rhs: UInt<33>
                            connect rename_table_special_1_dest1.addr, sub(dest_reg_28.normal_regs[1].value, UInt<32>(0hFE)) @[reg_alloc.rs 233:33]
                            connect rename_table_special_1_dest1.en, UInt<1>(0h1) @[reg_alloc.rs 235:29]
                        when and(geq(flag_reg_14.value, UInt<32>(0hFE)), lt(flag_reg_14.value, UInt<32>(0h100))): @[reg_alloc.rs 228:25]
                            connect rename_table_special_1_flag0_rFE.data, _match_arm_value_116 @[reg_alloc.rs 229:29]
                            ; connect different types:
                            ; lhs: UInt<1>
                            ; rhs: UInt<32>
                            connect rename_table_special_1_flag0_rFE.addr, UInt<32>(0h0) @[reg_alloc.rs 231:33]
                            connect rename_table_special_1_flag0_rFE.en, UInt<1>(0h1) @[reg_alloc.rs 235:29]
                        when and(geq(flag_reg_15.value, UInt<32>(0hFE)), lt(flag_reg_15.value, UInt<32>(0h100))): @[reg_alloc.rs 228:25]
                            connect rename_table_special_1_flag1_rFF.data, _match_arm_value_116 @[reg_alloc.rs 229:29]
                            ; connect different types:
                            ; lhs: UInt<1>
                            ; rhs: UInt<32>
                            connect rename_table_special_1_flag1_rFF.addr, UInt<32>(0h1) @[reg_alloc.rs 231:33]
                            connect rename_table_special_1_flag1_rFF.en, UInt<1>(0h1) @[reg_alloc.rs 235:29]
        wire selected_unit_index_leaf_1_0: Ty41 @[reg_alloc.rs 249:25]
        connect selected_unit_index_leaf_1_0, {|HdlNone, HdlSome: UInt<2>|}(HdlNone) @[reg_alloc.rs 252:21]
        wire unit_index_1_0: UInt<2> @[reg_alloc.rs 258:25]
        ; connect different types:
        ; lhs: UInt<2>
        ; rhs: UInt<64>
        connect unit_index_1_0, UInt<64>(0h0) @[reg_alloc.rs 261:21]
        when available_units[1][0]: @[reg_alloc.rs 263:21]
            connect selected_unit_index_leaf_1_0, {|HdlNone, HdlSome: UInt<2>|}(HdlSome, unit_index_1_0) @[reg_alloc.rs 264:25]
        wire selected_unit_index_leaf_1_1: Ty41 @[reg_alloc.rs 249:25]
        connect selected_unit_index_leaf_1_1, {|HdlNone, HdlSome: UInt<2>|}(HdlNone) @[reg_alloc.rs 252:21]
        wire unit_index_1_1: UInt<2> @[reg_alloc.rs 258:25]
        ; connect different types:
        ; lhs: UInt<2>
        ; rhs: UInt<64>
        connect unit_index_1_1, UInt<64>(0h1) @[reg_alloc.rs 261:21]
        when available_units[1][1]: @[reg_alloc.rs 263:21]
            connect selected_unit_index_leaf_1_1, {|HdlNone, HdlSome: UInt<2>|}(HdlSome, unit_index_1_1) @[reg_alloc.rs 264:25]
        wire selected_unit_index_node_1_0: Ty41 @[reg_alloc.rs 271:25]
        connect selected_unit_index_node_1_0, selected_unit_index_leaf_1_0 @[reg_alloc.rs 275:21]
        match selected_unit_index_leaf_1_0: @[reg_alloc.rs 277:21]
            HdlNone:
                connect selected_unit_index_node_1_0, selected_unit_index_leaf_1_1 @[reg_alloc.rs 278:25]
            HdlSome(_match_arm_value_139):
                skip
        connect selected_unit_indexes[1], selected_unit_index_node_1_0 @[reg_alloc.rs 241:9]
        match selected_unit_indexes[0]: @[reg_alloc.rs 291:13]
            HdlNone:
                skip
            HdlSome(_match_arm_value_140):
                connect available_units[1][_match_arm_value_140], UInt<1>(0h0) @[reg_alloc.rs 292:17]
        wire _array_literal_expr_74: Ty53[2]
        connect _array_literal_expr_74[0], {|HdlNone, HdlSome: Ty25|}(HdlNone)
        connect _array_literal_expr_74[1], {|HdlNone, HdlSome: Ty25|}(HdlNone)
        connect renamed_mops_out_reg, _array_literal_expr_74 @[reg_alloc.rs 299:5]
        inst unit_0 of alu_branch @[reg_alloc.rs 311:13]
        connect unit_0.cd, cd @[reg_alloc.rs 313:9]
        inst unit_0_free_regs_tracker of unit_free_regs_tracker @[reg_alloc.rs 326:13]
        connect unit_0_free_regs_tracker.cd, cd @[reg_alloc.rs 328:9]
        wire _uninit_expr_12: Ty78
        invalidate _uninit_expr_12
        connect unit_0_free_regs_tracker.free_in[0].data, _uninit_expr_12 @[reg_alloc.rs 330:9]
        connect unit_0_free_regs_tracker.alloc_out[0].ready, UInt<1>(0h0) @[reg_alloc.rs 334:9]
        connect unit_0.input_insn.data, {|HdlNone, HdlSome: Ty46|}(HdlNone) @[reg_alloc.rs 335:9]
        match unit_0_free_regs_tracker.alloc_out[0].data: @[reg_alloc.rs 341:13]
            HdlNone:
                connect available_units[0][0], UInt<1>(0h0) @[reg_alloc.rs 343:17]
            HdlSome(_match_arm_value_141):
                skip
        when not(unit_0.input_insn.ready): @[reg_alloc.rs 346:13]
            connect available_units[0][0], UInt<1>(0h0) @[reg_alloc.rs 348:17]
        match selected_unit_indexes[0]: @[reg_alloc.rs 351:13]
            HdlNone:
                skip
            HdlSome(_match_arm_value_142):
                when eq(_match_arm_value_142, UInt<64>(0h0)): @[reg_alloc.rs 353:17]
                    wire and_then_out: Ty62 @[reg_alloc.rs 357:25]
                    connect unit_0_free_regs_tracker.alloc_out[0].ready, UInt<1>(0h1) @[reg_alloc.rs 354:21]
                    match renamed_mops[0]: @[reg_alloc.rs 357:25]
                        HdlNone:
                            connect and_then_out, {|HdlNone, HdlSome: Ty46|}(HdlNone) @[reg_alloc.rs 357:25]
                        HdlSome(_match_arm_value_143):
                            wire alu_branch_mop: Ty62 @[unit.rs 129:1]
                            connect alu_branch_mop, {|HdlNone, HdlSome: Ty46|}(HdlNone) @[unit.rs 129:1]
                            match _match_arm_value_143: @[unit.rs 129:1]
                                AluBranch(_match_arm_value_144):
                                    connect alu_branch_mop, {|HdlNone, HdlSome: Ty46|}(HdlSome, _match_arm_value_144) @[unit.rs 129:1]
                                L2RegisterFile(_match_arm_value_145):
                                    skip
                                LoadStore(_match_arm_value_146):
                                    skip
                            connect and_then_out, alu_branch_mop @[reg_alloc.rs 357:25]
                    match and_then_out: @[reg_alloc.rs 356:21]
                        HdlNone:
                            wire _uninit_expr_13: Ty46
                            invalidate _uninit_expr_13
                            connect unit_0.input_insn.data, {|HdlNone, HdlSome: Ty46|}(HdlSome, _uninit_expr_13) @[reg_alloc.rs 361:25]
                        HdlSome(_match_arm_value_147):
                            connect unit_0.input_insn.data, {|HdlNone, HdlSome: Ty46|}(HdlSome, _match_arm_value_147) @[reg_alloc.rs 359:25]
                    match unit_0_free_regs_tracker.alloc_out[0].data: @[reg_alloc.rs 368:21]
                        HdlNone:
                            skip
                        HdlSome(_match_arm_value_148):
                            wire _bundle_literal_expr_206: Ty25
                            wire _bundle_literal_expr_207: Ty23
                            connect _bundle_literal_expr_207.adj_value, tail(UInt<64>(0h1), 62)
                            connect _bundle_literal_expr_206.unit_num, _bundle_literal_expr_207
                            wire _bundle_literal_expr_208: Ty24
                            connect _bundle_literal_expr_208.value, _match_arm_value_148
                            connect _bundle_literal_expr_206.unit_out_reg, _bundle_literal_expr_208
                            connect renamed_mops_out_reg[0], {|HdlNone, HdlSome: Ty25|}(HdlSome, _bundle_literal_expr_206) @[reg_alloc.rs 374:25]
        match unit_0_free_regs_tracker.alloc_out[0].data: @[reg_alloc.rs 341:13]
            HdlNone:
                connect available_units[1][0], UInt<1>(0h0) @[reg_alloc.rs 343:17]
            HdlSome(_match_arm_value_149):
                skip
        when not(unit_0.input_insn.ready): @[reg_alloc.rs 346:13]
            connect available_units[1][0], UInt<1>(0h0) @[reg_alloc.rs 348:17]
        match selected_unit_indexes[1]: @[reg_alloc.rs 351:13]
            HdlNone:
                skip
            HdlSome(_match_arm_value_150):
                when eq(_match_arm_value_150, UInt<64>(0h0)): @[reg_alloc.rs 353:17]
                    wire and_then_out_1: Ty62 @[reg_alloc.rs 357:25]
                    connect unit_0_free_regs_tracker.alloc_out[0].ready, UInt<1>(0h1) @[reg_alloc.rs 354:21]
                    match renamed_mops[1]: @[reg_alloc.rs 357:25]
                        HdlNone:
                            connect and_then_out_1, {|HdlNone, HdlSome: Ty46|}(HdlNone) @[reg_alloc.rs 357:25]
                        HdlSome(_match_arm_value_151):
                            wire alu_branch_mop_1: Ty62 @[unit.rs 129:1]
                            connect alu_branch_mop_1, {|HdlNone, HdlSome: Ty46|}(HdlNone) @[unit.rs 129:1]
                            match _match_arm_value_151: @[unit.rs 129:1]
                                AluBranch(_match_arm_value_152):
                                    connect alu_branch_mop_1, {|HdlNone, HdlSome: Ty46|}(HdlSome, _match_arm_value_152) @[unit.rs 129:1]
                                L2RegisterFile(_match_arm_value_153):
                                    skip
                                LoadStore(_match_arm_value_154):
                                    skip
                            connect and_then_out_1, alu_branch_mop_1 @[reg_alloc.rs 357:25]
                    match and_then_out_1: @[reg_alloc.rs 356:21]
                        HdlNone:
                            wire _uninit_expr_14: Ty46
                            invalidate _uninit_expr_14
                            connect unit_0.input_insn.data, {|HdlNone, HdlSome: Ty46|}(HdlSome, _uninit_expr_14) @[reg_alloc.rs 361:25]
                        HdlSome(_match_arm_value_155):
                            connect unit_0.input_insn.data, {|HdlNone, HdlSome: Ty46|}(HdlSome, _match_arm_value_155) @[reg_alloc.rs 359:25]
                    match unit_0_free_regs_tracker.alloc_out[0].data: @[reg_alloc.rs 368:21]
                        HdlNone:
                            skip
                        HdlSome(_match_arm_value_156):
                            wire _bundle_literal_expr_209: Ty25
                            wire _bundle_literal_expr_210: Ty23
                            connect _bundle_literal_expr_210.adj_value, tail(UInt<64>(0h1), 62)
                            connect _bundle_literal_expr_209.unit_num, _bundle_literal_expr_210
                            wire _bundle_literal_expr_211: Ty24
                            connect _bundle_literal_expr_211.value, _match_arm_value_156
                            connect _bundle_literal_expr_209.unit_out_reg, _bundle_literal_expr_211
                            connect renamed_mops_out_reg[1], {|HdlNone, HdlSome: Ty25|}(HdlSome, _bundle_literal_expr_209) @[reg_alloc.rs 374:25]
        wire _bundle_literal_expr_212: Ty68
        wire _array_literal_expr_75: Ty67[2]
        connect _array_literal_expr_75[0], {|HdlNone, HdlSome: Ty66|}(HdlNone)
        connect _array_literal_expr_75[1], {|HdlNone, HdlSome: Ty66|}(HdlNone)
        connect _bundle_literal_expr_212.unit_output_writes, _array_literal_expr_75
        wire _bundle_literal_expr_213: Ty2
        invalidate _bundle_literal_expr_213
        connect _bundle_literal_expr_212._phantom, _bundle_literal_expr_213
        connect unit_0.unit_forwarding_info, _bundle_literal_expr_212 @[reg_alloc.rs 389:9]
        connect unit_0.`output`.ready, UInt<1>(0h0) @[reg_alloc.rs 400:9]
        connect unit_0.cancel_input.data, {|HdlNone, HdlSome: Ty69|}(HdlNone) @[reg_alloc.rs 402:9]
        inst unit_1 of alu_branch_1 @[reg_alloc.rs 311:13]
        connect unit_1.cd, cd @[reg_alloc.rs 313:9]
        inst unit_1_free_regs_tracker of unit_free_regs_tracker_1 @[reg_alloc.rs 326:13]
        connect unit_1_free_regs_tracker.cd, cd @[reg_alloc.rs 328:9]
        wire _uninit_expr_15: Ty78
        invalidate _uninit_expr_15
        connect unit_1_free_regs_tracker.free_in[0].data, _uninit_expr_15 @[reg_alloc.rs 330:9]
        connect unit_1_free_regs_tracker.alloc_out[0].ready, UInt<1>(0h0) @[reg_alloc.rs 334:9]
        connect unit_1.input_insn.data, {|HdlNone, HdlSome: Ty46|}(HdlNone) @[reg_alloc.rs 335:9]
        match unit_1_free_regs_tracker.alloc_out[0].data: @[reg_alloc.rs 341:13]
            HdlNone:
                connect available_units[0][1], UInt<1>(0h0) @[reg_alloc.rs 343:17]
            HdlSome(_match_arm_value_157):
                skip
        when not(unit_1.input_insn.ready): @[reg_alloc.rs 346:13]
            connect available_units[0][1], UInt<1>(0h0) @[reg_alloc.rs 348:17]
        match selected_unit_indexes[0]: @[reg_alloc.rs 351:13]
            HdlNone:
                skip
            HdlSome(_match_arm_value_158):
                when eq(_match_arm_value_158, UInt<64>(0h1)): @[reg_alloc.rs 353:17]
                    wire and_then_out_2: Ty62 @[reg_alloc.rs 357:25]
                    connect unit_1_free_regs_tracker.alloc_out[0].ready, UInt<1>(0h1) @[reg_alloc.rs 354:21]
                    match renamed_mops[0]: @[reg_alloc.rs 357:25]
                        HdlNone:
                            connect and_then_out_2, {|HdlNone, HdlSome: Ty46|}(HdlNone) @[reg_alloc.rs 357:25]
                        HdlSome(_match_arm_value_159):
                            wire alu_branch_mop_2: Ty62 @[unit.rs 129:1]
                            connect alu_branch_mop_2, {|HdlNone, HdlSome: Ty46|}(HdlNone) @[unit.rs 129:1]
                            match _match_arm_value_159: @[unit.rs 129:1]
                                AluBranch(_match_arm_value_160):
                                    connect alu_branch_mop_2, {|HdlNone, HdlSome: Ty46|}(HdlSome, _match_arm_value_160) @[unit.rs 129:1]
                                L2RegisterFile(_match_arm_value_161):
                                    skip
                                LoadStore(_match_arm_value_162):
                                    skip
                            connect and_then_out_2, alu_branch_mop_2 @[reg_alloc.rs 357:25]
                    match and_then_out_2: @[reg_alloc.rs 356:21]
                        HdlNone:
                            wire _uninit_expr_16: Ty46
                            invalidate _uninit_expr_16
                            connect unit_1.input_insn.data, {|HdlNone, HdlSome: Ty46|}(HdlSome, _uninit_expr_16) @[reg_alloc.rs 361:25]
                        HdlSome(_match_arm_value_163):
                            connect unit_1.input_insn.data, {|HdlNone, HdlSome: Ty46|}(HdlSome, _match_arm_value_163) @[reg_alloc.rs 359:25]
                    match unit_1_free_regs_tracker.alloc_out[0].data: @[reg_alloc.rs 368:21]
                        HdlNone:
                            skip
                        HdlSome(_match_arm_value_164):
                            wire _bundle_literal_expr_214: Ty25
                            wire _bundle_literal_expr_215: Ty23
                            connect _bundle_literal_expr_215.adj_value, tail(UInt<64>(0h2), 62)
                            connect _bundle_literal_expr_214.unit_num, _bundle_literal_expr_215
                            wire _bundle_literal_expr_216: Ty24
                            connect _bundle_literal_expr_216.value, _match_arm_value_164
                            connect _bundle_literal_expr_214.unit_out_reg, _bundle_literal_expr_216
                            connect renamed_mops_out_reg[0], {|HdlNone, HdlSome: Ty25|}(HdlSome, _bundle_literal_expr_214) @[reg_alloc.rs 374:25]
        match unit_1_free_regs_tracker.alloc_out[0].data: @[reg_alloc.rs 341:13]
            HdlNone:
                connect available_units[1][1], UInt<1>(0h0) @[reg_alloc.rs 343:17]
            HdlSome(_match_arm_value_165):
                skip
        when not(unit_1.input_insn.ready): @[reg_alloc.rs 346:13]
            connect available_units[1][1], UInt<1>(0h0) @[reg_alloc.rs 348:17]
        match selected_unit_indexes[1]: @[reg_alloc.rs 351:13]
            HdlNone:
                skip
            HdlSome(_match_arm_value_166):
                when eq(_match_arm_value_166, UInt<64>(0h1)): @[reg_alloc.rs 353:17]
                    wire and_then_out_3: Ty62 @[reg_alloc.rs 357:25]
                    connect unit_1_free_regs_tracker.alloc_out[0].ready, UInt<1>(0h1) @[reg_alloc.rs 354:21]
                    match renamed_mops[1]: @[reg_alloc.rs 357:25]
                        HdlNone:
                            connect and_then_out_3, {|HdlNone, HdlSome: Ty46|}(HdlNone) @[reg_alloc.rs 357:25]
                        HdlSome(_match_arm_value_167):
                            wire alu_branch_mop_3: Ty62 @[unit.rs 129:1]
                            connect alu_branch_mop_3, {|HdlNone, HdlSome: Ty46|}(HdlNone) @[unit.rs 129:1]
                            match _match_arm_value_167: @[unit.rs 129:1]
                                AluBranch(_match_arm_value_168):
                                    connect alu_branch_mop_3, {|HdlNone, HdlSome: Ty46|}(HdlSome, _match_arm_value_168) @[unit.rs 129:1]
                                L2RegisterFile(_match_arm_value_169):
                                    skip
                                LoadStore(_match_arm_value_170):
                                    skip
                            connect and_then_out_3, alu_branch_mop_3 @[reg_alloc.rs 357:25]
                    match and_then_out_3: @[reg_alloc.rs 356:21]
                        HdlNone:
                            wire _uninit_expr_17: Ty46
                            invalidate _uninit_expr_17
                            connect unit_1.input_insn.data, {|HdlNone, HdlSome: Ty46|}(HdlSome, _uninit_expr_17) @[reg_alloc.rs 361:25]
                        HdlSome(_match_arm_value_171):
                            connect unit_1.input_insn.data, {|HdlNone, HdlSome: Ty46|}(HdlSome, _match_arm_value_171) @[reg_alloc.rs 359:25]
                    match unit_1_free_regs_tracker.alloc_out[0].data: @[reg_alloc.rs 368:21]
                        HdlNone:
                            skip
                        HdlSome(_match_arm_value_172):
                            wire _bundle_literal_expr_217: Ty25
                            wire _bundle_literal_expr_218: Ty23
                            connect _bundle_literal_expr_218.adj_value, tail(UInt<64>(0h2), 62)
                            connect _bundle_literal_expr_217.unit_num, _bundle_literal_expr_218
                            wire _bundle_literal_expr_219: Ty24
                            connect _bundle_literal_expr_219.value, _match_arm_value_172
                            connect _bundle_literal_expr_217.unit_out_reg, _bundle_literal_expr_219
                            connect renamed_mops_out_reg[1], {|HdlNone, HdlSome: Ty25|}(HdlSome, _bundle_literal_expr_217) @[reg_alloc.rs 374:25]
        wire _bundle_literal_expr_220: Ty68
        wire _array_literal_expr_76: Ty67[2]
        connect _array_literal_expr_76[0], {|HdlNone, HdlSome: Ty66|}(HdlNone)
        connect _array_literal_expr_76[1], {|HdlNone, HdlSome: Ty66|}(HdlNone)
        connect _bundle_literal_expr_220.unit_output_writes, _array_literal_expr_76
        wire _bundle_literal_expr_221: Ty2
        invalidate _bundle_literal_expr_221
        connect _bundle_literal_expr_220._phantom, _bundle_literal_expr_221
        connect unit_1.unit_forwarding_info, _bundle_literal_expr_220 @[reg_alloc.rs 389:9]
        connect unit_1.`output`.ready, UInt<1>(0h0) @[reg_alloc.rs 400:9]
        connect unit_1.cancel_input.data, {|HdlNone, HdlSome: Ty69|}(HdlNone) @[reg_alloc.rs 402:9]
    module alu_branch: @[alu_branch.rs 18:1]
        input cd: Ty0 @[alu_branch.rs 21:29]
        input input_insn: Ty63 @[alu_branch.rs 24:11]
        input unit_forwarding_info: Ty68 @[alu_branch.rs 27:11]
        input cancel_input: Ty71 @[alu_branch.rs 30:11]
        output `output`: Ty76 @[alu_branch.rs 33:11]
        inst unit_base of unit_base @[alu_branch.rs 35:21]
        connect unit_base.input_insn, input_insn @[alu_branch.rs 40:5]
        connect unit_base.cd, cd @[alu_branch.rs 41:5]
        connect unit_base.unit_forwarding_info, unit_forwarding_info @[alu_branch.rs 42:5]
        connect unit_base.cancel_input, cancel_input @[alu_branch.rs 43:5]
        connect unit_base.ready_mop.ready, UInt<1>(0h1) @[alu_branch.rs 45:5]
        connect `output`.data, {|HdlNone, HdlSome: Ty74|}(HdlNone) @[alu_branch.rs 46:5]
    module unit_free_regs_tracker: @[unit_free_regs_tracker.rs 7:1]
        input cd: Ty0 @[unit_free_regs_tracker.rs 14:29]
        input free_in: Ty79[1] @[unit_free_regs_tracker.rs 17:11]
        output alloc_out: Ty79[1] @[unit_free_regs_tracker.rs 20:11]
        wire _array_literal_expr: UInt<1>[16]
        connect _array_literal_expr[0], UInt<1>(0h0)
        connect _array_literal_expr[1], UInt<1>(0h0)
        connect _array_literal_expr[2], UInt<1>(0h0)
        connect _array_literal_expr[3], UInt<1>(0h0)
        connect _array_literal_expr[4], UInt<1>(0h0)
        connect _array_literal_expr[5], UInt<1>(0h0)
        connect _array_literal_expr[6], UInt<1>(0h0)
        connect _array_literal_expr[7], UInt<1>(0h0)
        connect _array_literal_expr[8], UInt<1>(0h0)
        connect _array_literal_expr[9], UInt<1>(0h0)
        connect _array_literal_expr[10], UInt<1>(0h0)
        connect _array_literal_expr[11], UInt<1>(0h0)
        connect _array_literal_expr[12], UInt<1>(0h0)
        connect _array_literal_expr[13], UInt<1>(0h0)
        connect _array_literal_expr[14], UInt<1>(0h0)
        connect _array_literal_expr[15], UInt<1>(0h0)
        regreset allocated_reg: UInt<1>[16], cd.clk, cd.rst, _array_literal_expr @[unit_free_regs_tracker.rs 27:25]
        connect free_in[0].ready, UInt<1>(0h1) @[unit_free_regs_tracker.rs 29:9]
        wire firing_data: Ty78 @[ready_valid.rs 30:27]
        connect firing_data, {|HdlNone, HdlSome: UInt<4>|}(HdlNone) @[ready_valid.rs 31:9]
        when free_in[0].ready: @[ready_valid.rs 33:9]
            connect firing_data, free_in[0].data @[ready_valid.rs 34:13]
        match firing_data: @[unit_free_regs_tracker.rs 31:9]
            HdlNone:
                skip
            HdlSome(_match_arm_value):
                connect allocated_reg[_match_arm_value], UInt<1>(0h0) @[unit_free_regs_tracker.rs 32:13]
        wire reduced_count_0_2: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_0_2, add(not(allocated_reg[0]), not(allocated_reg[1])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_0_2: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_0_2, or(or(neq(add(not(allocated_reg[0]), not(allocated_reg[1])), reduced_count_0_2), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_0_2: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[0]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_1: UInt<0>[1]
            connect _array_literal_expr_1[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_0_2[0], _array_literal_expr_1[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_2: UInt<0>[1]
            connect _array_literal_expr_2[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_0_2[0], add(_array_literal_expr_2[sub(UInt<64>(0h0), not(allocated_reg[0]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_2_4: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_2_4, add(not(allocated_reg[2]), not(allocated_reg[3])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_2_4: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_2_4, or(or(neq(add(not(allocated_reg[2]), not(allocated_reg[3])), reduced_count_2_4), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_2_4: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[2]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_3: UInt<0>[1]
            connect _array_literal_expr_3[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_2_4[0], _array_literal_expr_3[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_4: UInt<0>[1]
            connect _array_literal_expr_4[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_2_4[0], add(_array_literal_expr_4[sub(UInt<64>(0h0), not(allocated_reg[2]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_0_4: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_0_4, add(reduced_count_0_2, reduced_count_2_4) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_0_4: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_0_4, or(or(neq(add(reduced_count_0_2, reduced_count_2_4), reduced_count_0_4), reduced_count_overflowed_0_2), reduced_count_overflowed_2_4) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_0_4: UInt<2>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_0_2, gt(reduced_count_0_2, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<1>
            connect reduced_alloc_nums_0_4[0], reduced_alloc_nums_0_2[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_0_4[0], add(reduced_alloc_nums_2_4[sub(UInt<64>(0h0), reduced_count_0_2)], UInt<64>(0h2)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_4_6: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_4_6, add(not(allocated_reg[4]), not(allocated_reg[5])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_4_6: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_4_6, or(or(neq(add(not(allocated_reg[4]), not(allocated_reg[5])), reduced_count_4_6), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_4_6: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[4]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_5: UInt<0>[1]
            connect _array_literal_expr_5[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_4_6[0], _array_literal_expr_5[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_6: UInt<0>[1]
            connect _array_literal_expr_6[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_4_6[0], add(_array_literal_expr_6[sub(UInt<64>(0h0), not(allocated_reg[4]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_6_8: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_6_8, add(not(allocated_reg[6]), not(allocated_reg[7])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_6_8: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_6_8, or(or(neq(add(not(allocated_reg[6]), not(allocated_reg[7])), reduced_count_6_8), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_6_8: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[6]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_7: UInt<0>[1]
            connect _array_literal_expr_7[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_6_8[0], _array_literal_expr_7[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_8: UInt<0>[1]
            connect _array_literal_expr_8[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_6_8[0], add(_array_literal_expr_8[sub(UInt<64>(0h0), not(allocated_reg[6]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_4_8: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_4_8, add(reduced_count_4_6, reduced_count_6_8) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_4_8: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_4_8, or(or(neq(add(reduced_count_4_6, reduced_count_6_8), reduced_count_4_8), reduced_count_overflowed_4_6), reduced_count_overflowed_6_8) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_4_8: UInt<2>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_4_6, gt(reduced_count_4_6, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<1>
            connect reduced_alloc_nums_4_8[0], reduced_alloc_nums_4_6[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_4_8[0], add(reduced_alloc_nums_6_8[sub(UInt<64>(0h0), reduced_count_4_6)], UInt<64>(0h2)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_0_8: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_0_8, add(reduced_count_0_4, reduced_count_4_8) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_0_8: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_0_8, or(or(neq(add(reduced_count_0_4, reduced_count_4_8), reduced_count_0_8), reduced_count_overflowed_0_4), reduced_count_overflowed_4_8) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_0_8: UInt<3>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_0_4, gt(reduced_count_0_4, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<3>
            ; rhs: UInt<2>
            connect reduced_alloc_nums_0_8[0], reduced_alloc_nums_0_4[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<3>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_0_8[0], add(reduced_alloc_nums_4_8[sub(UInt<64>(0h0), reduced_count_0_4)], UInt<64>(0h4)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_8_10: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_8_10, add(not(allocated_reg[8]), not(allocated_reg[9])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_8_10: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_8_10, or(or(neq(add(not(allocated_reg[8]), not(allocated_reg[9])), reduced_count_8_10), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_8_10: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[8]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_9: UInt<0>[1]
            connect _array_literal_expr_9[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_8_10[0], _array_literal_expr_9[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_10: UInt<0>[1]
            connect _array_literal_expr_10[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_8_10[0], add(_array_literal_expr_10[sub(UInt<64>(0h0), not(allocated_reg[8]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_10_12: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_10_12, add(not(allocated_reg[10]), not(allocated_reg[11])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_10_12: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_10_12, or(or(neq(add(not(allocated_reg[10]), not(allocated_reg[11])), reduced_count_10_12), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_10_12: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[10]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_11: UInt<0>[1]
            connect _array_literal_expr_11[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_10_12[0], _array_literal_expr_11[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_12: UInt<0>[1]
            connect _array_literal_expr_12[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_10_12[0], add(_array_literal_expr_12[sub(UInt<64>(0h0), not(allocated_reg[10]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_8_12: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_8_12, add(reduced_count_8_10, reduced_count_10_12) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_8_12: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_8_12, or(or(neq(add(reduced_count_8_10, reduced_count_10_12), reduced_count_8_12), reduced_count_overflowed_8_10), reduced_count_overflowed_10_12) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_8_12: UInt<2>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_8_10, gt(reduced_count_8_10, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<1>
            connect reduced_alloc_nums_8_12[0], reduced_alloc_nums_8_10[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_8_12[0], add(reduced_alloc_nums_10_12[sub(UInt<64>(0h0), reduced_count_8_10)], UInt<64>(0h2)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_12_14: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_12_14, add(not(allocated_reg[12]), not(allocated_reg[13])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_12_14: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_12_14, or(or(neq(add(not(allocated_reg[12]), not(allocated_reg[13])), reduced_count_12_14), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_12_14: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[12]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_13: UInt<0>[1]
            connect _array_literal_expr_13[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_12_14[0], _array_literal_expr_13[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_14: UInt<0>[1]
            connect _array_literal_expr_14[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_12_14[0], add(_array_literal_expr_14[sub(UInt<64>(0h0), not(allocated_reg[12]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_14_16: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_14_16, add(not(allocated_reg[14]), not(allocated_reg[15])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_14_16: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_14_16, or(or(neq(add(not(allocated_reg[14]), not(allocated_reg[15])), reduced_count_14_16), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_14_16: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[14]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_15: UInt<0>[1]
            connect _array_literal_expr_15[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_14_16[0], _array_literal_expr_15[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_16: UInt<0>[1]
            connect _array_literal_expr_16[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_14_16[0], add(_array_literal_expr_16[sub(UInt<64>(0h0), not(allocated_reg[14]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_12_16: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_12_16, add(reduced_count_12_14, reduced_count_14_16) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_12_16: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_12_16, or(or(neq(add(reduced_count_12_14, reduced_count_14_16), reduced_count_12_16), reduced_count_overflowed_12_14), reduced_count_overflowed_14_16) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_12_16: UInt<2>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_12_14, gt(reduced_count_12_14, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<1>
            connect reduced_alloc_nums_12_16[0], reduced_alloc_nums_12_14[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_12_16[0], add(reduced_alloc_nums_14_16[sub(UInt<64>(0h0), reduced_count_12_14)], UInt<64>(0h2)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_8_16: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_8_16, add(reduced_count_8_12, reduced_count_12_16) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_8_16: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_8_16, or(or(neq(add(reduced_count_8_12, reduced_count_12_16), reduced_count_8_16), reduced_count_overflowed_8_12), reduced_count_overflowed_12_16) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_8_16: UInt<3>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_8_12, gt(reduced_count_8_12, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<3>
            ; rhs: UInt<2>
            connect reduced_alloc_nums_8_16[0], reduced_alloc_nums_8_12[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<3>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_8_16[0], add(reduced_alloc_nums_12_16[sub(UInt<64>(0h0), reduced_count_8_12)], UInt<64>(0h4)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_0_16: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_0_16, add(reduced_count_0_8, reduced_count_8_16) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_0_16: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_0_16, or(or(neq(add(reduced_count_0_8, reduced_count_8_16), reduced_count_0_16), reduced_count_overflowed_0_8), reduced_count_overflowed_8_16) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_0_16: UInt<4>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_0_8, gt(reduced_count_0_8, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<4>
            ; rhs: UInt<3>
            connect reduced_alloc_nums_0_16[0], reduced_alloc_nums_0_8[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<4>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_0_16[0], add(reduced_alloc_nums_8_16[sub(UInt<64>(0h0), reduced_count_0_8)], UInt<64>(0h8)) @[unit_free_regs_tracker.rs 83:21]
        wire firing_data_1: Ty78 @[ready_valid.rs 30:27]
        connect firing_data_1, {|HdlNone, HdlSome: UInt<4>|}(HdlNone) @[ready_valid.rs 31:9]
        when alloc_out[0].ready: @[ready_valid.rs 33:9]
            connect firing_data_1, alloc_out[0].data @[ready_valid.rs 34:13]
        match firing_data_1: @[unit_free_regs_tracker.rs 102:9]
            HdlNone:
                skip
            HdlSome(_match_arm_value_1):
                connect allocated_reg[_match_arm_value_1], UInt<1>(0h1) @[unit_free_regs_tracker.rs 103:13]
        when or(reduced_count_overflowed_0_16, gt(reduced_count_0_16, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 106:9]
            connect alloc_out[0].data, {|HdlNone, HdlSome: UInt<4>|}(HdlSome, reduced_alloc_nums_0_16[0]) @[unit_free_regs_tracker.rs 107:13]
        else:
            connect alloc_out[0].data, {|HdlNone, HdlSome: UInt<4>|}(HdlNone) @[unit_free_regs_tracker.rs 112:13]
    module alu_branch_1: @[alu_branch.rs 18:1]
        input cd: Ty0 @[alu_branch.rs 21:29]
        input input_insn: Ty63 @[alu_branch.rs 24:11]
        input unit_forwarding_info: Ty68 @[alu_branch.rs 27:11]
        input cancel_input: Ty71 @[alu_branch.rs 30:11]
        output `output`: Ty76 @[alu_branch.rs 33:11]
        inst unit_base of unit_base_1 @[alu_branch.rs 35:21]
        connect unit_base.input_insn, input_insn @[alu_branch.rs 40:5]
        connect unit_base.cd, cd @[alu_branch.rs 41:5]
        connect unit_base.unit_forwarding_info, unit_forwarding_info @[alu_branch.rs 42:5]
        connect unit_base.cancel_input, cancel_input @[alu_branch.rs 43:5]
        connect unit_base.ready_mop.ready, UInt<1>(0h1) @[alu_branch.rs 45:5]
        connect `output`.data, {|HdlNone, HdlSome: Ty74|}(HdlNone) @[alu_branch.rs 46:5]
    module unit_free_regs_tracker_1: @[unit_free_regs_tracker.rs 7:1]
        input cd: Ty0 @[unit_free_regs_tracker.rs 14:29]
        input free_in: Ty79[1] @[unit_free_regs_tracker.rs 17:11]
        output alloc_out: Ty79[1] @[unit_free_regs_tracker.rs 20:11]
        wire _array_literal_expr: UInt<1>[16]
        connect _array_literal_expr[0], UInt<1>(0h0)
        connect _array_literal_expr[1], UInt<1>(0h0)
        connect _array_literal_expr[2], UInt<1>(0h0)
        connect _array_literal_expr[3], UInt<1>(0h0)
        connect _array_literal_expr[4], UInt<1>(0h0)
        connect _array_literal_expr[5], UInt<1>(0h0)
        connect _array_literal_expr[6], UInt<1>(0h0)
        connect _array_literal_expr[7], UInt<1>(0h0)
        connect _array_literal_expr[8], UInt<1>(0h0)
        connect _array_literal_expr[9], UInt<1>(0h0)
        connect _array_literal_expr[10], UInt<1>(0h0)
        connect _array_literal_expr[11], UInt<1>(0h0)
        connect _array_literal_expr[12], UInt<1>(0h0)
        connect _array_literal_expr[13], UInt<1>(0h0)
        connect _array_literal_expr[14], UInt<1>(0h0)
        connect _array_literal_expr[15], UInt<1>(0h0)
        regreset allocated_reg: UInt<1>[16], cd.clk, cd.rst, _array_literal_expr @[unit_free_regs_tracker.rs 27:25]
        connect free_in[0].ready, UInt<1>(0h1) @[unit_free_regs_tracker.rs 29:9]
        wire firing_data: Ty78 @[ready_valid.rs 30:27]
        connect firing_data, {|HdlNone, HdlSome: UInt<4>|}(HdlNone) @[ready_valid.rs 31:9]
        when free_in[0].ready: @[ready_valid.rs 33:9]
            connect firing_data, free_in[0].data @[ready_valid.rs 34:13]
        match firing_data: @[unit_free_regs_tracker.rs 31:9]
            HdlNone:
                skip
            HdlSome(_match_arm_value):
                connect allocated_reg[_match_arm_value], UInt<1>(0h0) @[unit_free_regs_tracker.rs 32:13]
        wire reduced_count_0_2: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_0_2, add(not(allocated_reg[0]), not(allocated_reg[1])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_0_2: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_0_2, or(or(neq(add(not(allocated_reg[0]), not(allocated_reg[1])), reduced_count_0_2), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_0_2: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[0]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_1: UInt<0>[1]
            connect _array_literal_expr_1[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_0_2[0], _array_literal_expr_1[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_2: UInt<0>[1]
            connect _array_literal_expr_2[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_0_2[0], add(_array_literal_expr_2[sub(UInt<64>(0h0), not(allocated_reg[0]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_2_4: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_2_4, add(not(allocated_reg[2]), not(allocated_reg[3])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_2_4: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_2_4, or(or(neq(add(not(allocated_reg[2]), not(allocated_reg[3])), reduced_count_2_4), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_2_4: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[2]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_3: UInt<0>[1]
            connect _array_literal_expr_3[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_2_4[0], _array_literal_expr_3[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_4: UInt<0>[1]
            connect _array_literal_expr_4[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_2_4[0], add(_array_literal_expr_4[sub(UInt<64>(0h0), not(allocated_reg[2]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_0_4: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_0_4, add(reduced_count_0_2, reduced_count_2_4) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_0_4: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_0_4, or(or(neq(add(reduced_count_0_2, reduced_count_2_4), reduced_count_0_4), reduced_count_overflowed_0_2), reduced_count_overflowed_2_4) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_0_4: UInt<2>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_0_2, gt(reduced_count_0_2, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<1>
            connect reduced_alloc_nums_0_4[0], reduced_alloc_nums_0_2[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_0_4[0], add(reduced_alloc_nums_2_4[sub(UInt<64>(0h0), reduced_count_0_2)], UInt<64>(0h2)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_4_6: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_4_6, add(not(allocated_reg[4]), not(allocated_reg[5])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_4_6: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_4_6, or(or(neq(add(not(allocated_reg[4]), not(allocated_reg[5])), reduced_count_4_6), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_4_6: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[4]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_5: UInt<0>[1]
            connect _array_literal_expr_5[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_4_6[0], _array_literal_expr_5[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_6: UInt<0>[1]
            connect _array_literal_expr_6[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_4_6[0], add(_array_literal_expr_6[sub(UInt<64>(0h0), not(allocated_reg[4]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_6_8: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_6_8, add(not(allocated_reg[6]), not(allocated_reg[7])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_6_8: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_6_8, or(or(neq(add(not(allocated_reg[6]), not(allocated_reg[7])), reduced_count_6_8), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_6_8: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[6]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_7: UInt<0>[1]
            connect _array_literal_expr_7[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_6_8[0], _array_literal_expr_7[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_8: UInt<0>[1]
            connect _array_literal_expr_8[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_6_8[0], add(_array_literal_expr_8[sub(UInt<64>(0h0), not(allocated_reg[6]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_4_8: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_4_8, add(reduced_count_4_6, reduced_count_6_8) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_4_8: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_4_8, or(or(neq(add(reduced_count_4_6, reduced_count_6_8), reduced_count_4_8), reduced_count_overflowed_4_6), reduced_count_overflowed_6_8) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_4_8: UInt<2>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_4_6, gt(reduced_count_4_6, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<1>
            connect reduced_alloc_nums_4_8[0], reduced_alloc_nums_4_6[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_4_8[0], add(reduced_alloc_nums_6_8[sub(UInt<64>(0h0), reduced_count_4_6)], UInt<64>(0h2)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_0_8: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_0_8, add(reduced_count_0_4, reduced_count_4_8) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_0_8: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_0_8, or(or(neq(add(reduced_count_0_4, reduced_count_4_8), reduced_count_0_8), reduced_count_overflowed_0_4), reduced_count_overflowed_4_8) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_0_8: UInt<3>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_0_4, gt(reduced_count_0_4, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<3>
            ; rhs: UInt<2>
            connect reduced_alloc_nums_0_8[0], reduced_alloc_nums_0_4[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<3>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_0_8[0], add(reduced_alloc_nums_4_8[sub(UInt<64>(0h0), reduced_count_0_4)], UInt<64>(0h4)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_8_10: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_8_10, add(not(allocated_reg[8]), not(allocated_reg[9])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_8_10: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_8_10, or(or(neq(add(not(allocated_reg[8]), not(allocated_reg[9])), reduced_count_8_10), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_8_10: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[8]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_9: UInt<0>[1]
            connect _array_literal_expr_9[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_8_10[0], _array_literal_expr_9[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_10: UInt<0>[1]
            connect _array_literal_expr_10[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_8_10[0], add(_array_literal_expr_10[sub(UInt<64>(0h0), not(allocated_reg[8]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_10_12: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_10_12, add(not(allocated_reg[10]), not(allocated_reg[11])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_10_12: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_10_12, or(or(neq(add(not(allocated_reg[10]), not(allocated_reg[11])), reduced_count_10_12), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_10_12: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[10]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_11: UInt<0>[1]
            connect _array_literal_expr_11[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_10_12[0], _array_literal_expr_11[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_12: UInt<0>[1]
            connect _array_literal_expr_12[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_10_12[0], add(_array_literal_expr_12[sub(UInt<64>(0h0), not(allocated_reg[10]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_8_12: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_8_12, add(reduced_count_8_10, reduced_count_10_12) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_8_12: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_8_12, or(or(neq(add(reduced_count_8_10, reduced_count_10_12), reduced_count_8_12), reduced_count_overflowed_8_10), reduced_count_overflowed_10_12) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_8_12: UInt<2>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_8_10, gt(reduced_count_8_10, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<1>
            connect reduced_alloc_nums_8_12[0], reduced_alloc_nums_8_10[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_8_12[0], add(reduced_alloc_nums_10_12[sub(UInt<64>(0h0), reduced_count_8_10)], UInt<64>(0h2)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_12_14: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_12_14, add(not(allocated_reg[12]), not(allocated_reg[13])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_12_14: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_12_14, or(or(neq(add(not(allocated_reg[12]), not(allocated_reg[13])), reduced_count_12_14), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_12_14: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[12]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_13: UInt<0>[1]
            connect _array_literal_expr_13[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_12_14[0], _array_literal_expr_13[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_14: UInt<0>[1]
            connect _array_literal_expr_14[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_12_14[0], add(_array_literal_expr_14[sub(UInt<64>(0h0), not(allocated_reg[12]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_14_16: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_14_16, add(not(allocated_reg[14]), not(allocated_reg[15])) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_14_16: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_14_16, or(or(neq(add(not(allocated_reg[14]), not(allocated_reg[15])), reduced_count_14_16), UInt<1>(0h0)), UInt<1>(0h0)) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_14_16: UInt<1>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(UInt<1>(0h0), gt(not(allocated_reg[14]), UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            wire _array_literal_expr_15: UInt<0>[1]
            connect _array_literal_expr_15[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<0>
            connect reduced_alloc_nums_14_16[0], _array_literal_expr_15[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            wire _array_literal_expr_16: UInt<0>[1]
            connect _array_literal_expr_16[0], UInt<0>(0h0)
            ; connect different types:
            ; lhs: UInt<1>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_14_16[0], add(_array_literal_expr_16[sub(UInt<64>(0h0), not(allocated_reg[14]))], UInt<64>(0h1)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_12_16: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_12_16, add(reduced_count_12_14, reduced_count_14_16) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_12_16: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_12_16, or(or(neq(add(reduced_count_12_14, reduced_count_14_16), reduced_count_12_16), reduced_count_overflowed_12_14), reduced_count_overflowed_14_16) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_12_16: UInt<2>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_12_14, gt(reduced_count_12_14, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<1>
            connect reduced_alloc_nums_12_16[0], reduced_alloc_nums_12_14[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<2>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_12_16[0], add(reduced_alloc_nums_14_16[sub(UInt<64>(0h0), reduced_count_12_14)], UInt<64>(0h2)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_8_16: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_8_16, add(reduced_count_8_12, reduced_count_12_16) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_8_16: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_8_16, or(or(neq(add(reduced_count_8_12, reduced_count_12_16), reduced_count_8_16), reduced_count_overflowed_8_12), reduced_count_overflowed_12_16) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_8_16: UInt<3>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_8_12, gt(reduced_count_8_12, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<3>
            ; rhs: UInt<2>
            connect reduced_alloc_nums_8_16[0], reduced_alloc_nums_8_12[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<3>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_8_16[0], add(reduced_alloc_nums_12_16[sub(UInt<64>(0h0), reduced_count_8_12)], UInt<64>(0h4)) @[unit_free_regs_tracker.rs 83:21]
        wire reduced_count_0_16: UInt<1> @[unit_free_regs_tracker.rs 60:17]
        ; connect different types:
        ; lhs: UInt<1>
        ; rhs: UInt<2>
        connect reduced_count_0_16, add(reduced_count_0_8, reduced_count_8_16) @[unit_free_regs_tracker.rs 63:13]
        wire reduced_count_overflowed_0_16: UInt<1> @[unit_free_regs_tracker.rs 66:17]
        connect reduced_count_overflowed_0_16, or(or(neq(add(reduced_count_0_8, reduced_count_8_16), reduced_count_0_16), reduced_count_overflowed_0_8), reduced_count_overflowed_8_16) @[unit_free_regs_tracker.rs 69:13]
        wire reduced_alloc_nums_0_16: UInt<4>[1] @[unit_free_regs_tracker.rs 75:17]
        when or(reduced_count_overflowed_0_8, gt(reduced_count_0_8, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 80:17]
            ; connect different types:
            ; lhs: UInt<4>
            ; rhs: UInt<3>
            connect reduced_alloc_nums_0_16[0], reduced_alloc_nums_0_8[0] @[unit_free_regs_tracker.rs 81:21]
        else:
            ; connect different types:
            ; lhs: UInt<4>
            ; rhs: UInt<65>
            connect reduced_alloc_nums_0_16[0], add(reduced_alloc_nums_8_16[sub(UInt<64>(0h0), reduced_count_0_8)], UInt<64>(0h8)) @[unit_free_regs_tracker.rs 83:21]
        wire firing_data_1: Ty78 @[ready_valid.rs 30:27]
        connect firing_data_1, {|HdlNone, HdlSome: UInt<4>|}(HdlNone) @[ready_valid.rs 31:9]
        when alloc_out[0].ready: @[ready_valid.rs 33:9]
            connect firing_data_1, alloc_out[0].data @[ready_valid.rs 34:13]
        match firing_data_1: @[unit_free_regs_tracker.rs 102:9]
            HdlNone:
                skip
            HdlSome(_match_arm_value_1):
                connect allocated_reg[_match_arm_value_1], UInt<1>(0h1) @[unit_free_regs_tracker.rs 103:13]
        when or(reduced_count_overflowed_0_16, gt(reduced_count_0_16, UInt<64>(0h0))): @[unit_free_regs_tracker.rs 106:9]
            connect alloc_out[0].data, {|HdlNone, HdlSome: UInt<4>|}(HdlSome, reduced_alloc_nums_0_16[0]) @[unit_free_regs_tracker.rs 107:13]
        else:
            connect alloc_out[0].data, {|HdlNone, HdlSome: UInt<4>|}(HdlNone) @[unit_free_regs_tracker.rs 112:13]
    module unit_base: @[unit_base.rs 32:1]
        input cd: Ty0 @[unit_base.rs 39:29]
        input unit_forwarding_info: Ty68 @[unit_base.rs 42:11]
        input input_insn: Ty63 @[unit_base.rs 44:41]
        input cancel_input: Ty71 @[unit_base.rs 48:11]
        output ready_mop: Ty83 @[unit_base.rs 51:50]
        wire and_then_out: Ty85 @[unit_base.rs 66:17]
        wire and_then_out_1: Ty85 @[unit_base.rs 66:17]
        wire and_then_out_2: Ty85 @[unit_base.rs 66:17]
        wire and_then_out_3: Ty85 @[unit_base.rs 66:17]
        wire and_then_out_4: Ty85 @[unit_base.rs 66:17]
        wire and_then_out_5: Ty85 @[unit_base.rs 66:17]
        wire and_then_out_6: Ty85 @[unit_base.rs 66:17]
        wire and_then_out_7: Ty85 @[unit_base.rs 66:17]
        connect input_insn.ready, UInt<1>(0h0) @[unit_base.rs 45:5]
        connect cancel_input.ready, UInt<1>(0h1) @[unit_base.rs 49:5]
        connect ready_mop.data, {|HdlNone, HdlSome: Ty81|}(HdlNone) @[unit_base.rs 52:5]
        wire _array_literal_expr: Ty88[8]
        connect _array_literal_expr[0], {|HdlNone, HdlSome: Ty87|}(HdlNone)
        connect _array_literal_expr[1], {|HdlNone, HdlSome: Ty87|}(HdlNone)
        connect _array_literal_expr[2], {|HdlNone, HdlSome: Ty87|}(HdlNone)
        connect _array_literal_expr[3], {|HdlNone, HdlSome: Ty87|}(HdlNone)
        connect _array_literal_expr[4], {|HdlNone, HdlSome: Ty87|}(HdlNone)
        connect _array_literal_expr[5], {|HdlNone, HdlSome: Ty87|}(HdlNone)
        connect _array_literal_expr[6], {|HdlNone, HdlSome: Ty87|}(HdlNone)
        connect _array_literal_expr[7], {|HdlNone, HdlSome: Ty87|}(HdlNone)
        regreset in_flight_ops: Ty88[8], cd.clk, cd.rst, _array_literal_expr @[unit_base.rs 55:25]
        wire input_index: Ty85 @[unit_base.rs 61:23]
        match in_flight_ops[0]: @[unit_base.rs 66:17]
            HdlNone:
                connect and_then_out, {|HdlNone, HdlSome: UInt<3>|}(HdlNone) @[unit_base.rs 66:17]
            HdlSome(_match_arm_value):
                connect and_then_out, {|HdlNone, HdlSome: UInt<3>|}(HdlSome, tail(UInt<64>(0h0), 61)) @[unit_base.rs 66:17]
        match in_flight_ops[1]: @[unit_base.rs 66:17]
            HdlNone:
                connect and_then_out_1, {|HdlNone, HdlSome: UInt<3>|}(HdlNone) @[unit_base.rs 66:17]
            HdlSome(_match_arm_value_1):
                connect and_then_out_1, {|HdlNone, HdlSome: UInt<3>|}(HdlSome, tail(UInt<64>(0h1), 61)) @[unit_base.rs 66:17]
        wire or_out: Ty85 @[function.rs 166:5]
        connect or_out, and_then_out_1 @[function.rs 166:5]
        match and_then_out: @[function.rs 166:5]
            HdlNone:
                skip
            HdlSome(_match_arm_value_2):
                connect or_out, and_then_out @[function.rs 166:5]
        match in_flight_ops[2]: @[unit_base.rs 66:17]
            HdlNone:
                connect and_then_out_2, {|HdlNone, HdlSome: UInt<3>|}(HdlNone) @[unit_base.rs 66:17]
            HdlSome(_match_arm_value_3):
                connect and_then_out_2, {|HdlNone, HdlSome: UInt<3>|}(HdlSome, tail(UInt<64>(0h2), 61)) @[unit_base.rs 66:17]
        match in_flight_ops[3]: @[unit_base.rs 66:17]
            HdlNone:
                connect and_then_out_3, {|HdlNone, HdlSome: UInt<3>|}(HdlNone) @[unit_base.rs 66:17]
            HdlSome(_match_arm_value_4):
                connect and_then_out_3, {|HdlNone, HdlSome: UInt<3>|}(HdlSome, tail(UInt<64>(0h3), 61)) @[unit_base.rs 66:17]
        wire or_out_1: Ty85 @[function.rs 166:5]
        connect or_out_1, and_then_out_3 @[function.rs 166:5]
        match and_then_out_2: @[function.rs 166:5]
            HdlNone:
                skip
            HdlSome(_match_arm_value_5):
                connect or_out_1, and_then_out_2 @[function.rs 166:5]
        wire or_out_2: Ty85 @[function.rs 166:5]
        connect or_out_2, or_out_1 @[function.rs 166:5]
        match or_out: @[function.rs 166:5]
            HdlNone:
                skip
            HdlSome(_match_arm_value_6):
                connect or_out_2, or_out @[function.rs 166:5]
        match in_flight_ops[4]: @[unit_base.rs 66:17]
            HdlNone:
                connect and_then_out_4, {|HdlNone, HdlSome: UInt<3>|}(HdlNone) @[unit_base.rs 66:17]
            HdlSome(_match_arm_value_7):
                connect and_then_out_4, {|HdlNone, HdlSome: UInt<3>|}(HdlSome, tail(UInt<64>(0h4), 61)) @[unit_base.rs 66:17]
        match in_flight_ops[5]: @[unit_base.rs 66:17]
            HdlNone:
                connect and_then_out_5, {|HdlNone, HdlSome: UInt<3>|}(HdlNone) @[unit_base.rs 66:17]
            HdlSome(_match_arm_value_8):
                connect and_then_out_5, {|HdlNone, HdlSome: UInt<3>|}(HdlSome, tail(UInt<64>(0h5), 61)) @[unit_base.rs 66:17]
        wire or_out_3: Ty85 @[function.rs 166:5]
        connect or_out_3, and_then_out_5 @[function.rs 166:5]
        match and_then_out_4: @[function.rs 166:5]
            HdlNone:
                skip
            HdlSome(_match_arm_value_9):
                connect or_out_3, and_then_out_4 @[function.rs 166:5]
        match in_flight_ops[6]: @[unit_base.rs 66:17]
            HdlNone:
                connect and_then_out_6, {|HdlNone, HdlSome: UInt<3>|}(HdlNone) @[unit_base.rs 66:17]
            HdlSome(_match_arm_value_10):
                connect and_then_out_6, {|HdlNone, HdlSome: UInt<3>|}(HdlSome, tail(UInt<64>(0h6), 61)) @[unit_base.rs 66:17]
        match in_flight_ops[7]: @[unit_base.rs 66:17]
            HdlNone:
                connect and_then_out_7, {|HdlNone, HdlSome: UInt<3>|}(HdlNone) @[unit_base.rs 66:17]
            HdlSome(_match_arm_value_11):
                connect and_then_out_7, {|HdlNone, HdlSome: UInt<3>|}(HdlSome, tail(UInt<64>(0h7), 61)) @[unit_base.rs 66:17]
        wire or_out_4: Ty85 @[function.rs 166:5]
        connect or_out_4, and_then_out_7 @[function.rs 166:5]
        match and_then_out_6: @[function.rs 166:5]
            HdlNone:
                skip
            HdlSome(_match_arm_value_12):
                connect or_out_4, and_then_out_6 @[function.rs 166:5]
        wire or_out_5: Ty85 @[function.rs 166:5]
        connect or_out_5, or_out_4 @[function.rs 166:5]
        match or_out_3: @[function.rs 166:5]
            HdlNone:
                skip
            HdlSome(_match_arm_value_13):
                connect or_out_5, or_out_3 @[function.rs 166:5]
        wire or_out_6: Ty85 @[function.rs 166:5]
        connect or_out_6, or_out_5 @[function.rs 166:5]
        match or_out_2: @[function.rs 166:5]
            HdlNone:
                skip
            HdlSome(_match_arm_value_14):
                connect or_out_6, or_out_2 @[function.rs 166:5]
        connect input_index, or_out_6 @[unit_base.rs 62:5]
        wire input_in_flight_op: Ty88 @[unit_base.rs 73:30]
        connect input_in_flight_op, {|HdlNone, HdlSome: Ty87|}(HdlNone) @[unit_base.rs 74:5]
        wire firing_data: Ty62 @[ready_valid.rs 30:27]
        connect firing_data, {|HdlNone, HdlSome: Ty46|}(HdlNone) @[ready_valid.rs 31:9]
        when input_insn.ready: @[ready_valid.rs 33:9]
            connect firing_data, input_insn.data @[ready_valid.rs 34:13]
        match firing_data: @[unit_base.rs 76:5]
            HdlNone:
                skip
            HdlSome(_match_arm_value_15):
                wire input_in_flight_op_src_values: Ty86[3] @[unit_base.rs 79:13]
                wire _array_literal_expr_1: Ty86[3]
                wire _bundle_literal_expr: Ty65
                connect _bundle_literal_expr.int_fp, UInt<64>(0h0)
                wire _bundle_literal_expr_1: Ty64
                connect _bundle_literal_expr_1.pwr_ca_x86_cf, UInt<1>(0h0)
                connect _bundle_literal_expr_1.pwr_ca32_x86_af, UInt<1>(0h0)
                connect _bundle_literal_expr_1.pwr_ov_x86_of, UInt<1>(0h0)
                connect _bundle_literal_expr_1.pwr_ov32_x86_df, UInt<1>(0h0)
                connect _bundle_literal_expr_1.pwr_cr_lt_x86_sf, UInt<1>(0h0)
                connect _bundle_literal_expr_1.pwr_cr_gt_x86_pf, UInt<1>(0h0)
                connect _bundle_literal_expr_1.pwr_cr_eq_x86_zf, UInt<1>(0h0)
                connect _bundle_literal_expr_1.pwr_so, UInt<1>(0h0)
                connect _bundle_literal_expr.flags, _bundle_literal_expr_1
                connect _array_literal_expr_1[0], {|HdlNone, HdlSome: Ty65|}(HdlSome, _bundle_literal_expr)
                wire _bundle_literal_expr_2: Ty65
                connect _bundle_literal_expr_2.int_fp, UInt<64>(0h0)
                wire _bundle_literal_expr_3: Ty64
                connect _bundle_literal_expr_3.pwr_ca_x86_cf, UInt<1>(0h0)
                connect _bundle_literal_expr_3.pwr_ca32_x86_af, UInt<1>(0h0)
                connect _bundle_literal_expr_3.pwr_ov_x86_of, UInt<1>(0h0)
                connect _bundle_literal_expr_3.pwr_ov32_x86_df, UInt<1>(0h0)
                connect _bundle_literal_expr_3.pwr_cr_lt_x86_sf, UInt<1>(0h0)
                connect _bundle_literal_expr_3.pwr_cr_gt_x86_pf, UInt<1>(0h0)
                connect _bundle_literal_expr_3.pwr_cr_eq_x86_zf, UInt<1>(0h0)
                connect _bundle_literal_expr_3.pwr_so, UInt<1>(0h0)
                connect _bundle_literal_expr_2.flags, _bundle_literal_expr_3
                connect _array_literal_expr_1[1], {|HdlNone, HdlSome: Ty65|}(HdlSome, _bundle_literal_expr_2)
                wire _bundle_literal_expr_4: Ty65
                connect _bundle_literal_expr_4.int_fp, UInt<64>(0h0)
                wire _bundle_literal_expr_5: Ty64
                connect _bundle_literal_expr_5.pwr_ca_x86_cf, UInt<1>(0h0)
                connect _bundle_literal_expr_5.pwr_ca32_x86_af, UInt<1>(0h0)
                connect _bundle_literal_expr_5.pwr_ov_x86_of, UInt<1>(0h0)
                connect _bundle_literal_expr_5.pwr_ov32_x86_df, UInt<1>(0h0)
                connect _bundle_literal_expr_5.pwr_cr_lt_x86_sf, UInt<1>(0h0)
                connect _bundle_literal_expr_5.pwr_cr_gt_x86_pf, UInt<1>(0h0)
                connect _bundle_literal_expr_5.pwr_cr_eq_x86_zf, UInt<1>(0h0)
                connect _bundle_literal_expr_5.pwr_so, UInt<1>(0h0)
                connect _bundle_literal_expr_4.flags, _bundle_literal_expr_5
                connect _array_literal_expr_1[2], {|HdlNone, HdlSome: Ty65|}(HdlSome, _bundle_literal_expr_4)
                connect input_in_flight_op_src_values, _array_literal_expr_1 @[unit_base.rs 82:9]
                match _match_arm_value_15: @[instruction.rs 502:1]
                    AddSub(_match_arm_value_16):
                        wire _bundle_literal_expr_6: Ty25
                        wire _bundle_literal_expr_7: Ty23
                        connect _bundle_literal_expr_7.adj_value, tail(UInt<64>(0h0), 62)
                        connect _bundle_literal_expr_6.unit_num, _bundle_literal_expr_7
                        wire _bundle_literal_expr_8: Ty24
                        connect _bundle_literal_expr_8.value, tail(UInt<8>(0h0), 4)
                        connect _bundle_literal_expr_6.unit_out_reg, _bundle_literal_expr_8
                        wire _cast_bundle_to_bits_expr: Ty57
                        connect _cast_bundle_to_bits_expr.unit_num, _bundle_literal_expr_6.unit_num.adj_value
                        connect _cast_bundle_to_bits_expr.unit_out_reg, _bundle_literal_expr_6.unit_out_reg.value
                        wire _cast_to_bits_expr: UInt<6>
                        connect _cast_to_bits_expr, cat(_cast_bundle_to_bits_expr.unit_out_reg, _cast_bundle_to_bits_expr.unit_num)
                        when neq(_cast_to_bits_expr, _match_arm_value_16.alu_common.common.src[0]): @[unit_base.rs 88:13]
                            connect input_in_flight_op_src_values[0], {|HdlNone, HdlSome: Ty65|}(HdlNone) @[unit_base.rs 94:17]
                        wire _bundle_literal_expr_9: Ty25
                        wire _bundle_literal_expr_10: Ty23
                        connect _bundle_literal_expr_10.adj_value, tail(UInt<64>(0h0), 62)
                        connect _bundle_literal_expr_9.unit_num, _bundle_literal_expr_10
                        wire _bundle_literal_expr_11: Ty24
                        connect _bundle_literal_expr_11.value, tail(UInt<8>(0h0), 4)
                        connect _bundle_literal_expr_9.unit_out_reg, _bundle_literal_expr_11
                        wire _cast_bundle_to_bits_expr_1: Ty57
                        connect _cast_bundle_to_bits_expr_1.unit_num, _bundle_literal_expr_9.unit_num.adj_value
                        connect _cast_bundle_to_bits_expr_1.unit_out_reg, _bundle_literal_expr_9.unit_out_reg.value
                        wire _cast_to_bits_expr_1: UInt<6>
                        connect _cast_to_bits_expr_1, cat(_cast_bundle_to_bits_expr_1.unit_out_reg, _cast_bundle_to_bits_expr_1.unit_num)
                        when neq(_cast_to_bits_expr_1, _match_arm_value_16.alu_common.common.src[1]): @[unit_base.rs 88:13]
                            connect input_in_flight_op_src_values[1], {|HdlNone, HdlSome: Ty65|}(HdlNone) @[unit_base.rs 94:17]
                        wire _bundle_literal_expr_12: Ty25
                        wire _bundle_literal_expr_13: Ty23
                        connect _bundle_literal_expr_13.adj_value, tail(UInt<64>(0h0), 62)
                        connect _bundle_literal_expr_12.unit_num, _bundle_literal_expr_13
                        wire _bundle_literal_expr_14: Ty24
                        connect _bundle_literal_expr_14.value, tail(UInt<8>(0h0), 4)
                        connect _bundle_literal_expr_12.unit_out_reg, _bundle_literal_expr_14
                        wire _cast_bundle_to_bits_expr_2: Ty57
                        connect _cast_bundle_to_bits_expr_2.unit_num, _bundle_literal_expr_12.unit_num.adj_value
                        connect _cast_bundle_to_bits_expr_2.unit_out_reg, _bundle_literal_expr_12.unit_out_reg.value
                        wire _cast_to_bits_expr_2: UInt<6>
                        connect _cast_to_bits_expr_2, cat(_cast_bundle_to_bits_expr_2.unit_out_reg, _cast_bundle_to_bits_expr_2.unit_num)
                        when neq(_cast_to_bits_expr_2, _match_arm_value_16.alu_common.common.src[2]): @[unit_base.rs 88:13]
                            connect input_in_flight_op_src_values[2], {|HdlNone, HdlSome: Ty65|}(HdlNone) @[unit_base.rs 94:17]
                    AddSubI(_match_arm_value_17):
                        wire _bundle_literal_expr_15: Ty25
                        wire _bundle_literal_expr_16: Ty23
                        connect _bundle_literal_expr_16.adj_value, tail(UInt<64>(0h0), 62)
                        connect _bundle_literal_expr_15.unit_num, _bundle_literal_expr_16
                        wire _bundle_literal_expr_17: Ty24
                        connect _bundle_literal_expr_17.value, tail(UInt<8>(0h0), 4)
                        connect _bundle_literal_expr_15.unit_out_reg, _bundle_literal_expr_17
                        wire _cast_bundle_to_bits_expr_3: Ty57
                        connect _cast_bundle_to_bits_expr_3.unit_num, _bundle_literal_expr_15.unit_num.adj_value
                        connect _cast_bundle_to_bits_expr_3.unit_out_reg, _bundle_literal_expr_15.unit_out_reg.value
                        wire _cast_to_bits_expr_3: UInt<6>
                        connect _cast_to_bits_expr_3, cat(_cast_bundle_to_bits_expr_3.unit_out_reg, _cast_bundle_to_bits_expr_3.unit_num)
                        when neq(_cast_to_bits_expr_3, _match_arm_value_17.alu_common.common.src[0]): @[unit_base.rs 88:13]
                            connect input_in_flight_op_src_values[0], {|HdlNone, HdlSome: Ty65|}(HdlNone) @[unit_base.rs 94:17]
                        wire _bundle_literal_expr_18: Ty25
                        wire _bundle_literal_expr_19: Ty23
                        connect _bundle_literal_expr_19.adj_value, tail(UInt<64>(0h0), 62)
                        connect _bundle_literal_expr_18.unit_num, _bundle_literal_expr_19
                        wire _bundle_literal_expr_20: Ty24
                        connect _bundle_literal_expr_20.value, tail(UInt<8>(0h0), 4)
                        connect _bundle_literal_expr_18.unit_out_reg, _bundle_literal_expr_20
                        wire _cast_bundle_to_bits_expr_4: Ty57
                        connect _cast_bundle_to_bits_expr_4.unit_num, _bundle_literal_expr_18.unit_num.adj_value
                        connect _cast_bundle_to_bits_expr_4.unit_out_reg, _bundle_literal_expr_18.unit_out_reg.value
                        wire _cast_to_bits_expr_4: UInt<6>
                        connect _cast_to_bits_expr_4, cat(_cast_bundle_to_bits_expr_4.unit_out_reg, _cast_bundle_to_bits_expr_4.unit_num)
                        when neq(_cast_to_bits_expr_4, _match_arm_value_17.alu_common.common.src[1]): @[unit_base.rs 88:13]
                            connect input_in_flight_op_src_values[1], {|HdlNone, HdlSome: Ty65|}(HdlNone) @[unit_base.rs 94:17]
                    Logical(_match_arm_value_18):
                        wire _bundle_literal_expr_21: Ty25
                        wire _bundle_literal_expr_22: Ty23
                        connect _bundle_literal_expr_22.adj_value, tail(UInt<64>(0h0), 62)
                        connect _bundle_literal_expr_21.unit_num, _bundle_literal_expr_22
                        wire _bundle_literal_expr_23: Ty24
                        connect _bundle_literal_expr_23.value, tail(UInt<8>(0h0), 4)
                        connect _bundle_literal_expr_21.unit_out_reg, _bundle_literal_expr_23
                        wire _cast_bundle_to_bits_expr_5: Ty57
                        connect _cast_bundle_to_bits_expr_5.unit_num, _bundle_literal_expr_21.unit_num.adj_value
                        connect _cast_bundle_to_bits_expr_5.unit_out_reg, _bundle_literal_expr_21.unit_out_reg.value
                        wire _cast_to_bits_expr_5: UInt<6>
                        connect _cast_to_bits_expr_5, cat(_cast_bundle_to_bits_expr_5.unit_out_reg, _cast_bundle_to_bits_expr_5.unit_num)
                        when neq(_cast_to_bits_expr_5, _match_arm_value_18.alu_common.common.src[0]): @[unit_base.rs 88:13]
                            connect input_in_flight_op_src_values[0], {|HdlNone, HdlSome: Ty65|}(HdlNone) @[unit_base.rs 94:17]
                        wire _bundle_literal_expr_24: Ty25
                        wire _bundle_literal_expr_25: Ty23
                        connect _bundle_literal_expr_25.adj_value, tail(UInt<64>(0h0), 62)
                        connect _bundle_literal_expr_24.unit_num, _bundle_literal_expr_25
                        wire _bundle_literal_expr_26: Ty24
                        connect _bundle_literal_expr_26.value, tail(UInt<8>(0h0), 4)
                        connect _bundle_literal_expr_24.unit_out_reg, _bundle_literal_expr_26
                        wire _cast_bundle_to_bits_expr_6: Ty57
                        connect _cast_bundle_to_bits_expr_6.unit_num, _bundle_literal_expr_24.unit_num.adj_value
                        connect _cast_bundle_to_bits_expr_6.unit_out_reg, _bundle_literal_expr_24.unit_out_reg.value
                        wire _cast_to_bits_expr_6: UInt<6>
                        connect _cast_to_bits_expr_6, cat(_cast_bundle_to_bits_expr_6.unit_out_reg, _cast_bundle_to_bits_expr_6.unit_num)
                        when neq(_cast_to_bits_expr_6, _match_arm_value_18.alu_common.common.src[1]): @[unit_base.rs 88:13]
                            connect input_in_flight_op_src_values[1], {|HdlNone, HdlSome: Ty65|}(HdlNone) @[unit_base.rs 94:17]
                wire _bundle_literal_expr_27: Ty87
                connect _bundle_literal_expr_27.mop, _match_arm_value_15
                connect _bundle_literal_expr_27.src_values, input_in_flight_op_src_values
                connect input_in_flight_op, {|HdlNone, HdlSome: Ty87|}(HdlSome, _bundle_literal_expr_27) @[unit_base.rs 97:9]
        match in_flight_ops[0]: @[unit_base.rs 107:9]
            HdlNone:
                match input_index: @[unit_base.rs 122:16]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_19):
                        connect input_insn.ready, UInt<1>(0h1) @[unit_base.rs 123:13]
                        when eq(_match_arm_value_19, UInt<64>(0h0)): @[unit_base.rs 125:13]
                            connect in_flight_ops[0], input_in_flight_op @[unit_base.rs 126:17]
            HdlSome(_match_arm_value_20):
                wire firing_data_1: Ty70 @[ready_valid.rs 30:27]
                connect firing_data_1, {|HdlNone, HdlSome: Ty69|}(HdlNone) @[ready_valid.rs 31:9]
                when cancel_input.ready: @[ready_valid.rs 33:9]
                    connect firing_data_1, cancel_input.data @[ready_valid.rs 34:13]
                match firing_data_1: @[unit_base.rs 109:13]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_21):
                        wire dest_reg: Ty24 @[instruction.rs 502:1]
                        match _match_arm_value_20.mop: @[instruction.rs 502:1]
                            AddSub(_match_arm_value_22):
                                connect dest_reg, _match_arm_value_22.alu_common.common.dest @[instruction.rs 502:1]
                            AddSubI(_match_arm_value_23):
                                connect dest_reg, _match_arm_value_23.alu_common.common.dest @[instruction.rs 502:1]
                            Logical(_match_arm_value_24):
                                connect dest_reg, _match_arm_value_24.alu_common.common.dest @[instruction.rs 502:1]
                        when eq(_match_arm_value_21.which.value, dest_reg.value): @[unit_base.rs 113:17]
                            connect in_flight_ops[0], {|HdlNone, HdlSome: Ty87|}(HdlNone) @[unit_base.rs 115:21]
        match in_flight_ops[1]: @[unit_base.rs 107:9]
            HdlNone:
                match input_index: @[unit_base.rs 122:16]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_25):
                        connect input_insn.ready, UInt<1>(0h1) @[unit_base.rs 123:13]
                        when eq(_match_arm_value_25, UInt<64>(0h1)): @[unit_base.rs 125:13]
                            connect in_flight_ops[1], input_in_flight_op @[unit_base.rs 126:17]
            HdlSome(_match_arm_value_26):
                wire firing_data_2: Ty70 @[ready_valid.rs 30:27]
                connect firing_data_2, {|HdlNone, HdlSome: Ty69|}(HdlNone) @[ready_valid.rs 31:9]
                when cancel_input.ready: @[ready_valid.rs 33:9]
                    connect firing_data_2, cancel_input.data @[ready_valid.rs 34:13]
                match firing_data_2: @[unit_base.rs 109:13]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_27):
                        wire dest_reg_1: Ty24 @[instruction.rs 502:1]
                        match _match_arm_value_26.mop: @[instruction.rs 502:1]
                            AddSub(_match_arm_value_28):
                                connect dest_reg_1, _match_arm_value_28.alu_common.common.dest @[instruction.rs 502:1]
                            AddSubI(_match_arm_value_29):
                                connect dest_reg_1, _match_arm_value_29.alu_common.common.dest @[instruction.rs 502:1]
                            Logical(_match_arm_value_30):
                                connect dest_reg_1, _match_arm_value_30.alu_common.common.dest @[instruction.rs 502:1]
                        when eq(_match_arm_value_27.which.value, dest_reg_1.value): @[unit_base.rs 113:17]
                            connect in_flight_ops[1], {|HdlNone, HdlSome: Ty87|}(HdlNone) @[unit_base.rs 115:21]
        match in_flight_ops[2]: @[unit_base.rs 107:9]
            HdlNone:
                match input_index: @[unit_base.rs 122:16]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_31):
                        connect input_insn.ready, UInt<1>(0h1) @[unit_base.rs 123:13]
                        when eq(_match_arm_value_31, UInt<64>(0h2)): @[unit_base.rs 125:13]
                            connect in_flight_ops[2], input_in_flight_op @[unit_base.rs 126:17]
            HdlSome(_match_arm_value_32):
                wire firing_data_3: Ty70 @[ready_valid.rs 30:27]
                connect firing_data_3, {|HdlNone, HdlSome: Ty69|}(HdlNone) @[ready_valid.rs 31:9]
                when cancel_input.ready: @[ready_valid.rs 33:9]
                    connect firing_data_3, cancel_input.data @[ready_valid.rs 34:13]
                match firing_data_3: @[unit_base.rs 109:13]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_33):
                        wire dest_reg_2: Ty24 @[instruction.rs 502:1]
                        match _match_arm_value_32.mop: @[instruction.rs 502:1]
                            AddSub(_match_arm_value_34):
                                connect dest_reg_2, _match_arm_value_34.alu_common.common.dest @[instruction.rs 502:1]
                            AddSubI(_match_arm_value_35):
                                connect dest_reg_2, _match_arm_value_35.alu_common.common.dest @[instruction.rs 502:1]
                            Logical(_match_arm_value_36):
                                connect dest_reg_2, _match_arm_value_36.alu_common.common.dest @[instruction.rs 502:1]
                        when eq(_match_arm_value_33.which.value, dest_reg_2.value): @[unit_base.rs 113:17]
                            connect in_flight_ops[2], {|HdlNone, HdlSome: Ty87|}(HdlNone) @[unit_base.rs 115:21]
        match in_flight_ops[3]: @[unit_base.rs 107:9]
            HdlNone:
                match input_index: @[unit_base.rs 122:16]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_37):
                        connect input_insn.ready, UInt<1>(0h1) @[unit_base.rs 123:13]
                        when eq(_match_arm_value_37, UInt<64>(0h3)): @[unit_base.rs 125:13]
                            connect in_flight_ops[3], input_in_flight_op @[unit_base.rs 126:17]
            HdlSome(_match_arm_value_38):
                wire firing_data_4: Ty70 @[ready_valid.rs 30:27]
                connect firing_data_4, {|HdlNone, HdlSome: Ty69|}(HdlNone) @[ready_valid.rs 31:9]
                when cancel_input.ready: @[ready_valid.rs 33:9]
                    connect firing_data_4, cancel_input.data @[ready_valid.rs 34:13]
                match firing_data_4: @[unit_base.rs 109:13]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_39):
                        wire dest_reg_3: Ty24 @[instruction.rs 502:1]
                        match _match_arm_value_38.mop: @[instruction.rs 502:1]
                            AddSub(_match_arm_value_40):
                                connect dest_reg_3, _match_arm_value_40.alu_common.common.dest @[instruction.rs 502:1]
                            AddSubI(_match_arm_value_41):
                                connect dest_reg_3, _match_arm_value_41.alu_common.common.dest @[instruction.rs 502:1]
                            Logical(_match_arm_value_42):
                                connect dest_reg_3, _match_arm_value_42.alu_common.common.dest @[instruction.rs 502:1]
                        when eq(_match_arm_value_39.which.value, dest_reg_3.value): @[unit_base.rs 113:17]
                            connect in_flight_ops[3], {|HdlNone, HdlSome: Ty87|}(HdlNone) @[unit_base.rs 115:21]
        match in_flight_ops[4]: @[unit_base.rs 107:9]
            HdlNone:
                match input_index: @[unit_base.rs 122:16]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_43):
                        connect input_insn.ready, UInt<1>(0h1) @[unit_base.rs 123:13]
                        when eq(_match_arm_value_43, UInt<64>(0h4)): @[unit_base.rs 125:13]
                            connect in_flight_ops[4], input_in_flight_op @[unit_base.rs 126:17]
            HdlSome(_match_arm_value_44):
                wire firing_data_5: Ty70 @[ready_valid.rs 30:27]
                connect firing_data_5, {|HdlNone, HdlSome: Ty69|}(HdlNone) @[ready_valid.rs 31:9]
                when cancel_input.ready: @[ready_valid.rs 33:9]
                    connect firing_data_5, cancel_input.data @[ready_valid.rs 34:13]
                match firing_data_5: @[unit_base.rs 109:13]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_45):
                        wire dest_reg_4: Ty24 @[instruction.rs 502:1]
                        match _match_arm_value_44.mop: @[instruction.rs 502:1]
                            AddSub(_match_arm_value_46):
                                connect dest_reg_4, _match_arm_value_46.alu_common.common.dest @[instruction.rs 502:1]
                            AddSubI(_match_arm_value_47):
                                connect dest_reg_4, _match_arm_value_47.alu_common.common.dest @[instruction.rs 502:1]
                            Logical(_match_arm_value_48):
                                connect dest_reg_4, _match_arm_value_48.alu_common.common.dest @[instruction.rs 502:1]
                        when eq(_match_arm_value_45.which.value, dest_reg_4.value): @[unit_base.rs 113:17]
                            connect in_flight_ops[4], {|HdlNone, HdlSome: Ty87|}(HdlNone) @[unit_base.rs 115:21]
        match in_flight_ops[5]: @[unit_base.rs 107:9]
            HdlNone:
                match input_index: @[unit_base.rs 122:16]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_49):
                        connect input_insn.ready, UInt<1>(0h1) @[unit_base.rs 123:13]
                        when eq(_match_arm_value_49, UInt<64>(0h5)): @[unit_base.rs 125:13]
                            connect in_flight_ops[5], input_in_flight_op @[unit_base.rs 126:17]
            HdlSome(_match_arm_value_50):
                wire firing_data_6: Ty70 @[ready_valid.rs 30:27]
                connect firing_data_6, {|HdlNone, HdlSome: Ty69|}(HdlNone) @[ready_valid.rs 31:9]
                when cancel_input.ready: @[ready_valid.rs 33:9]
                    connect firing_data_6, cancel_input.data @[ready_valid.rs 34:13]
                match firing_data_6: @[unit_base.rs 109:13]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_51):
                        wire dest_reg_5: Ty24 @[instruction.rs 502:1]
                        match _match_arm_value_50.mop: @[instruction.rs 502:1]
                            AddSub(_match_arm_value_52):
                                connect dest_reg_5, _match_arm_value_52.alu_common.common.dest @[instruction.rs 502:1]
                            AddSubI(_match_arm_value_53):
                                connect dest_reg_5, _match_arm_value_53.alu_common.common.dest @[instruction.rs 502:1]
                            Logical(_match_arm_value_54):
                                connect dest_reg_5, _match_arm_value_54.alu_common.common.dest @[instruction.rs 502:1]
                        when eq(_match_arm_value_51.which.value, dest_reg_5.value): @[unit_base.rs 113:17]
                            connect in_flight_ops[5], {|HdlNone, HdlSome: Ty87|}(HdlNone) @[unit_base.rs 115:21]
        match in_flight_ops[6]: @[unit_base.rs 107:9]
            HdlNone:
                match input_index: @[unit_base.rs 122:16]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_55):
                        connect input_insn.ready, UInt<1>(0h1) @[unit_base.rs 123:13]
                        when eq(_match_arm_value_55, UInt<64>(0h6)): @[unit_base.rs 125:13]
                            connect in_flight_ops[6], input_in_flight_op @[unit_base.rs 126:17]
            HdlSome(_match_arm_value_56):
                wire firing_data_7: Ty70 @[ready_valid.rs 30:27]
                connect firing_data_7, {|HdlNone, HdlSome: Ty69|}(HdlNone) @[ready_valid.rs 31:9]
                when cancel_input.ready: @[ready_valid.rs 33:9]
                    connect firing_data_7, cancel_input.data @[ready_valid.rs 34:13]
                match firing_data_7: @[unit_base.rs 109:13]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_57):
                        wire dest_reg_6: Ty24 @[instruction.rs 502:1]
                        match _match_arm_value_56.mop: @[instruction.rs 502:1]
                            AddSub(_match_arm_value_58):
                                connect dest_reg_6, _match_arm_value_58.alu_common.common.dest @[instruction.rs 502:1]
                            AddSubI(_match_arm_value_59):
                                connect dest_reg_6, _match_arm_value_59.alu_common.common.dest @[instruction.rs 502:1]
                            Logical(_match_arm_value_60):
                                connect dest_reg_6, _match_arm_value_60.alu_common.common.dest @[instruction.rs 502:1]
                        when eq(_match_arm_value_57.which.value, dest_reg_6.value): @[unit_base.rs 113:17]
                            connect in_flight_ops[6], {|HdlNone, HdlSome: Ty87|}(HdlNone) @[unit_base.rs 115:21]
        match in_flight_ops[7]: @[unit_base.rs 107:9]
            HdlNone:
                match input_index: @[unit_base.rs 122:16]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_61):
                        connect input_insn.ready, UInt<1>(0h1) @[unit_base.rs 123:13]
                        when eq(_match_arm_value_61, UInt<64>(0h7)): @[unit_base.rs 125:13]
                            connect in_flight_ops[7], input_in_flight_op @[unit_base.rs 126:17]
            HdlSome(_match_arm_value_62):
                wire firing_data_8: Ty70 @[ready_valid.rs 30:27]
                connect firing_data_8, {|HdlNone, HdlSome: Ty69|}(HdlNone) @[ready_valid.rs 31:9]
                when cancel_input.ready: @[ready_valid.rs 33:9]
                    connect firing_data_8, cancel_input.data @[ready_valid.rs 34:13]
                match firing_data_8: @[unit_base.rs 109:13]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_63):
                        wire dest_reg_7: Ty24 @[instruction.rs 502:1]
                        match _match_arm_value_62.mop: @[instruction.rs 502:1]
                            AddSub(_match_arm_value_64):
                                connect dest_reg_7, _match_arm_value_64.alu_common.common.dest @[instruction.rs 502:1]
                            AddSubI(_match_arm_value_65):
                                connect dest_reg_7, _match_arm_value_65.alu_common.common.dest @[instruction.rs 502:1]
                            Logical(_match_arm_value_66):
                                connect dest_reg_7, _match_arm_value_66.alu_common.common.dest @[instruction.rs 502:1]
                        when eq(_match_arm_value_63.which.value, dest_reg_7.value): @[unit_base.rs 113:17]
                            connect in_flight_ops[7], {|HdlNone, HdlSome: Ty87|}(HdlNone) @[unit_base.rs 115:21]
    module unit_base_1: @[unit_base.rs 32:1]
        input cd: Ty0 @[unit_base.rs 39:29]
        input unit_forwarding_info: Ty68 @[unit_base.rs 42:11]
        input input_insn: Ty63 @[unit_base.rs 44:41]
        input cancel_input: Ty71 @[unit_base.rs 48:11]
        output ready_mop: Ty83 @[unit_base.rs 51:50]
        wire and_then_out: Ty85 @[unit_base.rs 66:17]
        wire and_then_out_1: Ty85 @[unit_base.rs 66:17]
        wire and_then_out_2: Ty85 @[unit_base.rs 66:17]
        wire and_then_out_3: Ty85 @[unit_base.rs 66:17]
        wire and_then_out_4: Ty85 @[unit_base.rs 66:17]
        wire and_then_out_5: Ty85 @[unit_base.rs 66:17]
        wire and_then_out_6: Ty85 @[unit_base.rs 66:17]
        wire and_then_out_7: Ty85 @[unit_base.rs 66:17]
        connect input_insn.ready, UInt<1>(0h0) @[unit_base.rs 45:5]
        connect cancel_input.ready, UInt<1>(0h1) @[unit_base.rs 49:5]
        connect ready_mop.data, {|HdlNone, HdlSome: Ty81|}(HdlNone) @[unit_base.rs 52:5]
        wire _array_literal_expr: Ty88[8]
        connect _array_literal_expr[0], {|HdlNone, HdlSome: Ty87|}(HdlNone)
        connect _array_literal_expr[1], {|HdlNone, HdlSome: Ty87|}(HdlNone)
        connect _array_literal_expr[2], {|HdlNone, HdlSome: Ty87|}(HdlNone)
        connect _array_literal_expr[3], {|HdlNone, HdlSome: Ty87|}(HdlNone)
        connect _array_literal_expr[4], {|HdlNone, HdlSome: Ty87|}(HdlNone)
        connect _array_literal_expr[5], {|HdlNone, HdlSome: Ty87|}(HdlNone)
        connect _array_literal_expr[6], {|HdlNone, HdlSome: Ty87|}(HdlNone)
        connect _array_literal_expr[7], {|HdlNone, HdlSome: Ty87|}(HdlNone)
        regreset in_flight_ops: Ty88[8], cd.clk, cd.rst, _array_literal_expr @[unit_base.rs 55:25]
        wire input_index: Ty85 @[unit_base.rs 61:23]
        match in_flight_ops[0]: @[unit_base.rs 66:17]
            HdlNone:
                connect and_then_out, {|HdlNone, HdlSome: UInt<3>|}(HdlNone) @[unit_base.rs 66:17]
            HdlSome(_match_arm_value):
                connect and_then_out, {|HdlNone, HdlSome: UInt<3>|}(HdlSome, tail(UInt<64>(0h0), 61)) @[unit_base.rs 66:17]
        match in_flight_ops[1]: @[unit_base.rs 66:17]
            HdlNone:
                connect and_then_out_1, {|HdlNone, HdlSome: UInt<3>|}(HdlNone) @[unit_base.rs 66:17]
            HdlSome(_match_arm_value_1):
                connect and_then_out_1, {|HdlNone, HdlSome: UInt<3>|}(HdlSome, tail(UInt<64>(0h1), 61)) @[unit_base.rs 66:17]
        wire or_out: Ty85 @[function.rs 166:5]
        connect or_out, and_then_out_1 @[function.rs 166:5]
        match and_then_out: @[function.rs 166:5]
            HdlNone:
                skip
            HdlSome(_match_arm_value_2):
                connect or_out, and_then_out @[function.rs 166:5]
        match in_flight_ops[2]: @[unit_base.rs 66:17]
            HdlNone:
                connect and_then_out_2, {|HdlNone, HdlSome: UInt<3>|}(HdlNone) @[unit_base.rs 66:17]
            HdlSome(_match_arm_value_3):
                connect and_then_out_2, {|HdlNone, HdlSome: UInt<3>|}(HdlSome, tail(UInt<64>(0h2), 61)) @[unit_base.rs 66:17]
        match in_flight_ops[3]: @[unit_base.rs 66:17]
            HdlNone:
                connect and_then_out_3, {|HdlNone, HdlSome: UInt<3>|}(HdlNone) @[unit_base.rs 66:17]
            HdlSome(_match_arm_value_4):
                connect and_then_out_3, {|HdlNone, HdlSome: UInt<3>|}(HdlSome, tail(UInt<64>(0h3), 61)) @[unit_base.rs 66:17]
        wire or_out_1: Ty85 @[function.rs 166:5]
        connect or_out_1, and_then_out_3 @[function.rs 166:5]
        match and_then_out_2: @[function.rs 166:5]
            HdlNone:
                skip
            HdlSome(_match_arm_value_5):
                connect or_out_1, and_then_out_2 @[function.rs 166:5]
        wire or_out_2: Ty85 @[function.rs 166:5]
        connect or_out_2, or_out_1 @[function.rs 166:5]
        match or_out: @[function.rs 166:5]
            HdlNone:
                skip
            HdlSome(_match_arm_value_6):
                connect or_out_2, or_out @[function.rs 166:5]
        match in_flight_ops[4]: @[unit_base.rs 66:17]
            HdlNone:
                connect and_then_out_4, {|HdlNone, HdlSome: UInt<3>|}(HdlNone) @[unit_base.rs 66:17]
            HdlSome(_match_arm_value_7):
                connect and_then_out_4, {|HdlNone, HdlSome: UInt<3>|}(HdlSome, tail(UInt<64>(0h4), 61)) @[unit_base.rs 66:17]
        match in_flight_ops[5]: @[unit_base.rs 66:17]
            HdlNone:
                connect and_then_out_5, {|HdlNone, HdlSome: UInt<3>|}(HdlNone) @[unit_base.rs 66:17]
            HdlSome(_match_arm_value_8):
                connect and_then_out_5, {|HdlNone, HdlSome: UInt<3>|}(HdlSome, tail(UInt<64>(0h5), 61)) @[unit_base.rs 66:17]
        wire or_out_3: Ty85 @[function.rs 166:5]
        connect or_out_3, and_then_out_5 @[function.rs 166:5]
        match and_then_out_4: @[function.rs 166:5]
            HdlNone:
                skip
            HdlSome(_match_arm_value_9):
                connect or_out_3, and_then_out_4 @[function.rs 166:5]
        match in_flight_ops[6]: @[unit_base.rs 66:17]
            HdlNone:
                connect and_then_out_6, {|HdlNone, HdlSome: UInt<3>|}(HdlNone) @[unit_base.rs 66:17]
            HdlSome(_match_arm_value_10):
                connect and_then_out_6, {|HdlNone, HdlSome: UInt<3>|}(HdlSome, tail(UInt<64>(0h6), 61)) @[unit_base.rs 66:17]
        match in_flight_ops[7]: @[unit_base.rs 66:17]
            HdlNone:
                connect and_then_out_7, {|HdlNone, HdlSome: UInt<3>|}(HdlNone) @[unit_base.rs 66:17]
            HdlSome(_match_arm_value_11):
                connect and_then_out_7, {|HdlNone, HdlSome: UInt<3>|}(HdlSome, tail(UInt<64>(0h7), 61)) @[unit_base.rs 66:17]
        wire or_out_4: Ty85 @[function.rs 166:5]
        connect or_out_4, and_then_out_7 @[function.rs 166:5]
        match and_then_out_6: @[function.rs 166:5]
            HdlNone:
                skip
            HdlSome(_match_arm_value_12):
                connect or_out_4, and_then_out_6 @[function.rs 166:5]
        wire or_out_5: Ty85 @[function.rs 166:5]
        connect or_out_5, or_out_4 @[function.rs 166:5]
        match or_out_3: @[function.rs 166:5]
            HdlNone:
                skip
            HdlSome(_match_arm_value_13):
                connect or_out_5, or_out_3 @[function.rs 166:5]
        wire or_out_6: Ty85 @[function.rs 166:5]
        connect or_out_6, or_out_5 @[function.rs 166:5]
        match or_out_2: @[function.rs 166:5]
            HdlNone:
                skip
            HdlSome(_match_arm_value_14):
                connect or_out_6, or_out_2 @[function.rs 166:5]
        connect input_index, or_out_6 @[unit_base.rs 62:5]
        wire input_in_flight_op: Ty88 @[unit_base.rs 73:30]
        connect input_in_flight_op, {|HdlNone, HdlSome: Ty87|}(HdlNone) @[unit_base.rs 74:5]
        wire firing_data: Ty62 @[ready_valid.rs 30:27]
        connect firing_data, {|HdlNone, HdlSome: Ty46|}(HdlNone) @[ready_valid.rs 31:9]
        when input_insn.ready: @[ready_valid.rs 33:9]
            connect firing_data, input_insn.data @[ready_valid.rs 34:13]
        match firing_data: @[unit_base.rs 76:5]
            HdlNone:
                skip
            HdlSome(_match_arm_value_15):
                wire input_in_flight_op_src_values: Ty86[3] @[unit_base.rs 79:13]
                wire _array_literal_expr_1: Ty86[3]
                wire _bundle_literal_expr: Ty65
                connect _bundle_literal_expr.int_fp, UInt<64>(0h0)
                wire _bundle_literal_expr_1: Ty64
                connect _bundle_literal_expr_1.pwr_ca_x86_cf, UInt<1>(0h0)
                connect _bundle_literal_expr_1.pwr_ca32_x86_af, UInt<1>(0h0)
                connect _bundle_literal_expr_1.pwr_ov_x86_of, UInt<1>(0h0)
                connect _bundle_literal_expr_1.pwr_ov32_x86_df, UInt<1>(0h0)
                connect _bundle_literal_expr_1.pwr_cr_lt_x86_sf, UInt<1>(0h0)
                connect _bundle_literal_expr_1.pwr_cr_gt_x86_pf, UInt<1>(0h0)
                connect _bundle_literal_expr_1.pwr_cr_eq_x86_zf, UInt<1>(0h0)
                connect _bundle_literal_expr_1.pwr_so, UInt<1>(0h0)
                connect _bundle_literal_expr.flags, _bundle_literal_expr_1
                connect _array_literal_expr_1[0], {|HdlNone, HdlSome: Ty65|}(HdlSome, _bundle_literal_expr)
                wire _bundle_literal_expr_2: Ty65
                connect _bundle_literal_expr_2.int_fp, UInt<64>(0h0)
                wire _bundle_literal_expr_3: Ty64
                connect _bundle_literal_expr_3.pwr_ca_x86_cf, UInt<1>(0h0)
                connect _bundle_literal_expr_3.pwr_ca32_x86_af, UInt<1>(0h0)
                connect _bundle_literal_expr_3.pwr_ov_x86_of, UInt<1>(0h0)
                connect _bundle_literal_expr_3.pwr_ov32_x86_df, UInt<1>(0h0)
                connect _bundle_literal_expr_3.pwr_cr_lt_x86_sf, UInt<1>(0h0)
                connect _bundle_literal_expr_3.pwr_cr_gt_x86_pf, UInt<1>(0h0)
                connect _bundle_literal_expr_3.pwr_cr_eq_x86_zf, UInt<1>(0h0)
                connect _bundle_literal_expr_3.pwr_so, UInt<1>(0h0)
                connect _bundle_literal_expr_2.flags, _bundle_literal_expr_3
                connect _array_literal_expr_1[1], {|HdlNone, HdlSome: Ty65|}(HdlSome, _bundle_literal_expr_2)
                wire _bundle_literal_expr_4: Ty65
                connect _bundle_literal_expr_4.int_fp, UInt<64>(0h0)
                wire _bundle_literal_expr_5: Ty64
                connect _bundle_literal_expr_5.pwr_ca_x86_cf, UInt<1>(0h0)
                connect _bundle_literal_expr_5.pwr_ca32_x86_af, UInt<1>(0h0)
                connect _bundle_literal_expr_5.pwr_ov_x86_of, UInt<1>(0h0)
                connect _bundle_literal_expr_5.pwr_ov32_x86_df, UInt<1>(0h0)
                connect _bundle_literal_expr_5.pwr_cr_lt_x86_sf, UInt<1>(0h0)
                connect _bundle_literal_expr_5.pwr_cr_gt_x86_pf, UInt<1>(0h0)
                connect _bundle_literal_expr_5.pwr_cr_eq_x86_zf, UInt<1>(0h0)
                connect _bundle_literal_expr_5.pwr_so, UInt<1>(0h0)
                connect _bundle_literal_expr_4.flags, _bundle_literal_expr_5
                connect _array_literal_expr_1[2], {|HdlNone, HdlSome: Ty65|}(HdlSome, _bundle_literal_expr_4)
                connect input_in_flight_op_src_values, _array_literal_expr_1 @[unit_base.rs 82:9]
                match _match_arm_value_15: @[instruction.rs 502:1]
                    AddSub(_match_arm_value_16):
                        wire _bundle_literal_expr_6: Ty25
                        wire _bundle_literal_expr_7: Ty23
                        connect _bundle_literal_expr_7.adj_value, tail(UInt<64>(0h0), 62)
                        connect _bundle_literal_expr_6.unit_num, _bundle_literal_expr_7
                        wire _bundle_literal_expr_8: Ty24
                        connect _bundle_literal_expr_8.value, tail(UInt<8>(0h0), 4)
                        connect _bundle_literal_expr_6.unit_out_reg, _bundle_literal_expr_8
                        wire _cast_bundle_to_bits_expr: Ty57
                        connect _cast_bundle_to_bits_expr.unit_num, _bundle_literal_expr_6.unit_num.adj_value
                        connect _cast_bundle_to_bits_expr.unit_out_reg, _bundle_literal_expr_6.unit_out_reg.value
                        wire _cast_to_bits_expr: UInt<6>
                        connect _cast_to_bits_expr, cat(_cast_bundle_to_bits_expr.unit_out_reg, _cast_bundle_to_bits_expr.unit_num)
                        when neq(_cast_to_bits_expr, _match_arm_value_16.alu_common.common.src[0]): @[unit_base.rs 88:13]
                            connect input_in_flight_op_src_values[0], {|HdlNone, HdlSome: Ty65|}(HdlNone) @[unit_base.rs 94:17]
                        wire _bundle_literal_expr_9: Ty25
                        wire _bundle_literal_expr_10: Ty23
                        connect _bundle_literal_expr_10.adj_value, tail(UInt<64>(0h0), 62)
                        connect _bundle_literal_expr_9.unit_num, _bundle_literal_expr_10
                        wire _bundle_literal_expr_11: Ty24
                        connect _bundle_literal_expr_11.value, tail(UInt<8>(0h0), 4)
                        connect _bundle_literal_expr_9.unit_out_reg, _bundle_literal_expr_11
                        wire _cast_bundle_to_bits_expr_1: Ty57
                        connect _cast_bundle_to_bits_expr_1.unit_num, _bundle_literal_expr_9.unit_num.adj_value
                        connect _cast_bundle_to_bits_expr_1.unit_out_reg, _bundle_literal_expr_9.unit_out_reg.value
                        wire _cast_to_bits_expr_1: UInt<6>
                        connect _cast_to_bits_expr_1, cat(_cast_bundle_to_bits_expr_1.unit_out_reg, _cast_bundle_to_bits_expr_1.unit_num)
                        when neq(_cast_to_bits_expr_1, _match_arm_value_16.alu_common.common.src[1]): @[unit_base.rs 88:13]
                            connect input_in_flight_op_src_values[1], {|HdlNone, HdlSome: Ty65|}(HdlNone) @[unit_base.rs 94:17]
                        wire _bundle_literal_expr_12: Ty25
                        wire _bundle_literal_expr_13: Ty23
                        connect _bundle_literal_expr_13.adj_value, tail(UInt<64>(0h0), 62)
                        connect _bundle_literal_expr_12.unit_num, _bundle_literal_expr_13
                        wire _bundle_literal_expr_14: Ty24
                        connect _bundle_literal_expr_14.value, tail(UInt<8>(0h0), 4)
                        connect _bundle_literal_expr_12.unit_out_reg, _bundle_literal_expr_14
                        wire _cast_bundle_to_bits_expr_2: Ty57
                        connect _cast_bundle_to_bits_expr_2.unit_num, _bundle_literal_expr_12.unit_num.adj_value
                        connect _cast_bundle_to_bits_expr_2.unit_out_reg, _bundle_literal_expr_12.unit_out_reg.value
                        wire _cast_to_bits_expr_2: UInt<6>
                        connect _cast_to_bits_expr_2, cat(_cast_bundle_to_bits_expr_2.unit_out_reg, _cast_bundle_to_bits_expr_2.unit_num)
                        when neq(_cast_to_bits_expr_2, _match_arm_value_16.alu_common.common.src[2]): @[unit_base.rs 88:13]
                            connect input_in_flight_op_src_values[2], {|HdlNone, HdlSome: Ty65|}(HdlNone) @[unit_base.rs 94:17]
                    AddSubI(_match_arm_value_17):
                        wire _bundle_literal_expr_15: Ty25
                        wire _bundle_literal_expr_16: Ty23
                        connect _bundle_literal_expr_16.adj_value, tail(UInt<64>(0h0), 62)
                        connect _bundle_literal_expr_15.unit_num, _bundle_literal_expr_16
                        wire _bundle_literal_expr_17: Ty24
                        connect _bundle_literal_expr_17.value, tail(UInt<8>(0h0), 4)
                        connect _bundle_literal_expr_15.unit_out_reg, _bundle_literal_expr_17
                        wire _cast_bundle_to_bits_expr_3: Ty57
                        connect _cast_bundle_to_bits_expr_3.unit_num, _bundle_literal_expr_15.unit_num.adj_value
                        connect _cast_bundle_to_bits_expr_3.unit_out_reg, _bundle_literal_expr_15.unit_out_reg.value
                        wire _cast_to_bits_expr_3: UInt<6>
                        connect _cast_to_bits_expr_3, cat(_cast_bundle_to_bits_expr_3.unit_out_reg, _cast_bundle_to_bits_expr_3.unit_num)
                        when neq(_cast_to_bits_expr_3, _match_arm_value_17.alu_common.common.src[0]): @[unit_base.rs 88:13]
                            connect input_in_flight_op_src_values[0], {|HdlNone, HdlSome: Ty65|}(HdlNone) @[unit_base.rs 94:17]
                        wire _bundle_literal_expr_18: Ty25
                        wire _bundle_literal_expr_19: Ty23
                        connect _bundle_literal_expr_19.adj_value, tail(UInt<64>(0h0), 62)
                        connect _bundle_literal_expr_18.unit_num, _bundle_literal_expr_19
                        wire _bundle_literal_expr_20: Ty24
                        connect _bundle_literal_expr_20.value, tail(UInt<8>(0h0), 4)
                        connect _bundle_literal_expr_18.unit_out_reg, _bundle_literal_expr_20
                        wire _cast_bundle_to_bits_expr_4: Ty57
                        connect _cast_bundle_to_bits_expr_4.unit_num, _bundle_literal_expr_18.unit_num.adj_value
                        connect _cast_bundle_to_bits_expr_4.unit_out_reg, _bundle_literal_expr_18.unit_out_reg.value
                        wire _cast_to_bits_expr_4: UInt<6>
                        connect _cast_to_bits_expr_4, cat(_cast_bundle_to_bits_expr_4.unit_out_reg, _cast_bundle_to_bits_expr_4.unit_num)
                        when neq(_cast_to_bits_expr_4, _match_arm_value_17.alu_common.common.src[1]): @[unit_base.rs 88:13]
                            connect input_in_flight_op_src_values[1], {|HdlNone, HdlSome: Ty65|}(HdlNone) @[unit_base.rs 94:17]
                    Logical(_match_arm_value_18):
                        wire _bundle_literal_expr_21: Ty25
                        wire _bundle_literal_expr_22: Ty23
                        connect _bundle_literal_expr_22.adj_value, tail(UInt<64>(0h0), 62)
                        connect _bundle_literal_expr_21.unit_num, _bundle_literal_expr_22
                        wire _bundle_literal_expr_23: Ty24
                        connect _bundle_literal_expr_23.value, tail(UInt<8>(0h0), 4)
                        connect _bundle_literal_expr_21.unit_out_reg, _bundle_literal_expr_23
                        wire _cast_bundle_to_bits_expr_5: Ty57
                        connect _cast_bundle_to_bits_expr_5.unit_num, _bundle_literal_expr_21.unit_num.adj_value
                        connect _cast_bundle_to_bits_expr_5.unit_out_reg, _bundle_literal_expr_21.unit_out_reg.value
                        wire _cast_to_bits_expr_5: UInt<6>
                        connect _cast_to_bits_expr_5, cat(_cast_bundle_to_bits_expr_5.unit_out_reg, _cast_bundle_to_bits_expr_5.unit_num)
                        when neq(_cast_to_bits_expr_5, _match_arm_value_18.alu_common.common.src[0]): @[unit_base.rs 88:13]
                            connect input_in_flight_op_src_values[0], {|HdlNone, HdlSome: Ty65|}(HdlNone) @[unit_base.rs 94:17]
                        wire _bundle_literal_expr_24: Ty25
                        wire _bundle_literal_expr_25: Ty23
                        connect _bundle_literal_expr_25.adj_value, tail(UInt<64>(0h0), 62)
                        connect _bundle_literal_expr_24.unit_num, _bundle_literal_expr_25
                        wire _bundle_literal_expr_26: Ty24
                        connect _bundle_literal_expr_26.value, tail(UInt<8>(0h0), 4)
                        connect _bundle_literal_expr_24.unit_out_reg, _bundle_literal_expr_26
                        wire _cast_bundle_to_bits_expr_6: Ty57
                        connect _cast_bundle_to_bits_expr_6.unit_num, _bundle_literal_expr_24.unit_num.adj_value
                        connect _cast_bundle_to_bits_expr_6.unit_out_reg, _bundle_literal_expr_24.unit_out_reg.value
                        wire _cast_to_bits_expr_6: UInt<6>
                        connect _cast_to_bits_expr_6, cat(_cast_bundle_to_bits_expr_6.unit_out_reg, _cast_bundle_to_bits_expr_6.unit_num)
                        when neq(_cast_to_bits_expr_6, _match_arm_value_18.alu_common.common.src[1]): @[unit_base.rs 88:13]
                            connect input_in_flight_op_src_values[1], {|HdlNone, HdlSome: Ty65|}(HdlNone) @[unit_base.rs 94:17]
                wire _bundle_literal_expr_27: Ty87
                connect _bundle_literal_expr_27.mop, _match_arm_value_15
                connect _bundle_literal_expr_27.src_values, input_in_flight_op_src_values
                connect input_in_flight_op, {|HdlNone, HdlSome: Ty87|}(HdlSome, _bundle_literal_expr_27) @[unit_base.rs 97:9]
        match in_flight_ops[0]: @[unit_base.rs 107:9]
            HdlNone:
                match input_index: @[unit_base.rs 122:16]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_19):
                        connect input_insn.ready, UInt<1>(0h1) @[unit_base.rs 123:13]
                        when eq(_match_arm_value_19, UInt<64>(0h0)): @[unit_base.rs 125:13]
                            connect in_flight_ops[0], input_in_flight_op @[unit_base.rs 126:17]
            HdlSome(_match_arm_value_20):
                wire firing_data_1: Ty70 @[ready_valid.rs 30:27]
                connect firing_data_1, {|HdlNone, HdlSome: Ty69|}(HdlNone) @[ready_valid.rs 31:9]
                when cancel_input.ready: @[ready_valid.rs 33:9]
                    connect firing_data_1, cancel_input.data @[ready_valid.rs 34:13]
                match firing_data_1: @[unit_base.rs 109:13]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_21):
                        wire dest_reg: Ty24 @[instruction.rs 502:1]
                        match _match_arm_value_20.mop: @[instruction.rs 502:1]
                            AddSub(_match_arm_value_22):
                                connect dest_reg, _match_arm_value_22.alu_common.common.dest @[instruction.rs 502:1]
                            AddSubI(_match_arm_value_23):
                                connect dest_reg, _match_arm_value_23.alu_common.common.dest @[instruction.rs 502:1]
                            Logical(_match_arm_value_24):
                                connect dest_reg, _match_arm_value_24.alu_common.common.dest @[instruction.rs 502:1]
                        when eq(_match_arm_value_21.which.value, dest_reg.value): @[unit_base.rs 113:17]
                            connect in_flight_ops[0], {|HdlNone, HdlSome: Ty87|}(HdlNone) @[unit_base.rs 115:21]
        match in_flight_ops[1]: @[unit_base.rs 107:9]
            HdlNone:
                match input_index: @[unit_base.rs 122:16]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_25):
                        connect input_insn.ready, UInt<1>(0h1) @[unit_base.rs 123:13]
                        when eq(_match_arm_value_25, UInt<64>(0h1)): @[unit_base.rs 125:13]
                            connect in_flight_ops[1], input_in_flight_op @[unit_base.rs 126:17]
            HdlSome(_match_arm_value_26):
                wire firing_data_2: Ty70 @[ready_valid.rs 30:27]
                connect firing_data_2, {|HdlNone, HdlSome: Ty69|}(HdlNone) @[ready_valid.rs 31:9]
                when cancel_input.ready: @[ready_valid.rs 33:9]
                    connect firing_data_2, cancel_input.data @[ready_valid.rs 34:13]
                match firing_data_2: @[unit_base.rs 109:13]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_27):
                        wire dest_reg_1: Ty24 @[instruction.rs 502:1]
                        match _match_arm_value_26.mop: @[instruction.rs 502:1]
                            AddSub(_match_arm_value_28):
                                connect dest_reg_1, _match_arm_value_28.alu_common.common.dest @[instruction.rs 502:1]
                            AddSubI(_match_arm_value_29):
                                connect dest_reg_1, _match_arm_value_29.alu_common.common.dest @[instruction.rs 502:1]
                            Logical(_match_arm_value_30):
                                connect dest_reg_1, _match_arm_value_30.alu_common.common.dest @[instruction.rs 502:1]
                        when eq(_match_arm_value_27.which.value, dest_reg_1.value): @[unit_base.rs 113:17]
                            connect in_flight_ops[1], {|HdlNone, HdlSome: Ty87|}(HdlNone) @[unit_base.rs 115:21]
        match in_flight_ops[2]: @[unit_base.rs 107:9]
            HdlNone:
                match input_index: @[unit_base.rs 122:16]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_31):
                        connect input_insn.ready, UInt<1>(0h1) @[unit_base.rs 123:13]
                        when eq(_match_arm_value_31, UInt<64>(0h2)): @[unit_base.rs 125:13]
                            connect in_flight_ops[2], input_in_flight_op @[unit_base.rs 126:17]
            HdlSome(_match_arm_value_32):
                wire firing_data_3: Ty70 @[ready_valid.rs 30:27]
                connect firing_data_3, {|HdlNone, HdlSome: Ty69|}(HdlNone) @[ready_valid.rs 31:9]
                when cancel_input.ready: @[ready_valid.rs 33:9]
                    connect firing_data_3, cancel_input.data @[ready_valid.rs 34:13]
                match firing_data_3: @[unit_base.rs 109:13]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_33):
                        wire dest_reg_2: Ty24 @[instruction.rs 502:1]
                        match _match_arm_value_32.mop: @[instruction.rs 502:1]
                            AddSub(_match_arm_value_34):
                                connect dest_reg_2, _match_arm_value_34.alu_common.common.dest @[instruction.rs 502:1]
                            AddSubI(_match_arm_value_35):
                                connect dest_reg_2, _match_arm_value_35.alu_common.common.dest @[instruction.rs 502:1]
                            Logical(_match_arm_value_36):
                                connect dest_reg_2, _match_arm_value_36.alu_common.common.dest @[instruction.rs 502:1]
                        when eq(_match_arm_value_33.which.value, dest_reg_2.value): @[unit_base.rs 113:17]
                            connect in_flight_ops[2], {|HdlNone, HdlSome: Ty87|}(HdlNone) @[unit_base.rs 115:21]
        match in_flight_ops[3]: @[unit_base.rs 107:9]
            HdlNone:
                match input_index: @[unit_base.rs 122:16]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_37):
                        connect input_insn.ready, UInt<1>(0h1) @[unit_base.rs 123:13]
                        when eq(_match_arm_value_37, UInt<64>(0h3)): @[unit_base.rs 125:13]
                            connect in_flight_ops[3], input_in_flight_op @[unit_base.rs 126:17]
            HdlSome(_match_arm_value_38):
                wire firing_data_4: Ty70 @[ready_valid.rs 30:27]
                connect firing_data_4, {|HdlNone, HdlSome: Ty69|}(HdlNone) @[ready_valid.rs 31:9]
                when cancel_input.ready: @[ready_valid.rs 33:9]
                    connect firing_data_4, cancel_input.data @[ready_valid.rs 34:13]
                match firing_data_4: @[unit_base.rs 109:13]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_39):
                        wire dest_reg_3: Ty24 @[instruction.rs 502:1]
                        match _match_arm_value_38.mop: @[instruction.rs 502:1]
                            AddSub(_match_arm_value_40):
                                connect dest_reg_3, _match_arm_value_40.alu_common.common.dest @[instruction.rs 502:1]
                            AddSubI(_match_arm_value_41):
                                connect dest_reg_3, _match_arm_value_41.alu_common.common.dest @[instruction.rs 502:1]
                            Logical(_match_arm_value_42):
                                connect dest_reg_3, _match_arm_value_42.alu_common.common.dest @[instruction.rs 502:1]
                        when eq(_match_arm_value_39.which.value, dest_reg_3.value): @[unit_base.rs 113:17]
                            connect in_flight_ops[3], {|HdlNone, HdlSome: Ty87|}(HdlNone) @[unit_base.rs 115:21]
        match in_flight_ops[4]: @[unit_base.rs 107:9]
            HdlNone:
                match input_index: @[unit_base.rs 122:16]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_43):
                        connect input_insn.ready, UInt<1>(0h1) @[unit_base.rs 123:13]
                        when eq(_match_arm_value_43, UInt<64>(0h4)): @[unit_base.rs 125:13]
                            connect in_flight_ops[4], input_in_flight_op @[unit_base.rs 126:17]
            HdlSome(_match_arm_value_44):
                wire firing_data_5: Ty70 @[ready_valid.rs 30:27]
                connect firing_data_5, {|HdlNone, HdlSome: Ty69|}(HdlNone) @[ready_valid.rs 31:9]
                when cancel_input.ready: @[ready_valid.rs 33:9]
                    connect firing_data_5, cancel_input.data @[ready_valid.rs 34:13]
                match firing_data_5: @[unit_base.rs 109:13]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_45):
                        wire dest_reg_4: Ty24 @[instruction.rs 502:1]
                        match _match_arm_value_44.mop: @[instruction.rs 502:1]
                            AddSub(_match_arm_value_46):
                                connect dest_reg_4, _match_arm_value_46.alu_common.common.dest @[instruction.rs 502:1]
                            AddSubI(_match_arm_value_47):
                                connect dest_reg_4, _match_arm_value_47.alu_common.common.dest @[instruction.rs 502:1]
                            Logical(_match_arm_value_48):
                                connect dest_reg_4, _match_arm_value_48.alu_common.common.dest @[instruction.rs 502:1]
                        when eq(_match_arm_value_45.which.value, dest_reg_4.value): @[unit_base.rs 113:17]
                            connect in_flight_ops[4], {|HdlNone, HdlSome: Ty87|}(HdlNone) @[unit_base.rs 115:21]
        match in_flight_ops[5]: @[unit_base.rs 107:9]
            HdlNone:
                match input_index: @[unit_base.rs 122:16]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_49):
                        connect input_insn.ready, UInt<1>(0h1) @[unit_base.rs 123:13]
                        when eq(_match_arm_value_49, UInt<64>(0h5)): @[unit_base.rs 125:13]
                            connect in_flight_ops[5], input_in_flight_op @[unit_base.rs 126:17]
            HdlSome(_match_arm_value_50):
                wire firing_data_6: Ty70 @[ready_valid.rs 30:27]
                connect firing_data_6, {|HdlNone, HdlSome: Ty69|}(HdlNone) @[ready_valid.rs 31:9]
                when cancel_input.ready: @[ready_valid.rs 33:9]
                    connect firing_data_6, cancel_input.data @[ready_valid.rs 34:13]
                match firing_data_6: @[unit_base.rs 109:13]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_51):
                        wire dest_reg_5: Ty24 @[instruction.rs 502:1]
                        match _match_arm_value_50.mop: @[instruction.rs 502:1]
                            AddSub(_match_arm_value_52):
                                connect dest_reg_5, _match_arm_value_52.alu_common.common.dest @[instruction.rs 502:1]
                            AddSubI(_match_arm_value_53):
                                connect dest_reg_5, _match_arm_value_53.alu_common.common.dest @[instruction.rs 502:1]
                            Logical(_match_arm_value_54):
                                connect dest_reg_5, _match_arm_value_54.alu_common.common.dest @[instruction.rs 502:1]
                        when eq(_match_arm_value_51.which.value, dest_reg_5.value): @[unit_base.rs 113:17]
                            connect in_flight_ops[5], {|HdlNone, HdlSome: Ty87|}(HdlNone) @[unit_base.rs 115:21]
        match in_flight_ops[6]: @[unit_base.rs 107:9]
            HdlNone:
                match input_index: @[unit_base.rs 122:16]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_55):
                        connect input_insn.ready, UInt<1>(0h1) @[unit_base.rs 123:13]
                        when eq(_match_arm_value_55, UInt<64>(0h6)): @[unit_base.rs 125:13]
                            connect in_flight_ops[6], input_in_flight_op @[unit_base.rs 126:17]
            HdlSome(_match_arm_value_56):
                wire firing_data_7: Ty70 @[ready_valid.rs 30:27]
                connect firing_data_7, {|HdlNone, HdlSome: Ty69|}(HdlNone) @[ready_valid.rs 31:9]
                when cancel_input.ready: @[ready_valid.rs 33:9]
                    connect firing_data_7, cancel_input.data @[ready_valid.rs 34:13]
                match firing_data_7: @[unit_base.rs 109:13]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_57):
                        wire dest_reg_6: Ty24 @[instruction.rs 502:1]
                        match _match_arm_value_56.mop: @[instruction.rs 502:1]
                            AddSub(_match_arm_value_58):
                                connect dest_reg_6, _match_arm_value_58.alu_common.common.dest @[instruction.rs 502:1]
                            AddSubI(_match_arm_value_59):
                                connect dest_reg_6, _match_arm_value_59.alu_common.common.dest @[instruction.rs 502:1]
                            Logical(_match_arm_value_60):
                                connect dest_reg_6, _match_arm_value_60.alu_common.common.dest @[instruction.rs 502:1]
                        when eq(_match_arm_value_57.which.value, dest_reg_6.value): @[unit_base.rs 113:17]
                            connect in_flight_ops[6], {|HdlNone, HdlSome: Ty87|}(HdlNone) @[unit_base.rs 115:21]
        match in_flight_ops[7]: @[unit_base.rs 107:9]
            HdlNone:
                match input_index: @[unit_base.rs 122:16]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_61):
                        connect input_insn.ready, UInt<1>(0h1) @[unit_base.rs 123:13]
                        when eq(_match_arm_value_61, UInt<64>(0h7)): @[unit_base.rs 125:13]
                            connect in_flight_ops[7], input_in_flight_op @[unit_base.rs 126:17]
            HdlSome(_match_arm_value_62):
                wire firing_data_8: Ty70 @[ready_valid.rs 30:27]
                connect firing_data_8, {|HdlNone, HdlSome: Ty69|}(HdlNone) @[ready_valid.rs 31:9]
                when cancel_input.ready: @[ready_valid.rs 33:9]
                    connect firing_data_8, cancel_input.data @[ready_valid.rs 34:13]
                match firing_data_8: @[unit_base.rs 109:13]
                    HdlNone:
                        skip
                    HdlSome(_match_arm_value_63):
                        wire dest_reg_7: Ty24 @[instruction.rs 502:1]
                        match _match_arm_value_62.mop: @[instruction.rs 502:1]
                            AddSub(_match_arm_value_64):
                                connect dest_reg_7, _match_arm_value_64.alu_common.common.dest @[instruction.rs 502:1]
                            AddSubI(_match_arm_value_65):
                                connect dest_reg_7, _match_arm_value_65.alu_common.common.dest @[instruction.rs 502:1]
                            Logical(_match_arm_value_66):
                                connect dest_reg_7, _match_arm_value_66.alu_common.common.dest @[instruction.rs 502:1]
                        when eq(_match_arm_value_63.which.value, dest_reg_7.value): @[unit_base.rs 113:17]
                            connect in_flight_ops[7], {|HdlNone, HdlSome: Ty87|}(HdlNone) @[unit_base.rs 115:21]
",
    };
    // let sim_debug = format!("{sim:#?}");
    // println!("#######\n{sim_debug}\n#######");
    // if sim_debug != include_str!("expected/reg_alloc.txt") {
    //     panic!();
    // }
}
